/**
 * Sortd
 * Copyright (c) 2019, Sortd Inc
 * https://www.sortd.com
 *
 * @fileoverview Event page for the Sortd extension
 * @name sortd-event-page.js
 */

'use strict';

let sortdErrorHandler = new SortdEventPageErrorHandler();
let sortdHttp = new SortdHttp();

let _sortActiveTab = {
	current: null,
	lastGmailTab: null
};
let _CSP = {};
let _tabs = {};

let _overrides = {};
let _config;
let _sortdLoginWindows = null;

if (typeof overrides !== 'undefined') {
	_overrides = overrides;
}

_config = new config(_overrides);

_CSP.hosts = _config.CSP.hosts.join(' ');
_CSP.iframeHosts = _config.CSP.iframeHosts.join(' ');

/**
 * Check if sortdLoginWindows object reference is set, if not load it from storage
 * @params {Function} next - The results callback
 */
let checkWindowReferences = function (next){

	if (!_sortdLoginWindows) {

		_sortdLoginWindows = {};

		try {

			chrome.storage.local.get('sortdLoginWindows', function (result) {

				if (!chrome.extension.lastError) {

					if (result && result.sortdLoginWindows) {
						_sortdLoginWindows = JSON.parse(result.sortdLoginWindows);
					}
				} else {
					sortdErrorHandler.reportBackgroundError(null, chrome.extension.lastError);
				}
				next(_sortdLoginWindows);
			});
		} catch (e) {
			sortdErrorHandler.reportBackgroundError(null, e);
			next(_sortdLoginWindows);
		}
	} else {
		next(_sortdLoginWindows);
	}
};

let saveWindowReferences = function(data, next) {

	data = (data !== null && typeof data !== 'undefined' && typeof data !== 'string') ?
		JSON.stringify(data) :
		data;

	chrome.storage.local.set({ 'sortdLoginWindows': data }, function() {
		if (next) next(chrome.extension.lastError || null);
	});
};

let trackUninstall = function (request, sender, sendResponse) {

	if (chrome.runtime.setUninstallURL) {
		chrome.runtime.setUninstallURL(request.data.url);
	}
	sendResponse(null);
};

let showLogin = function (request, sender, sendResponse, next) {

	try {

		let w = 550,
			h = 715,
			left = Math.round((screen.width / 2) - (w / 2)),
			top = Math.round((screen.height / 2) - (h / 2)),
			requestData = request.data,
			url = requestData.url;

		if (requestData.loginHint) {
			url += url.indexOf('?') === -1 ? '?' : '&';
			url += 'hint=' + encodeURIComponent(requestData.loginHint);
		}

		if (requestData.forceOAuth) {
			url += url.indexOf('?') === -1 ? '?' : '&';
			url += 'prompt=consent';
		}

		if (requestData.acceptedTerms) {
			url += url.indexOf('?') === -1 ? '?' : '&';
			url += 'acctms=Y';
		}

		if (requestData.fullGmailScope) {
			url += url.indexOf('?') === -1 ? '?' : '&';
			url += 'full=1';
		}

		console.log(url);

		chrome.windows.create({
			url,
//			focused: true,
			type: 'popup',
			height: h,
			width: w,
			left: left,
			top: top
		}, function (window) {

			try {
				let windowId = window.id.toString();

				if (!chrome.extension.lastError) {

					checkWindowReferences(function (sortdLoginWindows) {

						try {
							sortdLoginWindows[windowId] = {
								loginWindow: windowId,
								loginTab: window.tabs[0].id,
								sortdTab: sender.tab.id,
								baseUrl: requestData.baseUrl,
								authUrl: requestData.authUrl,
								url: requestData.url,
								fullscreen: (requestData.fullscreen === false) ? 'false' : 'true',
								closedLoginWindowWithoutAuth: 'true'
							};

							saveWindowReferences(sortdLoginWindows, function(err) {

								if (err) {
									sortdErrorHandler.reportBackgroundError(sender.tab.id, err);
								}
								sendResponse({ windowId: window.id });
								if (next) next(err || null);
							});
						} catch (ex) {
							sortdErrorHandler.reportBackgroundError(null, ex);
							if (next) next(ex);
						}
					});
				} else {
					sortdErrorHandler.reportBackgroundError(sender.tab.id, chrome.extension.lastError);
					if (next) next(chrome.extension.lastError);
				}
			} catch (ex) {
				sortdErrorHandler.reportBackgroundError(null, ex);
				if (next) next(ex);
			}
		});
	} catch(error) {
		sortdErrorHandler.reportBackgroundError(sender.tab.id, error);
		if (next) next(error);
	}
};

let afterLogin = function (request, sender, sendResponse) {

	const authTokenId = request.authTokenId;
	const hashedUser = request.hashedUserId;

	try {

		let windowId = sender.tab.windowId.toString();

		checkWindowReferences(function (sortdLoginWindows) {

			try {
				let loginData = sortdLoginWindows[windowId];

				if (loginData && loginData.loginTab && loginData.sortdTab) {

					let sortdTab = loginData.sortdTab;

					chrome.tabs.remove(sender.tab.id, function () {

						if (chrome.extension.lastError) {
							sortdErrorHandler.reportBackgroundError(sortdTab, chrome.extension.lastError);
						}
					});

					chrome.tabs.sendMessage(
						sortdTab, {
							action: 'afterLogin',
							data: {
								authTokenId: authTokenId,
								hashedUser: hashedUser,
								fullscreen: !(loginData.fullscreen === 'false')
							}
						},
						function () {

							if (chrome.extension.lastError) {
								sortdErrorHandler.reportBackgroundError(sortdTab, chrome.extension.lastError);
							}
						});
				}

				if (windowId === loginData.loginWindow) {

					loginData.closedLoginWindowWithoutAuth = false;
					sortdLoginWindows[windowId] = loginData;
					saveWindowReferences(sortdLoginWindows, function(_err) {
						// noop
					});
				}
			} catch (ex) {
				sortdErrorHandler.reportBackgroundError(null, ex);
			}
		});
		sendResponse({});
	}
	catch (error) {
		sortdErrorHandler.reportBackgroundError(null, error);
	}
};

let oAuthFailed = function (request, sender, sendResponse) {

	try {

		let windowId = sender.tab.windowId.toString();

		checkWindowReferences(function (sortdLoginWindows) {

			try {
				let loginData = sortdLoginWindows[windowId];

				if (loginData && loginData.loginTab && loginData.sortdTab) {

					let sortdTab = loginData.sortdTab;

					chrome.tabs.remove(sender.tab.id, function () {

						if (chrome.extension.lastError) {
							sortdErrorHandler.reportBackgroundError(sortdTab, chrome.extension.lastError);
						}
					});

					chrome.tabs.sendMessage(
						sortdTab, {
							action: 'oAuthFailed',
							data: { }
						},
						function () {

							if (chrome.extension.lastError) {
								sortdErrorHandler.reportBackgroundError(sortdTab, chrome.extension.lastError);
							}
						});
				}

				if (windowId === loginData.loginWindow) {

					loginData.closedLoginWindowWithoutAuth = false;
					sortdLoginWindows[windowId] = loginData;
					saveWindowReferences(sortdLoginWindows, function(_err) {
						// noop
					});
				}
			} catch (ex) {
				sortdErrorHandler.reportBackgroundError(null, ex);
			}
		});
		sendResponse({});
	} catch (error) {
		sortdErrorHandler.reportBackgroundError(null, error);
	}
};

let showAdditionalOAuthPermissionsRequest = function (request, sender, sendResponse, next) {

	try {

		let w = 550,
			h = 715,
			left = Math.round((screen.width / 2) - (w / 2)),
			top = Math.round((screen.height / 2) - (h / 2)),
			requestData = request.data,
			url = requestData.url,
			permission = requestData.permission;

		if (requestData.loginHint) {
			url += url.indexOf('?') === -1 ? '?' : '&';
			url += '&hint=' + encodeURIComponent(requestData.loginHint);
		}

		url += url.indexOf('?') === -1 ? '?' : '&';
		url += 'permission=' + permission;
		url += '&hd=*';
		url += '&prompt=consent';
		url += '&include_granted_scopes=true';

		chrome.windows.create({
			url,
			type: 'popup',
			height: h,
			width: w,
			left: left,
			top: top
		}, function (window) {

			try {
				let windowId = window.id.toString();

				if (!chrome.extension.lastError) {

					checkWindowReferences(function (sortdLoginWindows) {

						try {
							sortdLoginWindows[windowId] = {
								loginWindow: windowId,
								loginTab: window.tabs[0].id,
								sortdTab: sender.tab.id,
								baseUrl: requestData.baseUrl,
								authUrl: requestData.authUrl,
								url: requestData.url,
								fullscreen: (requestData.fullscreen === false) ? 'false' : 'true',
								closedLoginWindowWithoutAuth: 'true'
							};

							saveWindowReferences(sortdLoginWindows, function(err) {

								if (err) {
									sortdErrorHandler.reportBackgroundError(sender.tab.id, err);
								}
								sendResponse({ windowId: window.id });
								if (next) next(err || null);
							});
						} catch (ex) {
							sortdErrorHandler.reportBackgroundError(null, ex);
							if (next) next(ex);
						}
					});
				} else {
					sortdErrorHandler.reportBackgroundError(sender.tab.id, chrome.extension.lastError);
					if (next) next(chrome.extension.lastError);
				}
			} catch (ex) {
				sortdErrorHandler.reportBackgroundError(null, ex);
				if (next) next(ex);
			}
		});
	} catch(error) {
		sortdErrorHandler.reportBackgroundError(sender.tab.id, error);
		if (next) next(error);
	}
};

let saveUserSettings = function (request, sender, sendResponse, next) {

	try {

		let obj = {},
			key = request.data.email + '/settings'.toLowerCase(),
			settings = request.data.settings;

		obj[key] = settings;

		chrome.storage.sync.set(obj, function () {

			if (chrome.extension.lastError) {
				sortdErrorHandler.reportBackgroundError(_sortActiveTab.current, chrome.extension.lastError);
			}
			sendResponse(null);
			if (next) next(chrome.extension.lastError || null);
		});
	} catch (ex) {
		sortdErrorHandler.reportBackgroundError(_sortActiveTab.current, ex);
		sendResponse(null);
		if (next) next(ex)
	}
};

let getUserSettings = function (request, sender, sendResponse) {

	try {
		let key = request.data.email + '/settings'.toLowerCase();

		chrome.storage.sync.get(key, function (result) {

			try {
				let settings = null;

				if (!chrome.extension.lastError) {

					if (result && result[key]) {
						settings = result[key];
					}
					sendResponse(settings);
				} else {
					sortdErrorHandler.reportBackgroundError(null, chrome.extension.lastError);
					sendResponse(null);
				}
			} catch (ex) {
				sortdErrorHandler.reportBackgroundError(null, ex);
				sendResponse(null);
			}
		});
	} catch (ex) {
		sortdErrorHandler.reportBackgroundError(_sortActiveTab.current, ex);
		sendResponse(null);
	}
};

let saveExtensionSettings = function (request, sender, sendResponse, next) {

	try {

		let obj = {},
			key = 'extension/settings',
			settings = request.data;

		obj[key] = settings;

		chrome.storage.sync.set(obj, function () {

			try {
				if (chrome.extension.lastError) {
					sortdErrorHandler.reportBackgroundError(_sortActiveTab.current, chrome.extension.lastError);
				} else if (settings) {

					if (settings.cspHosts) {
						_CSP.hosts = settings.cspHosts.join(' ');
					}

					if (settings.cspIframeHosts) {
						_CSP.iframeHosts = settings.cspIframeHosts.join(' ');
					}
				}

				sendResponse(null);
				if (next) next(chrome.extension.lastError || null);
			} catch (ex) {
				sortdErrorHandler.reportBackgroundError(null, ex);
				sendResponse(null);
				if (next) next(null);
			}
		});
	} catch (ex) {
		sortdErrorHandler.reportBackgroundError(_sortActiveTab.current, ex);
		sendResponse(null);
		if (next) next(ex);
	}
};

let getExtensionSettings = function (request, sender, sendResponse) {

	try {
		let key = 'extension/settings';

		chrome.storage.sync.get(key, function (result) {

			try {
				let settings = null;

				if (!chrome.extension.lastError) {

					if (result && result[key]) {
						settings = result[key];
					}
					sendResponse(settings);
				} else {
					sortdErrorHandler.reportBackgroundError(null, chrome.extension.lastError);
					sendResponse(null);
				}
			} catch (ex) {
				sortdErrorHandler.reportBackgroundError(null, ex);
				sendResponse(null);
			}
		});
	} catch (ex) {
		sortdErrorHandler.reportBackgroundError(_sortActiveTab.current, ex);
		sendResponse(null);
	}
};

let setSortdUser = function (request, sender, sendResponse) {

	try {
		if (sender && sender.tab && sender.tab.id && request.data && request.data.hashedUser) {

			if (!_tabs[sender.tab.id]) {
				_tabs[sender.tab.id] = {
					users: [],
					blockImgUrl: []
				};
			}

			const tab = _tabs[sender.tab.id];

			if (tab.users.indexOf(request.data.hashedUser) === -1) {
				const re = new RegExp('https?:\/\/' + _config.chrome.host.mail3.realm.replace(/\./g, '\\.') + '[^\\/]*\\/rr\\/[a-z0-9\\-]*?\\?from=' + request.data.hashedUser + '$');

				tab.users.push(request.data.hashedUser);
				tab.blockImgUrl.push(re);
			}
		}
	} catch (ex) {
		sortdErrorHandler.reportBackgroundError(_sortActiveTab.current, ex);
	}
};

let hasCePermissions = function(request, sender, sendResponse) {

	try {
		const permissions = request.data && request.data.permissions;
		const origins = request.data && request.data.origins;

		if (permissions || origins) {
			chrome.permissions.contains({ permissions, origins }, (result) => {
				sendResponse(result);
			});
		} else {
			sendResponse(null);
		}
	} catch (ex) {
		sortdErrorHandler.reportBackgroundError(null, ex);
		sendResponse(null);
	}
};

let requestCePermissions = function(request, sender, sendResponse) {

	try {
		const permissions = request.data && request.data.permissions;
		const origins = request.data && request.data.origins;

		if (permissions || origins) {
			chrome.permissions.request({ permissions, origins }, (result) => {
				sendResponse(result);
			});
		} else {
			sendResponse(null);
		}
	} catch (ex) {
		sortdErrorHandler.reportBackgroundError(null, ex);
		sendResponse(null);
	}
};

let createNotification = function(request, sender, sendResponse) {

	try {
		if (request.data && request.data.options) {

			chrome.permissions.contains({ permissions: ['notifications'] }, (result) => {

				if (result) {
					try {
						chrome.notifications.create(request.data.id, request.data.options, (id) => {
							sendResponse(id);
						});
					} catch (ex) {
						sortdErrorHandler.reportBackgroundError(null, ex);
						sendResponse(null);
					}
				} else {
					sendResponse(null);
				}
			});

		}
	} catch (ex) {
		sortdErrorHandler.reportBackgroundError(null, ex);
		sendResponse(null);
	}
};

let saveLinkedAccounts = function (request, sender, sendResponse, next) {

	try {

		const obj = {
			'sortd/linkedaccounts': request.data
		};

		chrome.storage.sync.set(obj, function () {

			if (chrome.extension.lastError) {
				sortdErrorHandler.reportBackgroundError(_sortActiveTab.current, chrome.extension.lastError);
			}
			sendResponse(null);
			if (next) next(chrome.extension.lastError || null);
		});
	} catch (ex) {
		sortdErrorHandler.reportBackgroundError(_sortActiveTab.current, ex);
		sendResponse(null);
		if (next) next(ex)
	}
};

let getLinkedAccounts = function (request, sender, sendResponse) {

	try {
		const key = 'sortd/linkedaccounts';

		chrome.storage.sync.get(key, function (result) {

			try {
				let linkedAccounts = null;

				if (!chrome.extension.lastError) {

					if (result && result[key]) {
						linkedAccounts = result[key];
					}
					sendResponse(linkedAccounts);
				} else {
					sortdErrorHandler.reportBackgroundError(null, chrome.extension.lastError);
					sendResponse(null);
				}
			} catch (ex) {
				sortdErrorHandler.reportBackgroundError(null, ex);
				sendResponse(null);
			}
		});
	} catch (ex) {
		sortdErrorHandler.reportBackgroundError(_sortActiveTab.current, ex);
		sendResponse(null);
	}
};

const createCallbackWithTimeout = function(sendResponse) {

	let timer = setTimeout(() => {
		timer = null;
		console.log('Message timed out');
		sendResponse({ err: new Error('Message timed out') });
	}, 60000);

	const callback = (err, result) => {
		if (timer) {
			clearTimeout(timer);
			sendResponse({ err, result });
		}
	};

	return callback;
};

let handleMessage = function (request, sender, sendResponse) {

	try {

		switch (request.action) {
			case 'showLogin':
				showLogin(request, sender, sendResponse);
				break;
			case 'showAdditionalOAuthPermissions':
				showAdditionalOAuthPermissionsRequest(request, sender, sendResponse);
				break;
			case 'afterLogin':
				afterLogin(request, sender, sendResponse);
				break;
			case 'oAuthFailed':
				oAuthFailed(request, sender, sendResponse);
				break;
			case 'trackUninstall':
				trackUninstall(request, sender, sendResponse);
				break;
			case 'saveUserSettings':
				saveUserSettings(request, sender, sendResponse);
				break;
			case 'getUserSettings':
				getUserSettings(request, sender, sendResponse);
				break;
			case 'saveExtensionSettings':
				saveExtensionSettings(request, sender, sendResponse);
				break;
			case 'getExtensionSettings':
				getExtensionSettings(request, sender, sendResponse);
				break;
			case 'setSortdUser':
				setSortdUser(request, sender, sendResponse);
				break;
			case 'hasCePermissions':
				hasCePermissions(request, sender, sendResponse);
				break;
			case 'requestCePermissions':
				requestCePermissions(request, sender, sendResponse);
				break;
			case 'createNotification':
				createNotification(request, sender, sendResponse);
				break;
			case 'saveLinkedAccounts':
				saveLinkedAccounts(request, sender, sendResponse);
				break;
			case 'getLinkedAccounts':
				getLinkedAccounts(request, sender, sendResponse);
				break;
			case 'httpGet':
				sortdHttp.get(request.data.url, request.data.params, createCallbackWithTimeout(sendResponse), request.data.retries);
				break;
			case 'httpPost':			
				sortdHttp.post(request.data.url, request.data.params, createCallbackWithTimeout(sendResponse));
				break;
			default:
				break;
		}
		return true;
	} catch (ex) {
		sendResponse({ err: ex });
		sortdErrorHandler.reportBackgroundError(null, ex);
	}
};

let closeTab = function (tabId, removeInfo) {

	try {
		let windowId = removeInfo.windowId.toString();

		if (_tabs[tabId]) {
			delete _tabs[tabId];
		}

		checkWindowReferences(function(sortdLoginWindows) {

			try {
				let loginData = sortdLoginWindows[windowId],
					sortdTab;

				if (loginData && windowId === loginData.loginWindow && loginData.closedLoginWindowWithoutAuth) {

					sortdTab = loginData.sortdTab;

					chrome.tabs.sendMessage(
						sortdTab, {
							action: 'noAuthAttempt',
							data: loginData
						}, function () {

							if (chrome.extension.lastError) {
								sortdErrorHandler.reportBackgroundError(sortdTab, chrome.extension.lastError);
							}
						});
				}
			} catch (ex) {
				sortdErrorHandler.reportBackgroundError(null, ex);
			}
		});

		if (_sortActiveTab.lastGmailTab == tabId) {
			_sortActiveTab.lastGmailTab = null;
		}

		removeWindow(windowId);
	} catch (ex) {
		sortdErrorHandler.reportBackgroundError(null, ex);
	}
};

let removeWindow = function(windowId, next) {

	try {

		checkWindowReferences(function (sortdLoginWindows) {

			try {
				delete sortdLoginWindows[windowId];

				saveWindowReferences(sortdLoginWindows, function (err) {

					if (err) {
						sortdErrorHandler.reportBackgroundError(_sortActiveTab.current, err);
					}
					if (next) next(err || null);
				});
			} catch (ex) {
				sortdErrorHandler.reportBackgroundError(null, ex);
				if (next) next(ex);
			}
		});
	} catch (e) {
		sortdErrorHandler.reportBackgroundError(null, new Error('Cannot remove window reference from storage: ' + windowId));
		if (next) next(e);
	}
};

let loadGoogleTagManager = function () {

	try {
		if (_config.tracker && _config.tracker.googletagmanager) {
			let documentHead = (document.head || document.documentElement);
			let script = document.createElement('script');

			script.src = 'https://www.googletagmanager.com/gtag/js?id=' + _config.tracker.googletagmanager.token;
			script.async = true;
			documentHead.appendChild(script);
			window.dataLayer = window.dataLayer || [];
			window.gtag = function(){dataLayer.push(arguments);};
			gtag('js', new Date());
			gtag('config', _config.tracker.googletagmanager.token);
		}
	} catch (ex) {
		sortdErrorHandler.reportBackgroundError(null, ex);
	}
};

let loadFacebookPixel = function () {

	try {
		if (_config.tracker && _config.tracker.facebook) {
			
			!function(f,b,e,v,n,t,s){if(f.fbq)return;n=f.fbq=function(){n.callMethod?
			n.callMethod.apply(n,arguments):n.queue.push(arguments)};
			if(!f._fbq)f._fbq=n;n.push=n;n.loaded=!0;n.version='2.0';
			n.queue=[];t=b.createElement(e);t.async=!0;
			t.src=v;s=b.getElementsByTagName(e)[0];
			s.parentNode.insertBefore(t,s)}(window, document, 'script',
			'https://connect.facebook.net/en_US/fbevents.js');

			fbq.disablePushState = true;
			fbq('init', _config.tracker.facebook.token);
		}
	} catch (ex) {
		sortdErrorHandler.reportBackgroundError(null, ex);
	}
};

let onInstalled = function (details) {

	try {
		chrome.windows.getAll({ populate: true }, function (windows) {

			try {
				let currentWindow,
					tab,
					gmailLoaded = false,
					i, j;

				for (i = 0; i < windows.length; i++) {
					currentWindow = windows[i];

					for (j = 0; j < currentWindow.tabs.length; j++) {
						tab = currentWindow.tabs[j];
						if (tab.url.match(/https:\/\/mail\.google\.com\/*/gi)) {
							chrome.tabs.reload(tab.id);
							if (!gmailLoaded) {
								chrome.tabs.update(tab.id, {selected: true});
								gmailLoaded = true;
							}
						}
					}
				}

				if (!gmailLoaded) {
					chrome.tabs.create({url: 'https://mail.google.com'});
				}
			} catch (ex) {
				sortdErrorHandler.reportBackgroundError(null, ex);
			}
		});

		if (_config.tracker && _config.tracker.googletagmanager && _config.tracker.googletagmanager.enabled && _config.tracker.googletagmanager.conversions) {
			gtag('event', 'installed_ce', {
				'send_to': _config.tracker.googletagmanager.conversions.installed_ce
			});
		}
		if (_config.tracker && _config.tracker.facebook && _config.tracker.facebook.enabled) {
			const options = {};
			if (_config.tracker.facebook.testToken) {
				options.test_event_code = _config.tracker.facebook.testToken;
			}
			fbq('trackCustom', 'InstalledCE', options);
		}
	} catch(err) {
		sortdErrorHandler.reportBackgroundError(null, new Error('Error while installing application'));
	}
};

let onUpdated = function (details) {

	try {
		saveWindowReferences(null, function(err) {

			if (err) {
				sortdErrorHandler.reportBackgroundError(_sortActiveTab.current, err);
			}
		});

		// let thisVersion = chrome.runtime.getManifest().version;

		chrome.windows.getAll({
			populate: true
		}, function (windows) {

			try {
				let w = windows.length,
					currentWindow,
					t,
					currentTab;

				for (let i = 0; i < w; i++) {
					currentWindow = windows[i];
					t = currentWindow.tabs.length;

					for (let j = 0; j < t; j++) {
						currentTab = currentWindow.tabs[j];
						// Only send to gmail pages
						if (currentTab.url.match(/https:\/\/mail\.google\.com\/*/gi)) {
							chrome.tabs.executeScript(currentTab.id, {code: "window.postMessage({ owner: 'sortd', source: 'bgpage', destination: 'jsce', action: 'extensionUpdated'}, '" + _config.messageOriginGmail + "')"});
						}
					}
				}
			} catch (ex) {
				sortdErrorHandler.reportBackgroundError(null, ex);
			}
		});
	} catch(err) {
		sortdErrorHandler.reportBackgroundError(null, new Error('Error while trying to upgrade application'));
	}
};

/**
 * Load CSP settings from the extension storage
 */
let loadCSPSettings = function () {

	try {
		getExtensionSettings(null, null, function (settings) {

			try {
				if (settings) {
					if (settings.cspHosts) {
						_CSP.hosts = settings.cspHosts.join(' ');
					}
					if (settings.cspIframeHosts) {
						_CSP.iframeHosts = settings.cspIframeHosts.join(' ');
					}
				}
			} catch (ex) {
				sortdErrorHandler.reportBackgroundError(null, ex);
			}
		});
	} catch (ex) {
		sortdErrorHandler.reportBackgroundError(null, ex);
	}
};

/**
 * Add Sortd to any Content Security Policy whitelist headers.
 */
let rewriteCSPHeaders = function () {

	try {
		let headerPattern = /(content-security-policy|^x-.*-csp)$/i;

		if (chrome.webRequest && chrome.webRequest.onHeadersReceived) {

			chrome.webRequest.onHeadersReceived.addListener(function (details) {

				try {
					let replaceHeaders = false;
					let method = details && details.method && details.method.toUpperCase();

					if (details && details.responseHeaders && (!method || method === 'GET')) {
						details.responseHeaders.forEach(function (header) {

							if (headerPattern.test(header.name)) {
								replaceHeaders = true;

								['default-src', 'style-src', 'script-src', 'connect-src', 'img-src', 'font-src', 'media-src'].forEach(function (section) {
									header.value = header.value.replace(section + ' ', section + ' ' + _CSP.hosts + ' ');
								});
								['frame-src', 'child-src'].forEach(function (section) {
									header.value = header.value.replace(section + ' ', section + ' ' + _CSP.iframeHosts + ' ');
								});
							}
						});
					}

					if (replaceHeaders) {
						return { responseHeaders: details.responseHeaders };
					}
				} catch (ex) {
					sortdErrorHandler.reportBackgroundError(null, ex);
				}
			}, {
				urls: ['https://mail.google.com/*'],
				types: ['main_frame', 'xmlhttprequest', 'websocket']
			}, ['responseHeaders', 'blocking']);
		}
	} catch (ex) {
		sortdErrorHandler.reportBackgroundError(null, ex);
	}
};

let checkImageRequests = function () {

	try {
		if (chrome.webRequest && chrome.webRequest.onBeforeRequest) {

			chrome.webRequest.onBeforeRequest.addListener(
				(details) => {

					try {
						if (details.method !== 'GET') {
							return;
						}

						let blockRequest = false;
						const tab = _tabs[details.tabId];
						if (tab) {
							blockRequest = tab.blockImgUrl.some((blockUrl) => blockUrl.test(details.url));
						}

						if (blockRequest) {
							return {
								redirectUrl: 'data:image/svg+xml,<svg xmlns="http://www.w3.org/2000/svg"/>'
							};
						}
					} catch (ex) {
						sortdErrorHandler.reportBackgroundError(null, ex);
					}
				},
				{
					urls: [
						"*://*.googleusercontent.com/*",
						"*://*.sortd.com/*"
					],
					types: ["image"]
				},
				["blocking"]
			);
		}
	} catch (ex) {
		sortdErrorHandler.reportBackgroundError(null, ex);
	}
};

try {
	if (chrome) {
//		if (chrome && chrome.storage && chrome.storage.sync) {
//			loadCSPSettings();
//		}
//		rewriteCSPHeaders();
		checkImageRequests();
		loadGoogleTagManager();
		loadFacebookPixel();

		if (chrome.tabs) {
			chrome.tabs.onRemoved.addListener(closeTab);
		}

		if (chrome.runtime) {
			if (chrome.runtime.onMessage) {
				chrome.runtime.onMessage.addListener(handleMessage);
			}

			if (chrome.runtime.onConnect && chrome.runtime.onConnect.addListener) {
				chrome.runtime.onConnect.addListener(() => {});
			}

			if (chrome.runtime.onInstalled) {
				chrome.runtime.onInstalled.addListener(function (details) {
					if (details.reason == 'install') {
						onInstalled(details);
					} else if (details.reason == 'update') {
						onUpdated(details);
					}
				});
			}
		}
	}
} catch (ex) {
	sortdErrorHandler.reportBackgroundError(null, ex);
}