/**
 * Sortd
 * Copyright (c) 2019, Sortd Inc
 * https://www.sortd.com
 *
 * @fileoverview Perform http requests to the server
 * @name sortd-http
 */

var SortdHttp = (function () {

	'use strict';

	/**
	 * Initialize the class
	 * @class SortdHttp
	 */
	function SortdHttp () {
		this.FILE_NAME = 'sortd-http.js';
	}

	/**
	 * Get an XMLHttpRequest object
	 * @private
	 */
	SortdHttp.prototype._getXMLHttpRequest = function () {
		return new XMLHttpRequest();
	};

	/**
	 * Do an XMLHttpRequest get request
	 * @param {String} url - The url to get
	 * @param {Object} [params] - Optional. JSON object with key:value pairs to send with GET request
	 * @param {Function} next - The results callback
	 * @private
	 */
	SortdHttp.prototype.get = function (url, params, next, retries) {

		try {
			const xhr = this._getXMLHttpRequest();

			if (retries === undefined) {
				retries = 3;
			}

			xhr.onreadystatechange = () => {

				try {
					if (xhr.readyState < 4) {
						return;
					} else if (xhr.status === 0 && retries >= 0) {
						setTimeout(() => {
							retries--;
							this.get(url, params, next, retries);
						}, 1000)
						return;
					} else if (xhr.status !== 200) {
						const error = new Error('Error loading resource');
						error.status = xhr.status;
						error.statusText = xhr.statusText;
						error.url = url;
						next(error);
					} else {
						try {
							let result;
							if (xhr.responseText) {
								result = JSON.parse(xhr.responseText);
							} else {
								result = null;
							}
							next(null, result);
						} catch (ex) {
							next(ex, xhr && xhr.responseText);
						}
					}
				} catch (ex) {
					next(ex);
				}
			};

			if (params) {
				for (var prop in params) {
					url += (url.indexOf('?') === -1) ? '?' : '&';
					url += prop + '=' + encodeURIComponent(params[prop]);
				}
			}

			xhr.withCredentials = true;
			xhr.open('GET', url, true);
			xhr.setRequestHeader('Content-Type', 'application/json;charset=UTF-8');
			xhr.send();
		} catch (ex) {
			next(ex);
		}
	};

	/**
	 * Do an XMLHttpRequest post request
	 * @param {String} url - The url to post to
	 * @param {Object} [params] - Optional. JSON object with key:value pairs to send with POST request
	 * @param {Function} next - The results callback
	 * @private
	 */
	SortdHttp.prototype.post = function (url, params, next) {

		try {
			const xhr = this._getXMLHttpRequest();

			xhr.onreadystatechange = function () {

				try {
					if (xhr.readyState < 4) {
						return;
					} else if (xhr.status !== 200) {
						var error = new Error('Error loading resource');
						error.status = xhr.status;
						error.statusText = xhr.statusText;
						error.url = url;
						next(error);
					} else {
						try {
							let result;
							if (xhr.responseText) {
								result = JSON.parse(xhr.responseText);
							} else {
								result = null;
							}
							next(null, result);
						} catch (ex) {
							next(ex, xhr && xhr.responseText);
						}
					}
				} catch (ex) {
					next(ex);
				}
			};

			xhr.withCredentials = true;
			xhr.open('POST', url, true);
			xhr.setRequestHeader('Content-Type', 'application/json;charset=UTF-8');
			xhr.send(JSON.stringify(params));
		} catch (ex) {
			next(ex);
		}
	};

	return SortdHttp;
}());