/**
 * Sortd
 * Copyright (c) 2019, Sortd Inc
 * https://www.sortd.com
 *
 * @fileoverview Interact with Gmail DOM
 * @name dom
 */

class GmailApiDom {

	constructor(options) {

		this.__FILE_NAME = 'dom.js';

		this.config = options.config;

		this.shortcutHelper = SortdShortcutHelper.getInstance();

		this.activeNavigationCount = 0;
		this.activeNavigationIndex = 0;
		this.activeNavigationElement = null;
		this.activeNavigationOnDone = null;

		this.boundUpArrow = this.upArrow.bind(this);
		this.boundDownArrow = this.downArrow.bind(this);
		this.boundOnSelect = this.onSelect.bind(this);
	}

	onSelect(event) {

		if (event) {
			event.stopPropagation();
			event.preventDefault();
		}

		const activeElement = this.activeNavigationElement.querySelector('.active[data-sd-selection-index]');
		if (activeElement) {
			activeElement.click();
		}

		if (this.activeNavigationOnDone) {
			this.activeNavigationOnDone();
		}
	}

	updateSelection() {

		const selectedIndex = this.activeNavigationIndex.toString();
		this.activeNavigationElement.querySelectorAll('[data-sd-selection-index]').forEach((el) => {

			if (el.dataset.sdSelectionIndex === selectedIndex) {
				el.classList.add('active');
			} else {
				el.classList.remove('active');
			}
		});
	}

	arrowPressed(event, direction) {

		if (event) {
			event.stopPropagation();
			event.preventDefault();
		}

		this.activeNavigationIndex += direction;

		if (this.activeNavigationIndex >= this.activeNavigationCount) {
			this.activeNavigationIndex = 0;
		}

		if (this.activeNavigationIndex < 0) {
			this.activeNavigationIndex = this.activeNavigationCount - 1;
		}

		this.updateSelection();
	};

	upArrow(event) {
		this.arrowPressed(event, -1);
	}

	downArrow(event) {
		this.arrowPressed(event, 1);
	}

	bindArrowNavigation(el, onCancel, onDone, defaultSelected = 0) {

		this.activeNavigationIndex = defaultSelected;
		this.activeNavigationElement = el;
		this.activeNavigationOnDone = onDone;

		const elements = this.activeNavigationElement.querySelectorAll('[data-sd-selection-index]');
		this.activeNavigationCount = elements.length;

		this.updateSelection();

		el.addEventListener('mouseover', (evt) => {

			let mouseOverEl;
			if (evt.target.nodeName === 'LI' && evt.target.hasAttribute('data-sd-selection-index')) {
				mouseOverEl = evt.target;
			} else {
				mouseOverEl = evt.target.closest('li[data-sd-selection-index]');
			}

			if (mouseOverEl) {
				const hoverIndex = parseInt(mouseOverEl.getAttribute('data-sd-selection-index'), 10);
				if (this.activeNavigationIndex !== hoverIndex) {
					this.activeNavigationIndex = hoverIndex;
					this.updateSelection();
				}
			}
		});

		this.shortcutHelper.addListener(
			constants.SHORTCUTS.KEY_CODES.ESCAPE,
			onCancel);

		this.shortcutHelper.addListener(
			constants.SHORTCUTS.KEY_CODES.ENTER,
			this.boundOnSelect);

		this.shortcutHelper.addListener(
			constants.SHORTCUTS.KEY_CODES.UP_ARROW,
			this.boundUpArrow);

		this.shortcutHelper.addListener(
			constants.SHORTCUTS.KEY_CODES.DOWN_ARROW,
			this.boundDownArrow);
	}

	unbindArrowNavigation(onCancel) {

		this.activeNavigationCount = 0;
		this.activeNavigationIndex = 0;
		this.activeNavigationElement = null;
		this.activeNavigationOnDone = null;

		this.shortcutHelper.removeListener(
			constants.SHORTCUTS.KEY_CODES.ESCAPE,
			onCancel);

		this.shortcutHelper.removeListener(
			constants.SHORTCUTS.KEY_CODES.ENTER,
			this.boundOnSelect);

		this.shortcutHelper.removeListener(
			constants.SHORTCUTS.KEY_CODES.UP_ARROW,
			this.boundUpArrow);

		this.shortcutHelper.removeListener(
			constants.SHORTCUTS.KEY_CODES.DOWN_ARROW,
			this.boundDownArrow);
	}

    /**
     * Observes an HTML element for changes
     * @param {HTMLElement} element - The HTML element to observer
     * @param {Object} options - The options object
     * @returns {Observable} - Stream of element mutations
     */
	observeMutations(element, options) {

		return Kefir.stream((emitter) => {

			const observer = new MutationObserver((mutations) => {
				emitter.emit(mutations instanceof Array ? mutations : Array.from(mutations));
			});

			observer.observe(element, options);

			return () => {
				observer.disconnect();
			};
		});
	}

	selectorFromClassArray(classes) {
		return `.${classes.join('.')}`;
	}

	createVertSeparatorElement() {

		const el = document.createElement('span');
		el.classList.add('sd_vert_separator');

		return el;
	}

	createHorzSeparatorElement() {

		const el = document.createElement('div');
		el.classList.add('sd_horz_separator');

		return el;
	}

	createTextElement(text) {

		const el = document.createElement('span');
		el.textContent = text;
		el.classList.add('sortdText');

		return el;
	}

    /**
     * Create a new checbox element
     * @param {string} text - The text to appear next to the checkbox
     * @param {Function} onClick - The on click callback function
     * @param {string[]} [labelClasses] - Array of html classes to apply to the label
     * @param {string[]} [checkboxClasses] - Array of html classes to apply to the input
     * @param {Object} [data] - Key / value pairs to set on the input's dataset
     */
	createCheckboxElement(text, onClick, labelClasses, checkboxClasses, data) {

		const id = SortdCommon.newGuid();

		const el = document.createElement('label');
		el.classList.add(...labelClasses);
		el.setAttribute('for', id);

		const checkbox = document.createElement('input');
		checkbox.type = 'checkbox';
		checkbox.id = id;
		checkbox.classList.add(...checkboxClasses);

		if (data) {
			for (const key in data) {
				if (data.hasOwnProperty(key)) {
					checkbox.dataset[key] = data[key];
				}
			}
		}

		el.appendChild(checkbox);
		el.appendChild(document.createTextNode(text));

		if (onClick) {
			el.onclick = (event) => {
				onClick(event);
			}
		}

		return el;
	}

    /**
     * Create a new button element
     * @param {string} text - The text to appear on the button
     * @param {Function} onClick - The on click callback function
     * @param {string[]} [classes] - Array of html classes to apply to the button
     * @param {Object} [data] - Key / value pairs to set of the element's dataset
     */
	createGmailButtonElement(text, onClick, classes, data) {

		const el = document.createElement('span');
		el.tabIndex = 0;
		el.textContent = text;
		el.setAttribute('role', 'button');

		if (classes) {
			el.classList.add(...classes);
		}

		if (data) {
			for (const key in data) {
				if (data.hasOwnProperty(key)) {
					el.dataset[key] = data[key];
				}
			}
		}

		if (onClick) {
			el.onclick = (event) => {
				onClick(event);
			}
		}

		return el;
	}

    /**
     * Create a new button element
     * @param {string} text - The text to appear on the button
     * @param {Function} onClick - The on click callback function
     * @param {string[]} [classes] - Array of html classes to apply to the button
     * @param {Object} [data] - Key / value pairs to set of the element's dataset
     */
	createSortdButtonElement(text, onClick, classes, data) {

		const el = document.createElement('button');
		el.tabIndex = 0;
		el.textContent = text;
		el.classList.add('sortdButton');
		el.type = 'button';

		if (classes) {
			el.classList.add(...classes);
		}

		if (data) {
			for (const key in data) {
				if (data.hasOwnProperty(key)) {
					el.dataset[key] = data[key];
				}
			}
		}

		if (onClick) {
			el.onclick = (event) => {
				onClick(event);
			}
		}

		return el;
	}

    /**
     * @param {Object} options - The options object
     * @param {string} html - String to use as innerHTML of the drop down
     * @param {HTMLElement} [el] - HTML element to insert into the drop down element instead of the html string
	 * @param {Object} [options.positionToCoord] - x, y pair to position element to (if provided, positionToElement will be ignored)
     * @param {HTMLElement} [positionToElement] - HTML element to position the drop down to
     * @param {boolean} [centerToElement] - True to position the element centered to the positionToElement
     * @param {string} [className] - The css class name to add to the drop down element
     * @param {Function} [destroyedCallback] - Callback that is called when the drop down is destroyed.
     * @returns {Object} - Contains a destroy function that'll remove the element from the DOM
     */
	showDropDown(options) {

		const container = document.createElement('div');

		let destroyed = false;
		const destroy = () => {

			if (!destroyed) {
				container.remove();
				destroyed = true;
				if (options.destroyedCallback) {
					options.destroyedCallback();
				}
			}

			if (options.positionToElement) {
				options.positionToElement.classList.remove('sd_dropdown_active');
			}
		}

		const overlay = document.createElement('div');
		overlay.classList.add('sortd_default_Overlay');
		overlay.addEventListener('click', destroy);
		container.appendChild(overlay);

		const dropDownContainer = document.createElement('div');
		dropDownContainer.classList.add('sd_dropdown_container');

		const dropDownEl = document.createElement('div');
		dropDownEl.classList.add('sd_dropdown_body');
		if (options.el) {
			dropDownEl.appendChild(options.el);
		} else {
			dropDownEl.innerHTML = options.html;
		}

		if (options.className) {
			dropDownContainer.classList.add(options.className);
		}

		dropDownContainer.appendChild(dropDownEl);
		container.appendChild(dropDownContainer);
		document.body.appendChild(container);

		if (options.positionToElement) {
			options.positionToElement.classList.add('sd_dropdown_active');
		}

		setTimeout(() => {
			if (!destroyed) {
				if (options.positionToCoord) {
					this.positionToCoord(dropDownContainer, options.positionToCoord);
				} else if (options.positionToElement) {
					this.positionToElement(dropDownContainer, options.positionToElement, options.centerToElement);
				}
			}
		}, 200);

		return {
			destroy
		};
	}

    /**
     * Show the date picker and position it by the element
     * @param {HTMLElement} element - The element to position to
     * @param {Object} options - Options to set on the picker
     * @param {Function} next - The function to call with results
	 * @param {Function} [onclose] - The function to call when the picker closes 
	 * 	next will be called first if a date is selected, but not when the user cancels by clicking away
     */
	showDatePicker(element, options, next, onclose) {

		const selectDate = !!options.selectedTS;

		if (!options.inline) {
			const bodyRect = document.body.getBoundingClientRect();
			const elemRect = element.getBoundingClientRect();

			// If we're closer to the bottom than the picker's height, then align to bottom instead of top
			if (bodyRect.bottom - elemRect.bottom < 330) {
				element.classList.add('sd_algn_btm');
			}

			if (bodyRect.width - elemRect.left < 250) {
				element.classList.add('sd_algn_lft');
			}
		}

		const tomorrowMorning = new Date();
		tomorrowMorning.setDate(tomorrowMorning.getDate() + 1);
		tomorrowMorning.setHours(8, 0, 0, 0);

		const today = new Date();
		today.setHours(0, 0, 0, 0);

		let minDate = today;
		if (options.minDate || options.minDate === 0) {
			minDate = options.minDate;
		}

		options = Object.assign({
			selectedTS: tomorrowMorning.getTime(),
			dateOnly: false,
			closeOnSelect: options.dateOnly || false,
			min: minDate,
			max: options.maxDate || 200
		}, options);

		options.container = $(element);
		// Don't use arrow function below, as "this" would be incorrect inside the function
		options.onClose = function () {

			try {
				const selected = this.get('select');
				const highlighted = this.get('highlight');

				if (!this.get('cancelled')) {
					if (this.component && this.component.item && this.component.item.clear === true) {
						next(null, null);
					} else if (selected && highlighted && highlighted.pick) {
						const date = new Date(highlighted.pick);
						if (options.dateOnly) {
							date.setHours(0, 0, 0, 0);
						}
						next(null, date);
					} else {
						next(null, null);
					}
				}

				if (this.get('start')) {
					this.stop();
				}

				if (onclose) {
					onclose();
				}
			} catch (ex) {
				next(ex);
			}
		};

		options.container.pickadate(options);
		const picker = options.container.pickadate('picker');
		picker.stop(true).start().set('select', options.selectedTS).open();
		if (selectDate) {
			picker.set('highlight', options.selectedTS);
		}
	}

    /**
     * Adds the date time option selections to an existing element
	 * @param {Object} options - The options object
	 * @param {HTMLElement} options.element - The element to add the date options to
     * @param {number[]} options.dateIds - The date ids to be displayed in the drop down
	 * @param {boolean} options.dateOnly - True to show only date, false to show date & time
     * @param {Function} options.onDateSelected - Called when a date is selected
	 * @param {Function} options.close - Called when the dropdown needs to be closed
	 * @param {Number} options.selectionIndex - The selection index to start with
	 * @param {Date} [options.selectedDate] - The selected date
	 * @param {KefirStream} [options.dropDownActiveStream] - The stream to emit the drop down's active state to
	 * @param {boolean} [options.inlinePicker] - True if the custom selection needs to be an inline date picker
	 * @param {Number} [options.minDate] - Min date to allow for date picker
	 * @param {Number} [options.maxDate] - Max date to allow for date picker
	 * @return {Number} - The selection infox
     */
	addDateTimeOptionsToElement(options) {

		const el = options.element;
		let selectionIndex = options.selectionIndex || 0;

		let pickerEl = document.createElement('div');
		pickerEl.classList.add('sd_picker_container');
		el.appendChild(pickerEl);

		let ul = document.createElement('ul');
		let li;
		let customSelection = false;

		const dateIds = options.dateIds.slice();
		const customIndex = dateIds.indexOf(constants.DATE_ID.CUSTOM);
		if (customIndex !== -1) {
			customSelection = true;
			dateIds.splice(customIndex, 1);
		}

		dateIds.forEach((lblId) => {

			let date;
			li = document.createElement('li');
			li.dataset.sdSelectionIndex = selectionIndex;
			selectionIndex++;
			const span = document.createElement('span');
			let text;
			if (options.dateOnly) {
				date = SortdCommon.createDateFromTimeLabel(lblId);
				text = constants.DATE_LABELS[lblId];
				li.classList.add('sd_date_label', 'sd_id_' + lblId);
				if (date) {
					span.textContent = SortdCommon.formatShortDay(date);
				}
			} else {
				date = SortdCommon.createReminderFromTimeLabel(lblId);
				text = constants.TIME_LABELS[lblId];
				li.classList.add('sd_time_label', 'sd_id_' + lblId);
				if (date) {
					span.textContent = SortdCommon.formatShortDayTime(date);
				}
			}
			span.classList.add('sd_fl_rht');
			li.appendChild(document.createTextNode(text));
			li.appendChild(span);
			if (date) {
				span.dataset.sortdDate = date.toISOString();
				li.dataset.sortdDate = date.toISOString();
			} else {
				span.dataset.sortdLabel = lblId;
				li.dataset.sortdLabel = lblId;
			}
			ul.appendChild(li);
		});
		el.appendChild(ul);
		el.appendChild(this.createHorzSeparatorElement());

		if (customSelection) {

			const pickerOptions = { dateOnly: options.dateOnly, minDate: options.minDate, maxDate: options.maxDate };
			if (options.selectedDate) {
				pickerOptions.selectedTS = options.selectedDate.getTime();
			}

			const pickerDateCallback = (err, date) => {

				options.close();

				if (err) {
					this.handleError(err);
				} else if (date) {
					options.onDateSelected(date);
				} else {
					options.onDateSelected();
				}

				if (options.dropDownActiveStream) {
					options.dropDownActiveStream.emit(false);
				}
			};

			const pickerDestroyCallback = () => {

				options.close();

				if (options.dropDownActiveStream) {
					options.dropDownActiveStream.emit(false);
				}
			};

			if (options.inlinePicker) {
				const pickerContainer = document.createElement('div');
				pickerContainer.classList.add(options.dateOnly ? 'sd_date_picker_inline' : 'sd_datetime_picker_inline');
				pickerEl = document.createElement('div');
				pickerContainer.appendChild(pickerEl);
				el.appendChild(pickerContainer);
				pickerOptions.inline = true;

				setTimeout(() => {
					this.showDatePicker(pickerEl, pickerOptions, pickerDateCallback.bind(this), pickerDestroyCallback.bind(this));
				}, 300);
			} else {
				ul = document.createElement('ul');
				li = document.createElement('li');
				li.textContent = 'Pick a date';
				li.classList.add('sd_date_label', 'sd_id_custom');
				li.dataset.sdSelectionIndex = selectionIndex;
				selectionIndex++;
				ul.appendChild(li);
				li.addEventListener('click', () => {

					setTimeout(() => {

						if (options.dropDownActiveStream) {
							options.dropDownActiveStream.emit(true);
						}

						this.showDatePicker(pickerEl, pickerOptions, pickerDateCallback.bind(this));
					}, 10);
				});
				el.appendChild(ul);
			}
		}

		return selectionIndex;
	}

	/**
     * Adds the Remind me header, as well as the No date option
	 * @param {Object} options - The options object
	 * @param {HTMLElement} options.element - The element to add the date options to
	 * @param {Number} options.selectionIndex - The selection index to start with
	 * @return {Number} - The selection infox
     */
	addNextStepHeadingAndDateToElement(options) {
		
		const el = options.element;
		let selectionIndex = options.selectionIndex || 0;

		const heading = document.createElement('div');
		heading.classList.add('sd_selector_header');
		heading.innerText = 'Remind me';
		el.appendChild(heading);
		el.appendChild(this.createHorzSeparatorElement());

		let ul = document.createElement('ul');
		let li = document.createElement('li');
		li.dataset.sdSelectionIndex = selectionIndex;
		selectionIndex++;
		li.classList.add('sd_date_label', 'sd_id_nodate');
		li.dataset.sortdDate = 'nodate';		
		li.appendChild(document.createTextNode('No date'));
		ul.appendChild(li);
		el.appendChild(ul);

		return selectionIndex;
	}

    /**
     * Create a date/time or just date selector dropdown element
	 * @param {Object} options - The options object
     * @param {number[]} options.dateIds - The date ids to be displayed in the drop down
	 * @param {boolean} options.dateOnly - True to show only date, false to show date & time
	 * @param {String} options.heading - The heading to display
     * @param {Function} options.onDateSelected - Called when a date is selected
	 * @param {Object} [options.positionToCoord] - x, y pair to position element to (if provided, positionToElement will be ignored)
     * @param {HTMLElement} [options.positionToElement] - HTML element to position the drop down to
	 * @param {boolean} [options.centerToElement] - True when the button is the part of the active thread
	 * @param {Date} [options.selectedDate] - The selected date
	 * @param {KefirStream} [options.dropDownActiveStream] - The stream to emit the drop down's active state to
	 * @param {boolean} [options.inlinePicker] - True if the custom selection needs to be an inline date picker
	 * @param {Number} [options.minDate] - Min date to allow for date picker
	 * @param {Number} [options.maxDate] - Max date to allow for date picker
     */
	showDateTimeDropDown(options) {

		let result;
		let defaultSelected = 0;
		let checkboxContainer;

		const closeDropDown = () => {

			if (result && result.destroy) {
				result.destroy();
			}
		}

		const el = document.createElement('div');
		const heading = document.createElement('div');
		heading.classList.add('sd_selector_header');
		heading.innerText = options.heading || '';
		if (options.selectedDate) {
			const clear = document.createElement('div');
			clear.classList.add('sd_clear_lnk', 'sd_fl_rht');
			clear.innerText = 'Clear';
			clear.dataset.sortdDate = 'clear';
			heading.appendChild(clear);
		}
		el.appendChild(heading);

		this.addDateTimeOptionsToElement({
			element: el,
			dateIds: options.dateIds,
			dateOnly: options.dateOnly,
			onDateSelected: (dt) => {
				const checkbox = checkboxContainer && checkboxContainer.querySelector('input');
				const setDueDateToPlannedDate = checkbox && checkbox.checked || false;
				options.onDateSelected(dt, setDueDateToPlannedDate);
			},
			close: () => closeDropDown(),
			selectedDate: options.selectedDate,
			dropDownActiveStream: options.dropDownActiveStream,
			inlinePicker: options.inlinePicker,
			minDate: options.minDate,
			maxDate: options.maxDate
		});

		if (options.showSetDueDate) {
			el.appendChild(this.createHorzSeparatorElement());
			checkboxContainer = this.createCheckboxElement('Also update the deadline date', null, ['sd_chk_label', 'sd_vsmall_text'], []);
			checkboxContainer.querySelector('input').checked = true;
			el.appendChild(checkboxContainer);
		}

		el.addEventListener('click', (event) => {

			event.stopPropagation();

			const clickedEl = event.target;
			const sortdDate = clickedEl.dataset.sortdDate;
			const sortdLabel = clickedEl.dataset.sortdLabel;

			const checkbox = checkboxContainer && checkboxContainer.querySelector('input');
			const setDueDateToPlannedDate = checkbox && checkbox.checked || false;

			if (sortdDate === 'cancel') {
				closeDropDown();
			} else if (sortdDate === 'clear') {
				options.onDateSelected(null, setDueDateToPlannedDate);
				closeDropDown();
			} else if (sortdDate) {
				options.onDateSelected(new Date(sortdDate), setDueDateToPlannedDate);
				closeDropDown();
			} else if (sortdLabel) {
				options.onDateSelected(parseInt(sortdLabel, 10), setDueDateToPlannedDate);
				closeDropDown();
			}
		});

		if (options.dropDownActiveStream) {
			options.dropDownActiveStream.emit(true);
		}

		result = this.showDropDown({
			el,
			positionToCoord: options.positionToCoord,
			positionToElement: options.positionToElement,
			centerToElement: options.centerToElement,
			className: 'sd_with_picker',
			destroyedCallback: () => {

				if (options.dropDownActiveStream) {
					options.dropDownActiveStream.emit(false);
				}

				this.unbindArrowNavigation(closeDropDown, closeDropDown);
			}
		});

		setTimeout(() => this.bindArrowNavigation(el, closeDropDown, closeDropDown, defaultSelected), 500);
	}

    /**
     * Create a date/time or just date selector dropdown element
	 * @param {Object} options - The options object
     * @param {number[]} options.dateIds - The date ids to be displayed in the drop down
	 * @param {boolean} options.dateOnly - True to show only date, false to show date & time
     * @param {Function} options.onDateSelected - Called when a date is selected
     * @param {Function} [options.onClose] - Called when a component is closed
     * @param {HTMLElement} [options.positionToElement] - HTML element to position the drop down to
	 * @param {Date} [options.selectedDate] - The selected date
	 * @param {boolean} [options.inlinePicker] - True if the custom selection needs to be an inline date picker
	 * @param {Number} [options.minDate] - Min date to allow for date picker
	 * @param {Number} [options.maxDate] - Max date to allow for date picker
     */
	showActionOverrideDateTimeDropDown(options) {

		let result;
		let defaultSelected = 0;
		let selectedClose = false;

		const closeDropDown = () => {

			if (result && result.destroy) {
				result.destroy();
			}
		}

		const el = document.createElement('div');
		const heading = document.createElement('div');
		heading.classList.add('sd_selector_header');
		heading.innerText = 'When do you want to get back to this?';
		el.appendChild(heading);

		let selectionIndex = this.addDateTimeOptionsToElement({
			element: el,
			dateIds: options.dateIds,
			dateOnly: options.dateOnly,
			onDateSelected: options.onDateSelected,
			close: () => closeDropDown(),
			selectedDate: options.selectedDate,
			inlinePicker: options.inlinePicker,
			minDate: options.minDate,
			maxDate: options.maxDate
		});

		el.appendChild(this.createHorzSeparatorElement());

		const ul = document.createElement('ul');
		let li = document.createElement('li');
		li.classList.add('sd_date_label');
		li.dataset.sdSelectionIndex = selectionIndex;
		defaultSelected = selectionIndex;
		li.dataset.sortdDate = 'cancel';
		selectionIndex++;
		const span = document.createElement('span');
		span.textContent = 'ENTER';
		span.classList.add('sd_fl_rht', 'sd_active_only');
		li.appendChild(document.createTextNode('Skip'));
		li.appendChild(span);
		ul.appendChild(li);
		el.appendChild(ul);

		el.addEventListener('click', (event) => {

			event.stopPropagation();

			const clickedEl = event.target;
			const sortdDate = clickedEl.dataset.sortdDate;
			const sortdLabel = clickedEl.dataset.sortdLabel;

			if (sortdDate === 'cancel') {
				selectedClose = true;
				closeDropDown();
			} else if (sortdDate === 'clear') {
				options.onDateSelected();
				closeDropDown();
			} else if (sortdDate) {
				options.onDateSelected(new Date(sortdDate));
				closeDropDown();
			} else if (sortdLabel) {
				options.onDateSelected(parseInt(sortdLabel, 10));
				closeDropDown();
			}
		});

		result = this.showDropDown({
			el,
			positionToElement: options.positionToElement,
			className: 'sd_with_picker',
			destroyedCallback: () => {

				if (options.onClose) {
					options.onClose(selectedClose);
				}

				this.unbindArrowNavigation(closeDropDown, closeDropDown);
			}
		});

		setTimeout(() => this.bindArrowNavigation(el, closeDropDown, closeDropDown, defaultSelected), 500);
	}

    /**
     * Create a date/time or just date selector dropdown element
	 * @param {Object} options - The options object
     * @param {number[]} options.dateIds - The date ids to be displayed in the drop down
	 * @param {boolean} options.dateOnly - True to show only date, false to show date & time
	 * @param {String} options.heading - The heading to display
     * @param {Function} options.onDateSelected - Called when a date is selected
	 * @param {Boolean} options.originNextStep - True if called for setting the next step
	 * @param {Object} [options.positionToCoord] - x, y pair to position element to (if provided, positionToElement will be ignored)
     * @param {HTMLElement} [options.positionToElement] - HTML element to position the drop down to
	 * @param {boolean} [options.centerToElement] - True when the button is the part of the active thread
	 * @param {Date} [options.selectedDate] - The selected date
	 * @param {String} [options.nextStep] - The current next step value
	 * @param {KefirStream} [options.dropDownActiveStream] - The stream to emit the drop down's active state to
	 * @param {boolean} [options.inlinePicker] - True if the custom selection needs to be an inline date picker
     */
	showReminderTimeDropDown(options) {

		let result;
		let defaultSelected = 0;
		let checkboxContainer;
		let hasSaved = false;

		const closeDropDown = () => {

			if (result && result.destroy) {
				result.destroy();
			}
		}

		const selectDate = (date, fromCancel) => {

			const checkbox = checkboxContainer && checkboxContainer.querySelector('input');
			const newNextStep = nextStep.value || '';
			let addToDatePlan = checkbox && checkbox.checked || false;
			let saveNextStep = true;

			if (fromCancel) {
				const currentNextStep = options.nextStep || '';

				addToDatePlan = false;
				saveNextStep = (currentNextStep !== newNextStep);
			}

			if (saveNextStep) {
				options.onDateSelected(date, addToDatePlan, newNextStep);
			}
		}

		const el = document.createElement('div');
		const heading = document.createElement('div');
		heading.classList.add('sd_selector_header');
		heading.innerText = options.heading || '';
		if ((!options.originNextStep && options.selectedDate) || (options.originNextStep && options.nextStep)) {
			const clear = document.createElement('div');
			clear.classList.add('sd_clear_lnk', 'sd_fl_rht');
			clear.innerText = 'Clear';
			clear.dataset.sortdDate = 'clear';
			heading.appendChild(clear);
		}
		el.appendChild(heading);

		const nextStep = document.createElement('textarea');
		nextStep.placeholder = options.originNextStep ? 'Add a note, status or next step' : 'Add an optional note or next step';
		nextStep.value = options.nextStep || '';
		nextStep.classList.add('sd_text_fw');
		nextStep.addEventListener('focus', () => nextStep.select());
		nextStep.addEventListener('click', (evt) => evt.stopPropagation());
		el.appendChild(nextStep);

		if (options.originNextStep) {
			this.addNextStepHeadingAndDateToElement({
				element: el
			});
		}

		this.addDateTimeOptionsToElement({
			element: el,
			dateIds: options.dateIds,
			dateOnly: options.dateOnly,
			onDateSelected: selectDate,
			selectionIndex: 1,
			close: () => closeDropDown(),
			selectedDate: options.selectedDate,
			dropDownActiveStream: options.dropDownActiveStream,
			inlinePicker: options.inlinePicker
		});

		el.appendChild(this.createHorzSeparatorElement());
		checkboxContainer = this.createCheckboxElement('Add this item to My Day plan', null, ['sd_chk_label', 'sd_vsmall_text'], []);
		if (options.originNextStep) {
			checkboxContainer.querySelector('input').checked = true;
		}
		el.appendChild(checkboxContainer);

		el.addEventListener('click', (event) => {

			hasSaved = true;
			event.stopPropagation();

			const clickedEl = event.target;
			const sortdDate = clickedEl.dataset.sortdDate;
			const sortdLabel = clickedEl.dataset.sortdLabel;

			if (sortdDate === 'cancel') {
				closeDropDown();
			} else if (sortdDate === 'nodate') {
				selectDate();
				closeDropDown();
			} else if (sortdDate === 'clear') {
				nextStep.value = '';
				selectDate();
				closeDropDown();
			} else if (sortdDate) {
				selectDate(new Date(sortdDate));
				closeDropDown();
			} else if (sortdLabel) {
				selectDate(parseInt(sortdLabel, 10));
				closeDropDown();
			}
		});

		if (options.dropDownActiveStream) {
			options.dropDownActiveStream.emit(true);
		}

		result = this.showDropDown({
			el,
			positionToCoord: options.positionToCoord,
			positionToElement: options.positionToElement,
			centerToElement: options.centerToElement,
			className: 'sd_with_picker',
			destroyedCallback: () => {

				if (!hasSaved && options.originNextStep) {
					selectDate(options.selectedDate, true);
				}

				if (options.dropDownActiveStream) {
					options.dropDownActiveStream.emit(false);
				}

				this.unbindArrowNavigation(closeDropDown, closeDropDown);
			}
		});

		setTimeout(() => {
			this.bindArrowNavigation(el, closeDropDown, closeDropDown, defaultSelected);
			nextStep.focus();
			nextStep.select();
		}, 500);
	}

	ensureBottomPositionFits(bodyRect, elemRect, bottom) {

		const totalHeight = bottom + elemRect.height;
		if (totalHeight > bodyRect.height) {
			bottom = bottom - (totalHeight - bodyRect.height + 20);
		}

		return bottom;
	}

	/**
     * Position an element to the provided coordinates
     * @param {HTMLElement} elementToPosition - The element to position
     * @param {Object} coordinates - x, y pair to position to
     */
	positionToCoord(elementToPosition, coordinates) {

		if (elementToPosition && elementToPosition.style) {

			const bodyRect = document.body.getBoundingClientRect();
			const elemRect = elementToPosition.getBoundingClientRect();

			const totalHeight = coordinates.y + elemRect.height;
			if (totalHeight >= bodyRect.height) {
				const bottom = this.ensureBottomPositionFits(bodyRect, elemRect, (bodyRect.height - coordinates.y));
				elementToPosition.style.bottom = `${bottom}px`;
			} else {
				elementToPosition.style.top = `${coordinates.y}px`;
			}

			const totalWidth = coordinates.x + elemRect.width + 20;
			if (totalWidth >= bodyRect.width) {
				elementToPosition.style.right = '20px';
			} else {
				elementToPosition.style.left = `${coordinates.x}px`;
			}

			elementToPosition.classList.add('sd_coord_container');
		}
	}

    /**
     * Position an element relative to another element
     * @param {HTMLElement} elementToPosition - The element to position
     * @param {HTMLElement} relativeTo - The element to which is should be positioned to
     * @param {boolean} center - True to horizonally center to the element, false to align left sides
     */
	positionToElement(elementToPosition, relativeTo, center = false) {

		if (elementToPosition && elementToPosition.style) {

			const bodyRect = document.body.getBoundingClientRect();
			const elemRect = elementToPosition.getBoundingClientRect();
			const relativeToRect = relativeTo.getBoundingClientRect();

			let totalHeight = relativeToRect.top - bodyRect.top + relativeToRect.height + elemRect.height;
			if (totalHeight >= bodyRect.height) {
				const bottom = this.ensureBottomPositionFits(bodyRect, elemRect, Math.round(bodyRect.bottom - relativeToRect.bottom + relativeToRect.height) + 5);
				elementToPosition.style.bottom = `${bottom}px`;
				elementToPosition.classList.add('sd_btm_aligned');
			} else {
				elementToPosition.style.top = `${Math.round(relativeToRect.top - bodyRect.top + relativeToRect.height) + 10}px`;
				elementToPosition.classList.add('sd_top_aligned');
			}

			let left = Math.round(relativeToRect.left - bodyRect.left);
			if (center) {
				left += ((relativeToRect.width - elemRect.width) / 2);
			}

			const totalWidth = left + elemRect.width + 20;
			if (totalWidth >= bodyRect.width) {
				elementToPosition.style.right = '20px';
			} else {
				elementToPosition.style.left = `${left}px`;
			}
		}
	}

	/**
	 * Finds the first elements that match one of the selectors
	 * @param {string[]} selectors - Array of selectors
	 * @param {HTMLElement} [el] - The element which children must be searched. Defaults to document
	 */
	getElementsFromSelector(selectors, el) {

		if (selectors) {
			el = el || document;
			for (let i = 0; i < selectors.length; i++) {
				const result = el.querySelectorAll(selectors[i]);
				if (result && result.length > 0) {
					return result;
				}
			}
		}
	}

	/**
	 * Wait for a function to return one or more elements
	 * @param {string[]} selectors - Array of selectors to wait for
	 * @param {string[]} [el] - The element to do the querySelector on, if not provided will use document
	 * @param {number} [timeout] - The number of milliseconds to wait before returning an error
	 * @param {number} [interval] - The number of milliseconds to wait between each retry
	 */
	waitForElements(selectors, el, timeout, interval) {

		return SortdCommon.waitForResult(() => {
			return this.getElementsFromSelector(selectors, el);
		}, timeout, interval);
	}

	/**
     * Capture the mouse up event, and gives the option to call back to let the event propagate
     * @param {HTMLElement} el - The element the capture the mouseup event on
     * @param {Function} callback - The function to call when the event is captured,
	 * 		the first param is the event that was captured
	 * 		the second param is a function that can be called to propagate the event
     * @param {Function} [validationFn] - A validation function which is called with the event,
	 * 		if it returns false the event will not be captured but propagated
     */
	captureMouseUpEvent(el, callback, validationFn) {

		const events = [];

		el.addEventListener('mousedown', (evt) => events.push(evt));
		el.addEventListener('click', (evt) => events.push(evt));
		el.addEventListener('mouseup', (evt) => {

			if (evt.dontCaptureSortd || (validationFn && !validationFn(evt))) {
				return;
			}

			evt.stopPropagation();
			events.push(evt);
			const currentEvents = events.splice(0, events.length);

			callback(evt, () => {

				for (let i = 0; i < currentEvents.length; i++) {
					currentEvents[i].dontCaptureSortd = true;
					currentEvents[i].target.dispatchEvent(currentEvents[i]);
				}
			});
		});
	}

	/**
     * Capture the mouse click, and gives the option to call back to let the event propagate
     * @param {HTMLElement} el - The element the capture the click event on
     * @param {Function} callback - The function to call when the event is captured,
	 * 		the first param is the event that was captured
	 * 		the second param is a function that can be called to propagate the event
     * @param {Function} validationFn - A validation function which is called with the event,
	 * 		if it returns false the event will not be captured but propagated
     */
	captureMouseClickEvent(el, callback, validationFn) {

		el.addEventListener('click', (evt) => {

			if (evt.dontCaptureSortd || (validationFn && !validationFn(evt))) {
				return;
			}

			evt.stopPropagation();

			callback(evt, () => {

				evt.dontCaptureSortd = true;
				evt.target.dispatchEvent(evt);
			});
		});
	}
}