/**
 * Sortd
 * Copyright (c) 2019, Sortd Inc
 * https://www.sortd.com
 *
 * @fileoverview Display messages to the user
 * @name notifications
 */

const NOTIFICATION_TEXT_MARGIN = 24;

class GmailApiNotifications {

    constructor(options) {

        this.__FILE_NAME = 'notifications.js';

        this.config = options.config;
        this.errorHandler = options.errorHandler;
        this.dom = options.dom;

        this.notificationItems = null;
		this.currentNotification = null;
		this.isEnabled = true;

        this.BAR_DOM_MAP = SortdGmailApiDomMap.GMAIL_NOTIFICATION_BAR[0].CLASSES;
	}
	
	get enabled() {
		return this.isEnabled;
	}

    /**
     * Initialize the notification bar by creating the required observables
     */
    init() {

        this.setNotificationItems();

        this.gmailNotificationBarChanges = this.notificationItems.flatMapLatest((notifications) => {
            return this.dom.observeMutations(notifications.gmailNotification, { childList: true });
        });

        this.gmailNotificationBarAddedNodes = this.gmailNotificationBarChanges.filter((mutations) => {
            return mutations.some((m) => m.addedNodes.length > 0);
        });

        this.gmailNotificationBarRemovedNodes = this.gmailNotificationBarChanges.filter((mutations) => {
            return !mutations.some((m) => m.addedNodes.length > 0);
        });

        const obsAdded = Kefir.combine([this.notificationItems, this.gmailNotificationBarAddedNodes]).observe((value) => {

            if (this.currentNotification) {
                const notifications = value[0];

                if (this.currentNotification.keepCurrentNotification) {
                    this.updateCurrentNotification(notifications.gmailNotification, notifications.sortdNotification);
                } else {                    
                    this.setSortdNotificationBarInactive(
                        notifications.notificationContainer,
                        notifications.gmailNotification,
                        notifications.sortdNotification);
                    this.currentNotification = null;
                }
            }
        });

        const obsRemoved = Kefir.combine([this.notificationItems, this.gmailNotificationBarRemovedNodes]).observe(() => {

            if (this.currentNotification) {
                this.destroyNotification(
                    this.currentNotification.notificationId,
                    this.currentNotification.className,
                    this.currentNotification.canDestroyCallback,
                    this.currentNotification.destroyedCallback);
            }
		});
		
		// Wait for notification items to be set up, then check if we're going to have any clashes
		// If there are clashes remove the notification listeners again
		// and remove the sortd notification element
		this.notificationItems
			.take(1)
			.onValue((notificationItems) => {

				const selectors = SortdGmailApiDomMap.GMAIL_NOTIFICATION_BAR[0].DONT_SHOW_NOTIFICATION_BAR_SELECTORS;

				SortdCommon
					.waitForResult(() => {

					    if (selectors) {
                            for (let i = 0; i < selectors.length; i++) {
                                if (document.querySelector(selectors[i])) {
                                    return true;
                                }
                            }
                        }
					})
					.onValue(() => {

						obsAdded.unsubscribe();
						obsRemoved.unsubscribe();

						notificationItems.sortdNotification && notificationItems.sortdNotification.remove();

						this.isEnabled = false;
					});
			});
    }

    /**
     * Show a new notification
     * @param {Object} options - The options object
     * @param {string} text - Text to show
     * @param {string} [html] - String to use as innerHTML of the notification bar instead of using the given text
     * @param {HTMLElement} [el] HTML element to insert into the notification bar instead of using the given text
     * @param {string} [className] - The css class name to add to the notification bar
     * @param {number} [time] - Number of milliseconds to display message. Defaults to 15000.
     *      If it is empty and keepCurrentNotification is true the message will be displayed as long as the existing Gmail notification.
     * @param {boolean} [keepCurrentNotification] - True to keep the current message, and display this notification next to it. Defaults to false.
     * @param {Object []} [buttons]
     * @param {Function} [buttons[].onClick] - Callback when button clicked.  Event is passed in: function(event)
     * @param {string} [buttons[].title] - Button text
     * @param {Function} [canDestroyCallback] - Callback before a notification is destroyed.
     * @param {Function} [destroyedCallback] - Callback that is called when the notification is destroyed
     *      If it returns false the notification will remain active until manually destroyed, or until the next notification is added
     * @private
     */
    showNotification(options) {

		if (!this.isEnabled) {
			return;
		}

        this.currentNotification = options = Object.assign({
            notificationId: SortdCommon.newGuid(),
            keepCurrentNotification: false
        }, options);

        if (!options.time && !options.keepCurrentNotification) {
            options.time = 15000;
        }

        this.notificationItems
            .take(1)
            .onValue(({ notificationContainer, gmailNotification, sortdNotification }) => {

                const textEl = sortdNotification.querySelector('.sd_notification_bar_txt');
                textEl.innerHTML = '';

                if (options.html) {
                    textEl.innerHTML = options.html;
                } else if (options.el) {
                    textEl.appendChild(options.el);
                } else if (options.text) {
                    textEl.textContent = options.text;
                } else {
                    const selector = this.dom.selectorFromClassArray(this.BAR_DOM_MAP.MSG_TEXT);
                    const gmailTextEl = gmailNotification.querySelector(selector);
                    textEl.textContent = gmailTextEl.textContent;
                }

                // Replace the close button with a new one, so that all old event handlers are removed
                // Then add a new click event handler
                const oldSortdCloseBtn = sortdNotification.querySelector('.sd_notification_bar_close');
                const sortdCloseBtn = oldSortdCloseBtn.cloneNode(true);
                oldSortdCloseBtn.parentNode.replaceChild(sortdCloseBtn, oldSortdCloseBtn);
                sortdCloseBtn.addEventListener('click', () => {
                    this.destroyNotification(options.notificationId, options.className, null, options.destroyedCallback);
                });

                const buttonEl = sortdNotification.querySelector('.sd_notification_bar_btns');
                buttonEl.innerHTML = '';

                if (options.buttons) {
                    options.buttons.forEach((button) => {

                        const el = this.dom.createGmailButtonElement(
                            button.title,
                            (event) => button.onClick(event),
                            this.BAR_DOM_MAP.MSG_BUTTONS);

                        buttonEl.appendChild(el);
                    });
                }

                notificationContainer.style.visibility = 'visible';
                notificationContainer.style.top = '';
                notificationContainer.classList.add(...this.BAR_DOM_MAP.CONTAINER_ACTIVE);
                
                if (options.keepCurrentNotification) {
                    gmailNotification.style.boxShadow = 'none';
                    gmailNotification.style.zIndex = 2;
                    gmailNotification.style.paddingRight = '0';
                    sortdNotification.style.zIndex = 1;

                    const selector = this.dom.selectorFromClassArray(this.BAR_DOM_MAP.BUTTON);
                    const closeBtn = gmailNotification.querySelector(selector);
                    if (closeBtn) {
                        closeBtn.style.display = 'none';
                    }

                    const bodyRect = gmailNotification.getBoundingClientRect();
                    const resizeEl = textEl.querySelector('.sd_resize_curr') || textEl;
                    resizeEl.style.width = `${bodyRect.width - NOTIFICATION_TEXT_MARGIN}px`;
                } else {
                    gmailNotification.style.display = 'none';
                }

                if (options.className) {
                    sortdNotification.classList.add(options.className);
                }
                sortdNotification.style.display = '';
                sortdNotification.dataset.sortdId = options.notificationId;
            });

        if (options.time) {
            this.closeCurrentNotificationLater(options.time, options.className, options.canDestroyCallback, options.destroyedCallback);
        }

        return {
            destroy: () => {
                this.destroyNotification(options.notificationId, options.className, options.canDestroyCallback, options.destroyedCallback);
            }
        }
    }

    /**
     * Create the Sortd notification bar if it hasn't been created
     * Set notificationItems to an observable of the notification container as well as Sortd and Gmail notification bar elements
     * @private
     */
    setNotificationItems() {

        if (this.notificationItems) {
            return;
        }

        this.notificationItems = SortdCommon
            .waitForResult(() => {
                const selector = this.dom.selectorFromClassArray(this.BAR_DOM_MAP.CONTAINER);
                return document.body.querySelector(selector);
            })
            .map((notificationContainer) => {

                const selector = this.dom.selectorFromClassArray(this.BAR_DOM_MAP.BAR);
                const gmailNotification = notificationContainer.querySelector(`${selector}:not(.sd_notification_bar)`);
                gmailNotification.classList.add('gmailNotificationBar');

                let sortdNotification = notificationContainer.querySelector(`${selector}.sd_notification_bar`);

                if (!sortdNotification) {

                    sortdNotification = document.createElement('div');
					sortdNotification.classList.add(...this.BAR_DOM_MAP.BAR, 'sd_notification_bar');
                    sortdNotification.style.display = 'none';
                    
                    const messagePart = document.createElement('span');
                    messagePart.classList.add(...this.BAR_DOM_MAP.MSG_CONTAINER);

                    const messageText = document.createElement('span');
                    messageText.classList.add(...this.BAR_DOM_MAP.MSG_TEXT, 'sd_notification_bar_txt');
                    messageText.style.display = 'inline-block';
                    messagePart.appendChild(messageText);
                    
                    const messageButtons = document.createElement('span');
                    messageButtons.classList.add(...this.BAR_DOM_MAP.MSG_BUTTON_CONTAINER, 'sd_notification_bar_btns');
                    messagePart.appendChild(messageButtons);

                    sortdNotification.appendChild(messagePart);

                    const closeContainer = document.createElement('div');
                    closeContainer.classList.add(...this.BAR_DOM_MAP.CLOSE_CONTAINER, 'sd_notification_bar_close');
                    closeContainer.innerHTML = `<div class="${this.BAR_DOM_MAP.BUTTON}"></div>`;
                    closeContainer.tabIndex = 0;
                    closeContainer.setAttribute('role', 'button');
                    sortdNotification.appendChild(closeContainer);

                    gmailNotification.parentNode.appendChild(sortdNotification);
                }

                return {
                    notificationContainer,
                    gmailNotification,
                    sortdNotification
                };
            })
            .toProperty();
    }

    /**
     * Updates the current notification's width and text if required
     * @param {HTMLElement} gmailNotification - The Gmail notification bar element
     * @param {HTMLElement} sortdNotification - The Sortd notification bar element
     */
    updateCurrentNotification(gmailNotification, sortdNotification) {

		let activeElId;
		if (document.activeElement && document.activeElement.id) {
			activeElId = document.activeElement.id;
		}

        const textEl = sortdNotification.querySelector('.sd_notification_bar_txt');
        textEl.innerHTML = '';

        if (this.currentNotification.html) {
            textEl.innerHTML = this.currentNotification.html;
        } else if (this.currentNotification.el) {
            textEl.appendChild(this.currentNotification.el);
        } else if (this.currentNotification.text) {
            textEl.textContent = this.currentNotification.text;
        } else {
            const selector = this.dom.selectorFromClassArray(this.BAR_DOM_MAP.MSG_TEXT);
            const gmailTextEl = gmailNotification.querySelector(selector);
            if (gmailTextEl) {
                textEl.textContent = gmailTextEl.textContent;
            }
		}
		
		// We could have replaced the element above, in which case the active element
		// would have changed.  Put the focus back into the previously active element
		if (!document.activeElement || document.activeElement.id !== activeElId) {
			const activeEl = document.getElementById(activeElId);
			if (activeEl) {
				activeEl.focus();
			}
		}
        
        if (this.currentNotification.keepCurrentNotification) {
            gmailNotification.style.boxShadow = 'none';
            gmailNotification.style.zIndex = 2;
            gmailNotification.style.paddingRight = '0';

            const bodyRect = gmailNotification.getBoundingClientRect();
            const resizeEl = textEl.querySelector('.sd_resize_curr') || textEl;
            resizeEl.style.width = `${bodyRect.width - NOTIFICATION_TEXT_MARGIN}px`;

            const selector = this.dom.selectorFromClassArray(this.BAR_DOM_MAP.BUTTON);
            const closeBtn = gmailNotification.querySelector(selector);
            if (closeBtn) {
                closeBtn.style.display = 'none';
            }
        }
    }

    /**
     * Destroy the notification after a time period has elapsed
     * @param {number} time - The number of milliseconds after which to destroy the notification
     * @param {string} [className] - The class name of the notification
     * @param {Function} [canDestroyCallback] - The callback to call before destroying the notification
     * @param {Function} [destroyedCallback] - Callback that is called when the notification is destroyed
     */
    closeCurrentNotificationLater(time, className, canDestroyCallback, destroyedCallback) {

        Kefir.later(time, null)
            .take(1)
            .onValue(() => {

                if (this.currentNotification && this.currentNotification.notificationId) {
                    this.destroyNotification(this.currentNotification.notificationId, className, canDestroyCallback, destroyedCallback);
                }
            });
    }

    /**
     * Remove the notification if canDestroyCallback is not defined or it returns true
     * @param {string} notificationId - The id of the notification
     * @param {string} [className] - The custom class name that was added to the notification
     * @param {Function} [canDestroyCallback] - The callback function to determine if the notification can be removed
     * @param {Function} [destroyedCallback] - Callback that is called when the notification is destroyed
     * @private
     */
    destroyNotification(notificationId, className, canDestroyCallback, destroyedCallback) {

        this.notificationItems
            .take(1)
            .onValue(({ notificationContainer, gmailNotification, sortdNotification }) => {

                let canDestroy = true;

                if (canDestroyCallback) {
                    canDestroy = canDestroyCallback();
                }

                if (canDestroy) {
                    if (sortdNotification.dataset.sortdId === notificationId) {

                        if (className) {
                            sortdNotification.classList.remove(className);
                        }

                    }

                    this.resetNotificationBars(notificationContainer, gmailNotification, sortdNotification);
                    if (this.currentNotification && this.currentNotification.notificationId === notificationId) {
                        this.currentNotification = null;
                    }

                    if (destroyedCallback) {
                        destroyedCallback();
                    }
                } else if (sortdNotification.dataset.sortdId === notificationId) {
                    notificationContainer.classList.add(...this.BAR_DOM_MAP.CONTAINER_ACTIVE);
                    this.closeCurrentNotificationLater(1000, className, canDestroyCallback, destroyedCallback);
                }
            });
    }

    /**
     * Reset the message container and bars to their original states
     * @param {HTMLElement} notificationContainer - The notification container element
     * @param {HTMLElement} gmailNotification - The Gmail notification bar element
     * @param {HTMLElement} sortdNotification - The Sortd notification bar element
     * @private
     */
    resetNotificationBars(notificationContainer, gmailNotification, sortdNotification) {

        notificationContainer.style.top = '-10000px';
        notificationContainer.style.position = 'relative';
        notificationContainer.classList.remove(...this.BAR_DOM_MAP.CONTAINER_ACTIVE);

        this.setSortdNotificationBarInactive(notificationContainer, gmailNotification, sortdNotification);
    }

    /**
     * Set the Sortd message bar to inactive, and add close button back to Gmail bar
     * @param {HTMLElement} notificationContainer - The notification container element
     * @param {HTMLElement} gmailNotification - The Gmail notification bar element
     * @param {HTMLElement} sortdNotification - The Sortd notification bar element
     * @private
     */
    setSortdNotificationBarInactive(notificationContainer, gmailNotification, sortdNotification) {

        sortdNotification.style.display = 'none';
        sortdNotification.style.zIndex = '';
        sortdNotification.dataset.sortdId = '';

        gmailNotification.style.zIndex = '';
        gmailNotification.style.boxShadow = '';
        gmailNotification.style.paddingRight = '';
        gmailNotification.style.display = '';

        const selector = this.dom.selectorFromClassArray(this.BAR_DOM_MAP.BUTTON);
        const closeBtn = gmailNotification.querySelector(selector);
        if (closeBtn) {
            closeBtn.style.display = '';
        }
    }
}