/**
 * Sortd
 * Copyright (c) 2019, Sortd Inc
 * https://www.sortd.com
 *
 * @fileoverview Common Sortd helper functions
 * @name sortd-common
 */

class SortdCommon {}

SortdCommon.__FILE_NAME = 'sortd-common.js';

SortdCommon.daysShort = ['Sun', 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat'];
SortdCommon.monthsShort = [ 'Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec' ];
SortdCommon.htmlEscaper = /[&<>"'\/]/g;
SortdCommon.htmlEscapes = {
	'&': '&amp;',
	'<': '&lt;',
	'>': '&gt;',
	'"': '&quot;',
	"'": '&#x27;',
	'/': '&#x2F;'
};

/**
 * Wait for a function to return a value
 * @param {Function} fn - The function to evaluate
 * @param {number} [timeout] - The number of milliseconds to wait before returning an error
 * @param {number} [interval] - The number of milliseconds to wait between each retry
 */
SortdCommon.waitForResult = function(fn, timeout, interval) {

	timeout = timeout || 20000;
	interval = interval || 250;

	const checkStream = Kefir.interval(interval, null).map(() => fn());

	const timeoutStream = Kefir.later(timeout, null).flatMap(() => {
		const result = fn();
		return result ? Kefir.constant(result) : Kefir.constantError(new Error('Timeout in waitForResult'));
	});

	return Kefir.merge([checkStream, timeoutStream])
		.filter((v) => v !== undefined && v !== null)
		.take(1)
		.takeErrors(1);
};

/**
 * Wait for a function to return one or more elements
 * @param {string[]} selectors - Array of selectors to wait for
 * @param {string[]} [el] - The element to do the querySelector on, if not provided will use document
 * @param {number} [timeout] - The number of milliseconds to wait before returning an error
 * @param {number} [interval] - The number of milliseconds to wait between each retry
 */
SortdCommon.waitForElements = function(selectors, el, timeout, interval) {

	return SortdCommon.waitForResult(() => {

		el = el || document;
		if (selectors) {
			for (let i = 0; i < selectors.length; i++) {
				const result = el.querySelectorAll(selectors[i]);
				if (result) {
					return result;
				}
			}
		}
	}, timeout, interval);
}

/**
 * Creates an observable stream of messages for a certain action
 * @param {SortdMessages} messages - An instance of SortdMessages
 * @param {string} action - The message action to find
 */
SortdCommon.observeMessages = function(messages, action) {

	return Kefir.stream((emitter) => {

		const callback = (message) => emitter.emit(message);

		messages.addListener(action, callback);

		return () => {
			messages.removeListener(action, callback);
		};
	});
};

/**
 * Returns a new GUID
 */
SortdCommon.newGuid = function() {

	return 'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx'.replace(/[xy]/g, function (c) {
		var r = Math.random() * 16 | 0, v = c == 'x' ? r : (r & 0x3 | 0x8);
		return v.toString(16);
	});
};

SortdCommon.createDateFromTimeLabel = function(timeLabel) {

	const date = new Date();
	date.setHours(0, 0, 0, 0);

	switch (parseInt(timeLabel, 10)) {
		case constants.DATE_ID.TOMORROW:
			date.setDate(date.getDate() + 1);
			return date;
		case constants.DATE_ID.IN_TWO_DAYS:
			date.setDate(date.getDate() + 2);
			return date;
		case constants.DATE_ID.NEXT_WEEK:
			const day = date.getDay() || 7;
			date.setDate(date.getDate() + (8 - day));
			return date;
		case constants.DATE_ID.IN_A_MONTH:
			date.setMonth(date.getMonth() + 1);
			return date;
		case constants.DATE_ID.SOME_DAY:
			date.setMonth(date.getMonth() + 3);
			return date;
		default:	// defaults to constants.DATE_ID.TODAY
			return date;
	}
};

SortdCommon.createReminderFromTimeLabel = function(timeLabel) {

	const date = SortdCommon.createDateFromTimeLabel(timeLabel);
	if (date) {
		switch (parseInt(timeLabel, 10)) {
			case constants.DATE_ID.TOMORROW:
			case constants.DATE_ID.IN_TWO_DAYS:
			case constants.DATE_ID.NEXT_WEEK:
			case constants.DATE_ID.IN_A_MONTH:
			case constants.DATE_ID.SOME_DAY:
				date.setHours(8, 0, 0, 0);
				break;
			default:	// defaults to constants.DATE_ID.TODAY
				const now = new Date();
				let time = now.getHours() + 3;
				if (time < 8) {
					time = 8;
				}
				date.setHours(time, 0, 0, 0);
		}
	}

	return date;
};

/**
 * Gets the date id from the date
 * @param {Date} date - The date to check
 * @returns {Number} constants.DATE_ID
 */
SortdCommon.getDateIdFromDate = function(date) {

	date.setHours(0, 0, 0, 0);

	const dateTS = date.getTime();
	const compareDate = new Date();
	compareDate.setHours(0, 0, 0, 0);

	if (dateTS < compareDate.getTime()) {
		compareDate.setDate(compareDate.getDate() - 1);
		if (dateTS >= compareDate.getTime()) {
			return constants.DATE_ID.YESTERDAY;
		} else {
			return constants.DATE_ID.ALL_OVERDUE;
		}
	} else if (dateTS === compareDate.getTime()) {
		return constants.DATE_ID.TODAY;
	}

	compareDate.setDate(compareDate.getDate() + 1);
	if (dateTS === compareDate.getTime()) {
		return constants.DATE_ID.TOMORROW;
	}

	return constants.DATE_ID.ALL_UPCOMING;
};

/**
 * Format the date as Jan 4
 * @param {Date} date - The date to format
 * @param {number} [dateId] - The date id assocated with the date
 */
SortdCommon.formatShortDate = function(date, dateId) {

	dateId = dateId || this.getDateIdFromDate(date);

	if (dateId === constants.DATE_ID.TODAY) {
		return 'Today';
	} else if (dateId === constants.DATE_ID.TOMORROW) {
		return 'Tomorrow';
	} else {
		const day = date.getDate();
		const monthName = SortdCommon.monthsShort[date.getMonth()];

		return `${monthName} ${day}`;
	}
};

/**
 * Format the date as Jan 4, 9:29 PM
 * @param {Date} date - The date to format
 */
SortdCommon.formatShortDateTime = function(date) {

	const day = date.getDate();
	const monthName = SortdCommon.monthsShort[date.getMonth()];
	const hours = date.getHours() % 12 || 12;
	const mins = date.getMinutes().toString().padStart(2, '0');
	const amPm = date.getHours() >= 12 ? 'PM' : 'AM';

	return `${monthName} ${day}, ${hours}:${mins} ${amPm}`;
};

/**
 * Format the date as Mon
 * @param {Date} date - The date to format
 */
SortdCommon.formatShortDay = function(date) {

	const day = SortdCommon.daysShort[date.getDay()];

	return `${day}`;
};

/**
 * Format the date as Mon, 9:29 PM
 * @param {Date} date - The date to format
 */
SortdCommon.formatShortDayTime = function(date) {

	const hours = date.getHours() % 12 || 12;
	const mins = date.getMinutes().toString().padStart(2, '0');
	const amPm = date.getHours() >= 12 ? 'PM' : 'AM';

	return `${this.formatShortDay(date)}, ${hours}:${mins} ${amPm}`;
};

SortdCommon.escapeHtml = function(str) {
	return ('' + str).replace(SortdCommon.htmlEscaper, (match) => SortdCommon.htmlEscapes[match]);
};

SortdCommon.timeout = function(ttl) {

	return new Promise((resolve, reject) => {
		setTimeout(() => reject(new Error('Timeout occurred')), ttl);
	});
};

/**
 * Simulate a mouse click at the elements positions
 * @param {$.Element} element - The jQuery element to click
*/
SortdCommon.simulateRawClick = function(element) {

	if (element && element.length > 0) {

		var events = ['mousedown', 'mouseup', 'click'],
			pos = element.offset();

		try {
			// Try MouseEvent, only supported from Chrome 47
			events.forEach(function (evtType) {
				var evt = new MouseEvent(evtType, {
					bubbles: true,
					cancelable: true,
					view: window,
					screenX: pos.left,
					screenY: pos.top,
					dontCaptureSortd: true
				});
				element[0].dispatchEvent(evt);
			});

			return false;
		} catch (ex) {
			// If MouseEvent is not supported fall through to createEvent('MouseEvents')
		}

		events.forEach(function (evtType) {
			var evt = document.createEvent('MouseEvents');
			evt.initMouseEvent(evtType, true, true, window, 0, pos.left, pos.top, 0, 0, false, false, false, false, 0, null);
			element[0].dispatchEvent(evt);
		});
	}
};
