/**
 * Sortd
 * Copyright (c) 2019, Sortd Inc
 * https://www.sortd.com
 *
 * @fileoverview Defines the SortdDataService class
 * @name sortd-data
 */

class SortdDataService {

	constructor(options) {

		this.__FILE_NAME = 'sortd-data.js';

		this.messages = options.messages;

		this.messages.addListener('userInfo', this.handleUserInfo.bind(this));
	}

	showLoginRequired() {

		const destroy = () => {
			container.remove();
		}
		
		const container = document.createElement('div');
		const overlay = document.createElement('div');
        overlay.classList.add('sortd_default_Overlay');
        overlay.addEventListener('click', destroy);
		container.appendChild(overlay);
		
		const dropDownContainer = document.createElement('div');
		dropDownContainer.classList.add('sd_dropdown_container', 'sortdNotLoggedIn');
		const headingEl = document.createElement('div');
		headingEl.textContent = 'Sign in';
		headingEl.classList.add('sortdHeader');
		dropDownContainer.appendChild(headingEl);

		const form = document.createElement('form');
		let div = document.createElement('div');
		div.classList.add('form-group');
		div.innerHTML = 'To use this feature you need to<br>be signed into Sortd.';
		form.appendChild(div);

		div = document.createElement('div');
		div.classList.add('form-group', 'sd_txt_aln_rht');
		let button = document.createElement('button');
		button.textContent = 'Cancel';
		button.classList.add('sortdButton');
		button.onclick = destroy;
		div.appendChild(button);

		button = document.createElement('button');
		button.textContent = 'Sign In';
		button.classList.add('sortdButton', 'primary');
		button.onclick = (event) => {

			event.preventDefault();
			this.messages.sendMessageToExtension('showLogin', { fullscreen: false });
			destroy();
		};
		div.appendChild(button);
		form.appendChild(div);
		dropDownContainer.appendChild(form);
		container.appendChild(dropDownContainer);
		document.body.appendChild(container);		
	}

	sendMessageToFrame(message, data) {

		return new Promise((resolve, reject) => {

			this.messages.sendMessageToFrame(message, data || null, (result) => {

				if (result && result.errorStatus) {
					reject(new Error(`An error occurred (${result.errorStatus})`));
				} else {
					resolve(result);
				}
			});
		});
	}

	sendMessageToFrameWithTimeout(message, data, ttl) {

		return Promise.race([
			SortdCommon.timeout(ttl || constants.MESSAGE_TIMEOUT.TO_FRAME),
			this.sendMessageToFrame(message, data)
		]);
	}

	isLoggedIn() {		
		return this.sendMessageToFrameWithTimeout('ping');
	}

	callIfLoggedIn(fn, args = []) {

		this.isLoggedIn()
			.then(() => {
				try {
					fn(...args);
				} catch (ex) {
					console.log(ex);
				}
			})
			.catch(() => this.showLoginRequired());
	}

	handleUserInfo(message, next) {

		const user = message && message.data && message.data.user;
		if (user) {
			SortdDataService.userId = user.userId;
		}

		if (next) {
			next(null);
		}
	};

	getUserSettingValue(settingKey) {

		return this.getUserSetting(settingKey)
			.catch(() => this.getUserSettingFromExtension(settingKey))
			.then((setting) => setting && setting.value);
	}

	getUserSetting(settingKey) {

		return this.sendMessageToFrameWithTimeout('getUserSettings')
			.then((eventData) => {

				if (eventData && eventData.settings) {
					return eventData.settings && eventData.settings[settingKey];
				} else {
					return Promise.reject(new Error('Could not load user settings'));
				}
			});
	}

	getUserSettingFromExtension(settingKey) {
		
		return new Promise((resolve, reject) => {

			const email = sessionStorage.getItem(constants.STORAGE.KEYS.EMAIL);

			this.messages.sendMessageToEventPage('getUserSettings', { email }, (err, settings) => {

				if (!err && !settings) {
					err = new Error('Can\'t read user settings');
				}

				if (err) {
					reject(err);	
				} else {
					resolve(settings[settingKey]);
				}				
			});
		});
	}

	getListsByDashboard() {

		return this.sendMessageToFrameWithTimeout('getDashboardTaskLists')
			.then((eventData) => {

				try {
					let boards;
					if (eventData && eventData.sortdBoards) {
						boards = eventData.sortdBoards.map((entry) => {
		
							let name;
							if (entry.organisation.name) {
								if (entry.organisation.name.length > 10) {
									name = entry.organisation.name.slice(0, 7) + '...';
								} else {
									name = entry.organisation.name;
								}
								name += ' - ' + entry.dashboard.name;
							} else {
								name = entry.dashboard.name;
							}

							return {
								id: entry.dashboard.id,
								type: entry.dashboard.type,
								name,
								lists: entry.lists
							};
						});
					} else {
						boards = [];
					}

					return boards;
				} catch (ex) {
					return Promise.reject(ex);
				}
			});
	}

	/**
	 * Get the email reminder thread linked to threadId
	 * @param {string} threadId - The threadId to find
	 */
	getEmailReminderForThreadId(threadId) {
		return this.sendMessageToFrameWithTimeout('getEmailReminderForThreadId', { threadId }, 10000);
	}

	getReadReceiptEnabled() {

		return this.sendMessageToFrameWithTimeout('isReadReceiptsEnabled', { showPopup: false })
			.then((result) => (result && result.enabled) || false);
	};
}

SortdDataService.userId = null;
SortdDataService.instance = null;

SortdDataService.getInstance = function(options) {

	if (SortdDataService.instance === null) {
		SortdDataService.instance = new SortdDataService(options);
	}

	return SortdDataService.instance;
}