/**
 * Sortd
 * Copyright (c) 2019, Sortd Inc
 * https://www.sortd.com
 *
 * @fileoverview Defines services to be used by Sortd extension
 * @name services
 */

'use strict';

/**
* Define the SortdErrorHandler module.
*/
const SortdErrorHandler = (function () {

	const messages = {
		RETRYING: 'Unable to contact Sortd. Retrying...',
		FAILED: 'Unable to contact Sortd. <a href="FAIL_URL" target="_blank" style="color:#ffffff">Click here for more info.</a>'
	};

	const MAX_BODY_SIZE = 1024 * 70;

	function SortdErrorHandler(options) {

		var self = this,
			myOverrides = {};

		if (options && options.config) {
			self._config = options.config;
		} else {
			if (typeof overrides !== 'undefined') {
				myOverrides = overrides;
			}
			self._config = new config(myOverrides);
		}

		this.emailAddress = null;
	}

	SortdErrorHandler.prototype.getEmailAddress = function () {

		if (!this.emailAddress) {
			const emailAddress = sessionStorage.getItem(constants.STORAGE.KEYS.EMAIL);
			if (emailAddress) {
				this.emailAddress = emailAddress;
			}
		}

		return this.emailAddress;
	};

    /**
	 * Handle errors by the loader / messages
	 */
	SortdErrorHandler.prototype.handle = function (error, fileName) {

		var self = this,
			options;

		options = {
			message: constants.ERROR_MESSAGE.EXTENSION_ERROR,
			technical: error.message,
			fileName: fileName,
			originalError: error
		};
		if (error.message !== 'Timeout occurred' &&
			(error.status !== 0 && error.status !== 502 && error.status !== 503) &&
			((error.message && error.message !== '{}') || options.technical || options.stack)) {

			if ((!error.message || error.message === 'Script error.') &&
				(!error.stack || error.stack === 'url: ; lineNumber: 0; column: 0' ||
					error.stack === 'url: null; lineNumber: null; column: null' ||
					error.stack === 'url: ; lineNumber: ; column: ')) {

				// Don't log empty errors
				return;
			}

			self._handle(options);
		}
	};

	/**
	 * Handle errors by the loader / messages
	 */
	SortdErrorHandler.prototype._handle = function (error) {

		const self = this;

		self.logMessage({ level: constants.LOG.TYPE.ERROR, message: error.message, meta: error });

		const frame = document.getElementById('sortdWrapper');
		let target = frame && frame.contentWindow;
		if (!(target && target.postMessage)) {
			target = window;
		}

		if (target && target.postMessage && typeof target.postMessage === 'function' && !error.logConsoleOnly && !error.logOnly) {
			target.postMessage({
				action: 'handleExtensionError',
				owner: 'sortd',
				source: 'ce',
				destination: 'frame',
				data: JSON.stringify({
					error: error.originalError,
					fileName: error.fileName || '',
					notifyOnly: true
				})
			}, 'chrome-extension://' + chrome.runtime.id);
		}
	};
	/**
	 * Log a message to the server
	 * @param {Object} options - the log level (error, debug, etc)
	 * @param {String} options.level - the log level (error, debug, etc)
	 * @param {String} options.message - The text message to log
	 * @param {Object} options.meta - metadata related to the message
	 * @param {Function} next - the result callback
	 */
	SortdErrorHandler.prototype.logMessage = function (options, next) {

		var self = this;

		try {
			if (!options.meta) {
				options.meta = {};
			}
			let logToServer = !options.meta.logConsoleOnly;
			options.meta = self.createMetaObject(options.meta);
			if (Object.keys(options.meta).length === 0) {
				logToServer = false;
			}

			options.meta.time = new Date().toISOString();
			options.meta.ceVersion = self._config.version;
			options.name = 'LogMessage';

			console.log(options.meta.time + ': ' + JSON.stringify(options));

			if (logToServer) {
				let data = JSON.stringify(options.meta);
				if (data.length > MAX_BODY_SIZE) {
					options.meta = {
						truncated: data.slice(0, MAX_BODY_SIZE - 1000)
					};
				}

				self._sendMessageToServer(options);
			}

			if (next) {
				next(null);
			}
		} catch (ex) {
			console.log('Error in extension sortd error handler. Cannot log message: ' + options.message.toString() + ' error: ' + ex.message, ex);
			if (next) {
				next(null);
			}
		}
	};

	/**
	 * @param {Object} logMessage - the log message
	 * @param {String} logMessage.level - the log level (error, debug, etc)
	 * @param {String} logMessage.message - The text message to log
	 * @param {Object} logMessage.meta - metadata related to the message
	 * @param {Function} callback - the result callback
	 */
	SortdErrorHandler.prototype._sendMessageToServer = function (logMessage, callback) {

		logMessage.emailAddress = this.getEmailAddress();

		new SortdHttp().post(
			this._config.chrome.host.mail3.baseUrl + '/clientlog/logMessage',
			logMessage,
			(err, result) => {

				if (err) {
					console.log('Error in extension sortd error handler. Cannot sendMessageToServer message: ' + logMessage.message.toString() + ' error: ' + err.message, err);
				} else {
					callback && callback(null, result);
				}
			});
	};

	/**
	 * Checks if meta data is an Error, if so convert to normal object
	 * If meta has an original error, also convert that to a meta object
	 * @param {Object} data - The data object
	 * @returns {Object} - The resulting meta data
	 * @private
	 */
	SortdErrorHandler.prototype.createMetaObject = function (data) {

		var self = this,
			obj = {};

		if (data) {
			if (data instanceof Error) {
				obj = {
					message: data.message,
					name: data.name,
					technical: data.technical,
					stack: data.stack,
					fileName: data.fileName,
					lineNumber: data.lineNumber,
					columnNumber: data.columnNumber,
					status: data.status,
					statusText: data.statusText,
					url: data.url,
					retriesRemaining: data.retriesRemaining
				};
			} else if (typeof data === 'string') {
				obj.message = data;
			} else {
				obj = data;
			}

			if (data.originalError) {
				obj.originalError = self.createMetaObject(data.originalError);
			}

			if (data.err) {
				obj.err = self.createMetaObject(data.err);
			}

			if (data.error) {
				obj.error = self.createMetaObject(data.error);
			}

			if (data.originalData && data.status !== 502) {
				obj.originalData = self.createMetaObject(data.originalData);
			}
		}

		return obj;
	};

	/**
	 * Remove the notification
	 */
	SortdErrorHandler.prototype.removeNotification = function () {
		$('#sortdLoaderErrorNotification').remove();
	};

	/**
	 * Notify the user that we could not load the extension
	 */
	SortdErrorHandler.prototype.notifyUser = function (isRetry) {

		const self = this;
		let message;

		self.removeNotification();

		if (isRetry) {
			message = messages.RETRYING;
		} else {
			message = messages.FAILED.replace('FAIL_URL', self._config.chrome.sortd.failedToLoadURL);
		}

		const el = $('<div id="sortdLoaderErrorNotification" ' +
			'style="z-index:50000;position:absolute;top:0px;right:0px;background-color:#606060;color:#ffffff;' +
			'font-size:83%;padding:3px 10px;cursor:pointer">' + message + '</div>');
		el.click(() => self.removeNotification());
		el.find('a').click((event) => event.stopPropagation());
		el.appendTo('body');
	};

	return SortdErrorHandler;
}());