/**
 * Sortd
 * Copyright (c) 2019, Sortd Inc
 * https://www.sortd.com
 *
 * @fileoverview Functions to handle messaging between Sortd Frame and Extension
 * @name sortd-loader
 */

var SortdFrameWrapper = (function () {

	'use strict';

	/**
	 * Create the class
	 * @class SortdFrameWrapper
	 */
	function SortdFrameWrapper () {

		this.FILE_NAME = 'sortd-frame-wrapper.js';
		this._errorHandler = new SortdErrorHandler();
		this._gaService = null;

		var myOverrides = {};
		if (typeof overrides !== 'undefined') {
			myOverrides = overrides;
		}
		this._config = new config(myOverrides);

		this._messageOriginExtension = 'chrome-extension://' + chrome.runtime.id;
		this._allowedMessageOrigins = [
			this._config.messageOriginGmail,
			this._config.chrome.host.mail3.baseUrl,
			this._messageOriginExtension
		];
	}

	/**
	 * Logs the error
	 * @param {Error} err - The error
	 * @private
	 */
	SortdFrameWrapper.prototype._handleError = function (err) {
		this._errorHandler.handle(err, this.FILE_NAME);
	};

	/**
	 * Logs the error
	 * @param {String} action - The action
	 * @param {Object} [data] - The data to send
	 * @param {String} [destination] - Where to send the message to, defaults to 'ce'
	 * @private
	 */
	SortdFrameWrapper.prototype._sendMessage = function (action, data, destination) {

		let targetOrigin = this._config.messageOriginGmail;
		if (destination === 'frame') {
			targetOrigin = this._messageOriginExtension;
		}

		window.parent.postMessage({
			action: action,
			owner: 'sortd',
			source: 'wrapper',
			destination: (destination || 'ce'),
			data: data
		}, targetOrigin);
	};

	/**
	 * Receive a message from the window
	 * @param {Event} evt - The message event that was received
	 */
	SortdFrameWrapper.prototype._handleMessage = function (evt) {

		try {
			const message = evt.data;

			if (message && message.owner === 'sortd' && this._allowedMessageOrigins.indexOf(evt.origin) !== -1) {

				let target;
				let targetOrigin;

				if (message.action === 'setSortdFrameUrl') {
					document.getElementById('sortdFrame').src = message.data.url;
					this._sendMessage('timer', { timerEvent: 'SetFrameURL' });
					return;
				}

				switch (message.source) {
					case 'frame':
						if (message.destination === 'ce' || message.destination === 'bgpage' || message.destination === 'jsce') {
							target = window.parent;
							targetOrigin = this._config.messageOriginGmail;
						}
						break;
					case 'ce':
					case 'jsce':
					case 'bgpage':
						if (message.destination === 'frame') {
							target = document.getElementById('sortdFrame').contentWindow;
							targetOrigin = this._config.chrome.host.mail3.baseUrl;
						}
						break;
				}

				if (target && target.postMessage && targetOrigin) {
					target.postMessage(message, targetOrigin);
				}
			}
		} catch (ex) {
			this._handleError(ex);
		}
	};

	/**
	 * Create listeners to pass messages between extension to event page
	 * @private
	 */
	SortdFrameWrapper.prototype.createListeners = function () {
		window.addEventListener('message', this._handleMessage.bind(this), false);
	};

	/**
	 * Initialize listeners
	 * @private
	 */
	SortdFrameWrapper.prototype.init = function () {

		this.createListeners();

		this._sendMessage('frameWrapperLoaded');

		// This is to get drag & drop working for Gmail compose / reply
		// When Sortd it open fullscreen
		document.addEventListener('dragstart', () => {
			this._sendMessage('dragstartSortdFrame', {}, 'jsce');
		}, false);

		document.addEventListener('dragover', (event) => {
			event.preventDefault();
			event.stopPropagation();

			this._sendMessage('dragoverSortdFrame', {}, 'jsce');
		}, false);
	};


	return SortdFrameWrapper;
}());

(function () {
	/**
	 * Init a new instance and initialize
	 */
	new SortdFrameWrapper().init();
}());
