/**
 * Sortd
 * Copyright (c) 2019, Sortd Inc
 * https://www.sortd.com
 *
 * @fileoverview Defines the SortdGmailSendService class
 * @name sortd-gmail-send-service
 */

class SortdGmailSendService {

	constructor(options) {

		this.__FILE_NAME = 'sortd-gmail-send-service.js';

		this.config = options.config;
		this.messages = options.messages;
		this.errorHandler = options.errorHandler;
		this.gmailApi = options.gmailApi;
		this.listItems = options.listItems;
		this.inboxSdk = options.inboxSdk;
		this.sortdDataService = SortdDataService.getInstance({ messages: this.messages });
		this.shortcutHelper = SortdShortcutHelper.getInstance({
			config: this.config,
			messages: this.messages,
			errorHandler: this.errorHandler
		});

		this.isDestroyed = false;
		this.dropDownActive = null;
		this.isActive = false;
		this.handleOpenedThreadViewTimer = null;
		this.lastPinHandlerResult = null;

		this.buttons = {};
		this.activeButtonId = null;
		this.activeNotification = null;
		this.activeDropDown = null;
		this.showDayPlanLinksEnabled = true;
		this.displayMode = null;

		this.openThread = {
			buttonId: null,
			listItemSub: null,
			bodyEl: null,
			archiveCompleteAdded: false
		};

		this.BAR_DOM_MAP = SortdGmailApiDomMap.GMAIL_NOTIFICATION_BAR[0].CLASSES;

		this.messages.addListener('checkForUpdatedThreadInfo', this.handleCheckForUpdatedThreadInfo.bind(this));
		this.messages.addListener('showDayPlanLinks', this.handleShowDayPlanLinks.bind(this));
		this.messages.addListener('showAddToBoard', this.handleShowAddToBoard.bind(this));
		this.messages.addListener('cancelShowAddToBoard', this.handleCancelShowAddToBoard.bind(this));

		this.shortcutHelper.addListener(
			constants.SHORTCUTS.KEY_CODES.E,
			this.handleArchiveMessageShortcut.bind(this));

		this.shortcutHelper.addListener(
			`${constants.SHORTCUTS.KEY_CODES.SHIFT}_${constants.SHORTCUTS.KEY_CODES.U}`,
			this.handleMarkUnreadShortcut.bind(this));

		this.shortcutHelper.addListener(
			constants.SHORTCUTS.KEY_CODES.S,
			this.handleMarkAsStarShortcut.bind(this));
	}

	handleError(error) {

		if (this.errorHandler) {
			this.errorHandler.handle(error, this.__FILE_NAME);
		}
	}

	handleShowAddToBoard(options, next) {

		if (this.openThread && this.openThread.buttonId && options && options.data) {

			this.pinClickHandler(
				{ position: options.data.position },
				this.openThread.buttonId,
				options.data.eventOrigin);

			next();
		}
	}

	handleCancelShowAddToBoard() {

		if (this.lastPinHandlerResult && this.lastPinHandlerResult.destroy) {
			this.lastPinHandlerResult.destroy();
		}
	}

	handleOpenedThreadViewAfterDelay(threadView) {

		const buttonId = SortdCommon.newGuid();
		let loggedIn = true;
		let button;
		let threadId;
		let deregisterMainMoreMenuOverrides;
		let deregisterMessageMoreMenuOverrides;

		if (this.handleOpenedThreadViewTimer) {
			clearTimeout(this.handleOpenedThreadViewTimer);
			this.handleOpenedThreadViewTimer = null;
		}

		this.messages.sendMessageToExtension('getDisplayMode', {}, (result) => {
			this.displayMode = result.displayMode;
		});

		return threadView.getThreadIDAsync()
			.then((thId) => {

				threadId = thId;
				button = this.buttons[buttonId] = {
					eventOrigins: {},
					openedFromThread: true,
					buttonId,
					threadId,
					itemInfo: {},
					currentItemInfo: {},
					getSubject: () => threadView.getSubject()
				};

				this.openThread.buttonId = buttonId;
				this.openThread.threadId = threadId;

				threadView.on('destroy', (() => {

					if (this.openThread.listItemSub) {
						this.openThread.listItemSub.unsubscribe();
					}

					if (deregisterMainMoreMenuOverrides) {
						deregisterMainMoreMenuOverrides();
					}

					if (deregisterMessageMoreMenuOverrides) {
						deregisterMessageMoreMenuOverrides();
					}

					if (buttonId !== this.activeButtonId) {
						delete this.buttons[buttonId];
					}

					this.openThread.buttonId = null;
					this.openThread.messageId = null;
					this.openThread.threadId = null;
					this.openThread.listItemSub = null;
					this.openThread.bodyEl = null
					if (this.openThread.archiveCompleteAdded) {
						this.openThread.archiveCompleteAdded = false;
						const button = document.getElementById('sortdCompleteArchiveThread');
						if (button) {
							button.remove();
						}
					}
				}));
			})
			.then(() => this.sortdDataService.isLoggedIn())
			.catch(() => loggedIn = false)
			.then(() => this.sortdDataService.getUserSettingValue(constants.USER_SETTING_KEYS.GMAIL_MODE_PREVIEW_BAR))
			.catch(() => true)
			.then((settingValue) => {

				if (!settingValue) {
					return;
				}

				const bar = threadView.addNoticeBar();
				const barElement = this.createSortdListItemBarElement(buttonId, constants.TRACKER.EVENT_ORIGINS.PREVIEW_BAR);
				if (barElement) {
					bar.el.appendChild(barElement);
					bar.el.classList.add('sd_threadview_bar');
				}
			})
			.then(() => threadView.getMessageViews())
			.then((messageViews) => {
				if (messageViews && messageViews.length && messageViews[0]) {
					return messageViews[0].getMessageIDAsync();
				}
			})
			.then((messageId) => {

				if (messageId && !button.messageId && this.openThread.threadId === threadId) {
					this.openThread.messageId = messageId;
					button.messageId = messageId;
				}

				if (loggedIn) {
					this.listenForListItemChanges(buttonId);

					if (this.showDayPlanLinksEnabled) {
						this.registerMarkAsUnreadOverride();
					}
					deregisterMainMoreMenuOverrides = this.registerMainMoreMenuOverrides();

					const selectors = SortdGmailApiDomMap.VIEW_MAIL[0].BODY && SortdGmailApiDomMap.VIEW_MAIL[0].BODY.SELECTORS;

					this.gmailApi.Dom.waitForElements(selectors).onValue((elements) => {

						if (elements && elements.length > 0) {
							this.openThread.bodyEl = elements[0];
						}

						if (this.showDayPlanLinksEnabled) {
							this.registerStarOverride();
						}
						deregisterMessageMoreMenuOverrides = this.registerMessageMoreMenuOverrides();
					});
				}

			})
			.catch((err) => this.handleError(err))
	}

	handleOpenedThreadView(threadView) {

		// Only handle the open after a short delay as the way the thread is opened by inboxSdk
		// makes it reload immediately
		let delay = 1000;
		if (this.handleOpenedThreadViewTimer) {
			clearTimeout(this.handleOpenedThreadViewTimer);
			delay = 500;
		}

		this.handleOpenedThreadViewTimer = setTimeout(() => {
			this.handleOpenedThreadViewAfterDelay(threadView);
		}, delay);
	}

	handleOpenedComposeView(composeView, composeData) {

		if (!this.gmailApi.Notifications.enabled) {
			return Promise.resolve();
		}

		return this.sortdDataService
			.getUserSettingValue(constants.USER_SETTING_KEYS.GMAIL_MODE_PREVIEW_BAR)
			.catch(() => true)
			.then((settingValue) => {
				
				if (composeData) {

					if (composeData.clearBody === true) {
						composeView.setBodyText('');

						if (composeData.recipients && composeData.recipients.to && composeData.recipients.to.length > 0) {
							setTimeout(() => composeView.getBodyElement().focus(), 10);
						}
					}

					if (typeof composeData.subject !== 'undefined') {
						composeView.setSubject(composeData.subject);
					}

					if (composeData.recipients) {
						if (composeData.recipients.to) {
							composeView.setToRecipients(composeData.recipients.to);
						}
						if (composeData.recipients.cc) {
							composeView.setCcRecipients(composeData.recipients.cc);
						}
						if (composeData.recipients.bcc) {
							composeView.setBccRecipients(composeData.recipients.bcc);
						}
					}
				}

				let buttonId;
				let threadId = composeView.getThreadID();
				let itemInfo = {};
				let currentItemInfo = {};
				
				if (threadId) {
					for (const key in this.buttons) {
						if (this.buttons.hasOwnProperty(key)) {
							const button = this.buttons[key];
							if (button.threadId === threadId) {
								// Sent would be true for a compose use case if the user cancelled the send, or pressed undo
								// Mark button as cancelled, set threadId to null and hide the bar (which will also remove the current button)
								// so that it gets treated like a new compose window (with the previous selection copied)
								if (button.sent === true) {
									itemInfo = button.itemInfo;
									currentItemInfo = button.currentItemInfo;
									button.cancelled = true;
									threadId = null;
									this.hideSendAndRemindBar(button.buttonId);
								} else {
									buttonId = button.buttonId;
								}
								break;
							}
						}
					}
				}
				
				if (!buttonId) {
					buttonId = SortdCommon.newGuid();
					this.buttons[buttonId] = {
						eventOrigins: {},
						openedFromThread: false,
						buttonId,
						threadId,
						itemInfo,
						currentItemInfo,
						sent: false,
						getSubject: () => composeView.getSubject()
					}
				}

				if (composeData && composeData.linkedTask) {
					this.buttons[buttonId].addThreadToTask = true;
					this.buttons[buttonId].itemInfo = {
						item: {
							itemId: composeData.linkedTask.taskId,
							title: composeData.linkedTask.title
						},
						list: {
							listId: composeData.linkedTask.listId,
							title: composeData.linkedTask.listTitle
						},
						dashboard: {
							dashboardId: composeData.linkedTask.boardId,
							boardType: composeData.linkedTask.boardType,
							title: composeData.linkedTask. boardTitle
						}
					};
					this.buttons[buttonId].currentItemInfo = {
						item: Object.assign({}, this.buttons[buttonId].itemInfo.item),
						list: Object.assign({}, this.buttons[buttonId].itemInfo.list),
						dashboard: Object.assign({}, this.buttons[buttonId].itemInfo.dashboard)
					};
				}

				if (settingValue) {
					const bar = composeView.addComposeNotice();
					const barElement = this.createSortdListItemBarElement(buttonId, constants.TRACKER.EVENT_ORIGINS.SEND_BUTTON_MENU);
					if (barElement) {
						bar.el.appendChild(barElement);
						if (composeView.isInlineReplyForm()) {
							bar.el.classList.add('sd_threadview_bar', 'sd_reply');
						} else {
							bar.el.classList.add('sd_compose_bar');
						}
						setTimeout(() => this.setActiveBarValues(buttonId), 10);
					}
				}

				const fixerEl = document.body.querySelector(this.BAR_DOM_MAP.INBOXSDK_FIXER);
				if (fixerEl) {
					fixerEl.style.setProperty('position', 'fixed', 'important');
				}

				composeView.on('destroy', () => {

					try {
						if (buttonId) {
							const button = this.buttons[buttonId];
							// Check for explicit false, not falsy, otherwise the button
							// will be deleted when a reply is cancelled
							if (button && button.sent === false) {
								delete this.buttons[buttonId];
							}
						}
					} catch (ex) {
						this.handleError(ex);
						this.hideSendAndRemindBar(buttonId);
					}
				});

				composeView.on('sending', () => {

					try {
						if (buttonId) {
							const button = this.buttons[buttonId];

							if (button) {
								button.sent = true;
							}

							this.activeButtonId = buttonId;
							this.showSendAndRemindBar(buttonId);
						}
					} catch (ex) {
						this.handleError(ex);
						this.hideSendAndRemindBar(buttonId);
					}
				});

				composeView.on('sent', (evt) => {

					try {
						const subject = composeView.getSubject() || '(no subject)';
						const to = composeView.getToRecipients();

						let fromPromise;
						try {
							fromPromise = Promise.resolve(composeView.getFromContact());
						} catch (ex) {
							fromPromise = this.gmailApi.User.getEmailAddress().then((address) => {
								return { emailAddress: address, name: address };
							});
						}

						return Promise.all([evt.getMessageID(), evt.getThreadID(), fromPromise])
							.then((result) => {

								if (buttonId) {
									const button = this.buttons[buttonId];

									if (button) {
										button.sent = true;
										button.messageId = result[0];
										button.threadId = result[1];
										const from = result[2];

										if (button.addThreadToTask) {
											button.mailContentItem = {
												hexThreadId: button.threadId,
												flags: { seen: true, flagged: false, important: false },
												labelIds: ['SENT'],
												threadSummary: {
													title: subject,
													messageCount: 1,
													mostRecentMessage: {
														addresses: {
															from: { email: from.emailAddress, name: from.name },
															to: to.map((r) => ({ name: r.name, email: r.emailAddress }))
														},
														date: new Date(),
														dateReceived: new Date(),
														subject,
													},
													recentSenders: [{
														dateReceived: new Date(),
														seen: true,
														sender: { email: from.emailAddress, name: from.name }
													}]
												},
												folderName: 'sent'
											};
										}
									}
								}
							})
							.catch((ex) => {
								this.handleError(ex);
								this.hideSendAndRemindBar(buttonId);
							});
					} catch (ex) {
						this.handleError(ex);
						this.hideSendAndRemindBar(buttonId);
					}
				});
			}).catch((err) => this.handleError(err));
	}

	handleArchiveMessageShortcut() {

		// If the popup is already showing, just archive by propagating the event
		const existingPopup = document.getElementById('sortdCompleteArchivePopup');
		if (existingPopup) {
			existingPopup.remove();
			return true;
		}

		if (this.openThread.buttonId && this.openThread.threadId) {

			const button = this.buttons[this.openThread.buttonId];
			const completeEl = document.getElementById('sortdCompleteArchiveThread');
			if (button && completeEl) {

				let result;

				const closeDropdown = () => {
					result && result.destroy();
				};

				const el = document.createElement('div');
				el.id = 'sortdCompleteArchivePopup';
				const headingEl = document.createElement('div');
				headingEl.textContent = 'Archive Conversation';
				headingEl.classList.add('sortdHeader');
				el.appendChild(headingEl);
				el.appendChild(this.gmailApi.Dom.createHorzSeparatorElement());

				const ul = document.createElement('ul');
				ul.classList.add('sd_with_hdr');
				let li = document.createElement('li');
				li.textContent = 'Archive';
				li.dataset.sdSelectionIndex = 0;
				li.addEventListener('click', () => {

					this.clickArchiveButton();
					closeDropdown();
				});
				ul.appendChild(li);
				li = document.createElement('li');
				li.textContent = 'Archive & Complete';
				li.dataset.sdSelectionIndex = 1;
				li.addEventListener('click', () => {

					if (this.openThread.threadId) {
						SortdCommon.simulateRawClick($(completeEl));
					} else {
						this.clickArchiveButton();
					}
					closeDropdown();
				});
				ul.appendChild(li);
				el.appendChild(ul);

				result = this.gmailApi.Dom.showDropDown({
					el,
					positionToElement: document.body,
					centerToElement: true,
					destroyedCallback: () => {
						this.gmailApi.Dom.unbindArrowNavigation(closeDropdown, closeDropdown);
					}
				});

				setTimeout(() => this.gmailApi.Dom.bindArrowNavigation(el, closeDropdown, closeDropdown), 500);

				return false;
			}
		}

		// Otherwise propagate the event so Gmail handles it
		return true;
	}

	handleMarkUnreadShortcut() {

		if (this.openThread.buttonId && this.openThread.threadId) {

			const button = this.buttons[this.openThread.buttonId];
			if (button) {

				const popup = document.querySelector('.sd_with_picker');
				if (!popup) {
					this.clickMarkAsUnreadButton();
				}

				return false;
			}
		}

		// Otherwise propagate the event so Gmail handles it
		return true;
	}

	handleMarkAsStarShortcut() {

		if (this.openThread.buttonId && this.openThread.threadId) {

			const button = this.buttons[this.openThread.buttonId];
			if (button) {

				const popup = document.querySelector('.sd_with_picker');
				if (!popup) {
					this.clickMarkAsStar();
					return false;
				}
			}
		}

		// Otherwise propagate the event so Gmail handles it
		return true;
	}

	checkForEmailReminder(buttonId, threadId) {

		this.sortdDataService.getEmailReminderForThreadId(threadId).then((reminder) => {

			if (reminder && reminder.reminderId) {
				const button = this.buttons[buttonId];
				if (!button) {
					return;
				}
		
				button.itemInfo = button.itemInfo || {};
				button.currentItemInfo = button.currentItemInfo || {};

				button.itemInfo.reminderDate = reminder.reminderDate ? new Date(reminder.reminderDate) : null;

				button.currentItemInfo.reminderDate = reminder.reminderDate ? new Date(reminder.reminderDate) : null;
				button.currentItemInfo.reminderId = reminder.reminderId;

				this.setActiveBarValues(buttonId);
			}
		});
	}

	handleNewListItemValue(buttonId, listItemInfo) {

		const button = this.buttons[buttonId];
		if (!button) {
			return;
		}

		const itemId = listItemInfo && listItemInfo.itemId;
		if (!itemId && button.addThreadToTask) {
			return;
		}

		button.itemInfo = {};
		button.currentItemInfo = {};

		if (listItemInfo) {

			button.itemInfo.dashboard = Object.assign({}, listItemInfo.dashboard);
			button.itemInfo.list = Object.assign({}, listItemInfo.list);
			button.itemInfo.item = this.mapItem(listItemInfo);

			button.currentItemInfo.dashboard = Object.assign({}, listItemInfo.dashboard);
			button.currentItemInfo.list = Object.assign({}, listItemInfo.list);
			button.currentItemInfo.item = this.mapItem(listItemInfo);

			if (listItemInfo && listItemInfo.content && listItemInfo.content.reminders) {
				const reminder = this.listItems.getUserPopupReminder(listItemInfo.content.reminders);
				if (reminder && reminder.date) {
					button.itemInfo.reminderDate = new Date(reminder.date);
					button.currentItemInfo.reminderDate = new Date(reminder.date);
				}
			}
		}

		if (!(button.itemInfo.reminderDate)) {
			this.checkForEmailReminder(buttonId, this.openThread.threadId);
		}

		this.setActiveBarValues(buttonId);
		this.addArchiveCompleteButton();

		if (button.itemInfo.item) {
			this.messages.sendMessageToExtension('gmailThreadTaskChanged', {
				taskId: button.itemInfo.item.itemId
			});
		}
	}

	/**
	 * Maps a list item to a flat structure
	 * @param {ListItem} listItem 
	 */
	mapItem(listItem) {

		let item;
		if (listItem) {
			item = {
				itemId: listItem.itemId,
				title: listItem.title
			};

			if (listItem.content && (listItem.contentType === constants.items.CONTENT_TYPES.TASK ||
				listItem.contentType === constants.items.CONTENT_TYPES.SALE)
			) {
				item.nextStep = listItem.content.message && listItem.content.message.text;
				item.currentWorkflowStatusId = listItem.content.currentWorkflowStatusId;

				if (listItem.content.userPlannedDates) {
					const plannedDate = listItem.content.userPlannedDates.find((p) => p.createdBy === SortdDataService.userId);
					if (plannedDate) {
						item.plannedDate = plannedDate.date ? new Date(plannedDate.date) : null;
						item.plannedDateOrder = plannedDate.displayOrder || null;
					}
				}

				if (listItem.content.deal && listItem.contentType === constants.items.CONTENT_TYPES.SALE) {
					item.dealValue = listItem.content.deal.value;
				}
			}
		}

		return item;
	}

	handleCheckForUpdatedThreadInfo() {

		if (this.openThread && this.openThread.buttonId) {
			this.listenForListItemChanges(this.openThread.buttonId);
		}
	}

	handleShowDayPlanLinks(message, next) {
		this.showDayPlanLinksEnabled = message && message.data && message.data.enabled || false;
		next({});
	};

	setActiveBarValues(buttonId) {

		const button = this.buttons[buttonId];
		if (!button) {
			return;
		}

		if (button.itemInfo) {
			this.pinToSelectedList(buttonId, null, button.itemInfo.dashboard, button.itemInfo.list, false);
			this.setReminder(buttonId, null, button.itemInfo.reminderDate, false);

			const item = button.itemInfo.item;
			if (item) {
				this.setNextStep(buttonId, null, item.nextStep, null, false);
				this.setTaskDetails(buttonId, null, item.title, item.dealName, false);
				this.setPlannedDate(buttonId, null, item.plannedDate, item.plannedDateOrder, false);
			} else {
				this.setNextStep(buttonId, null, null, null, false);
				this.setTaskDetails(buttonId, null, null, null, false);
				this.setPlannedDate(buttonId, null, null, null, false);
			}
		} else {
			this.pinToSelectedList(buttonId, null, null, null, false);
			this.setReminder(buttonId, null, null, false);
			this.setNextStep(buttonId, null, null, null, false);
			this.setTaskDetails(buttonId, null, null, null, false);
		}
	}

	listenForListItemChanges(buttonId) {

		if (this.openThread.listItemSub || this.openThread.buttonId !== buttonId) {
			return;
		}

		this.openThread.listItemSub = this.listItems.getObsListItemByThreadId(this.openThread.threadId)
			.observe((listItem) => this.handleNewListItemValue(buttonId, listItem));	
	}

	showSendAndRemindBar(buttonId) {

		if (!buttonId || !this.gmailApi.Notifications.enabled) {
			return;
		}

		const button = this.buttons[buttonId];
		if (!button) {
			return;
		}

		this.isDestroyed = false;

		const notificationContent = document.createElement('span');
		const buttonClasses = this.BAR_DOM_MAP.MSG_BUTTONS.slice();
		buttonClasses.push('sd_gmail_btn');
		buttonClasses.push('sd_icon')

		let textEl = this.gmailApi.Dom.createTextElement('Message Sent.');
		textEl.classList.add('sd_resize_curr');
		notificationContent.appendChild(textEl);
		notificationContent.appendChild(this.gmailApi.Dom.createVertSeparatorElement());

		const remindEl = document.createElement('span');
		const remindButton = this.gmailApi.Dom.createGmailButtonElement(
			'Remind me',
			(event) => this.sortdDataService.callIfLoggedIn(this.reminderClickHandler.bind(this), [event, buttonId, constants.TRACKER.EVENT_ORIGINS.SEND_CONFIRMATION]),
			[...buttonClasses, 'sd_reminder'],
			{ defaultText: 'Remind me', sdRemindBtn: 'bar' });
		remindEl.appendChild(remindButton);

		const datePlanEl = document.createElement('span');
		const datePlanButton = this.gmailApi.Dom.createGmailButtonElement(
			'Add to Day plan',
			(event) => this.sortdDataService.callIfLoggedIn(this.datePlanClickHandler.bind(this), [event, buttonId, constants.TRACKER.EVENT_ORIGINS.SEND_CONFIRMATION]),
			[...buttonClasses, 'sd_dateplan'],
			{ defaultText: 'Add to Day plan', sdDatePlanBtn: 'bar' });
		datePlanEl.appendChild(datePlanButton);
		notificationContent.appendChild(datePlanEl);

		const pinEl = document.createElement('span');
		const pinButton = this.gmailApi.Dom.createGmailButtonElement('Add to board',
			(event) => this.sortdDataService.callIfLoggedIn(this.pinClickHandler.bind(this), [event, buttonId, constants.TRACKER.EVENT_ORIGINS.SEND_CONFIRMATION]),
			[...buttonClasses, 'sd_pinto'],
			{ sortdPin: 'N', defaultText: 'Add to board', sdPinBtn: 'bar' });
		pinEl.appendChild(pinButton);
		notificationContent.appendChild(pinEl);

		notificationContent.appendChild(remindEl);

		const nextStepEl = document.createElement('span');
		const nextStepText = document.createElement('input');
		nextStepText.type = 'text';
		nextStepText.placeholder = 'Next step ...';
		nextStepText.classList.add('sd_ns_text_bar');
		nextStepText.id = 'sd_ns_text_bar_txt_input';
		nextStepText.addEventListener('blur', (event) => this.setNextStep(buttonId, constants.TRACKER.EVENT_ORIGINS.SEND_CONFIRMATION, event.target.value));
		nextStepText.addEventListener('focus', () => nextStepText.classList.remove('selected', 'animate'));
		nextStepText.addEventListener('keyup', (event) => {
			if (event.keyCode === 13) {
				document.activeElement.blur();
			}
		});
		nextStepText.addEventListener('click', () => nextStepText.select());
		nextStepEl.appendChild(nextStepText);
		notificationContent.appendChild(nextStepEl);

		if (button.itemInfo && button.itemInfo.item && button.itemInfo.item.itemId) {
			const moreOptionsButton = this.gmailApi.Dom.createGmailButtonElement(
				'...',
				(event) => this.sortdDataService.callIfLoggedIn(this.moreOptionsClickHandler.bind(this), [event, buttonId, constants.TRACKER.EVENT_ORIGINS.SEND_CONFIRMATION]),
				[...this.BAR_DOM_MAP.MSG_BUTTONS],
				{ sdMoreBtn: buttonId });
			notificationContent.appendChild(moreOptionsButton);
		}

		this.activeNotification = this.gmailApi.Notifications.showNotification({
			el: notificationContent,
			className: 'sd_cust_bar',
			keepCurrentNotification: true,
			canDestroyCallback: () => !this.isActive,
			destroyedCallback: () => {

				if (!this.isDestroyed) {
					this.isDestroyed = true;
					this.processSendAndRemind(buttonId);
				}
			}
		});

		const mouseOver = Kefir.merge([
			Kefir.fromEvents(notificationContent, 'mouseenter').map(() => true),
			Kefir.fromEvents(notificationContent, 'mouseleave').map(() => false)
		]);

		const dropDownActive = Kefir.stream((emitter) => {
			this.dropDownActive = emitter;
			this.dropDownActive.emit(false);
		});

		Kefir.combine([mouseOver, dropDownActive], (bar, dropDown) => bar || dropDown)
			.flatMapLatest((active) => active ? Kefir.constant(active) : Kefir.later(5000, active))
			.takeWhile(() => !this.isDestroyed)
			.onValue((val) => this.isActive = val);

		setTimeout(() => this.setActiveBarValues(buttonId), 10);
	}

	hideSendAndRemindBar(buttonId) {

		if (buttonId === this.activeButtonId) {

			this.isActive = false;
			if (this.activeNotification) {
				this.activeNotification.destroy();
				this.activeNotification = null;
			}

			this.activeButtonId = null;
		}

		if (this.openThread.buttonId === buttonId) {
			this.buttons[buttonId].sent = false;
		} else {
			delete this.buttons[buttonId];
		}
	}

	processSendAndRemind(buttonId) {

		this.saveButtonInfo(buttonId);
		this.hideSendAndRemindBar(buttonId);
	}

	createSortdListItemBarElement(buttonId, eventOrigin) {

		if (!buttonId) {
			return;
		}

		const button = this.buttons[buttonId];
		if (!button) {
			return;
		}

		const bar = document.createElement('div');
		let leftSide = document.createElement('div');
		leftSide.classList.add('sd_fl_lft');
		let buttonEl;

		const remindEl = this.gmailApi.Dom.createSortdButtonElement('Remind me',
			(event) => this.sortdDataService.callIfLoggedIn(this.reminderClickHandler.bind(this), [event, buttonId, eventOrigin]),
			['primary', 'sd_icon', 'sd_fixwdth', 'sd_reminder'],
			{ defaultText: 'Remind me', sdRemindBtn: buttonId });

		buttonEl = this.gmailApi.Dom.createSortdButtonElement('Add to Day plan',
			(event) => this.sortdDataService.callIfLoggedIn(this.datePlanClickHandler.bind(this), [event, buttonId, eventOrigin]),
			['primary', 'sd_icon', 'sd_fixwdth', 'sd_dateplan'],
			{ defaultText: 'Add to Day plan', sdDatePlanBtn: buttonId });
		leftSide.appendChild(buttonEl);

		buttonEl = this.gmailApi.Dom.createSortdButtonElement('Add to board',
			(event) => this.sortdDataService.callIfLoggedIn(this.pinClickHandler.bind(this), [event, buttonId, eventOrigin]),
			['primary', 'sd_icon', 'sd_fixwdth', 'sd_pinto'],
			{ defaultText: 'Add to board', sdPinBtn: buttonId });
		leftSide.appendChild(buttonEl);
		leftSide.appendChild(remindEl);

		const nbsp = String.fromCharCode(160);
		buttonEl = this.gmailApi.Dom.createSortdButtonElement(nbsp,
			(event) => this.sortdDataService.callIfLoggedIn(this.addNextStepClickHandler.bind(this), [event, buttonId, eventOrigin]),
			['sd_icon', 'sd_more'],
			{ defaultText: nbsp });
		buttonEl.id = `sortdMoreButton_${buttonId}`;
		leftSide.appendChild(buttonEl);

		bar.appendChild(leftSide);

		const rightSide = document.createElement('div');
		rightSide.classList.add('sd_fl_rht');

		buttonEl = this.gmailApi.Dom.createSortdButtonElement('More',
			(event) => this.sortdDataService.callIfLoggedIn(this.moreOptionsClickHandler.bind(this), [event, buttonId, eventOrigin]),
			['primary', 'sd_dd'],
			{ sdMoreBtn: buttonId });
		buttonEl.style.display = 'none';
		rightSide.appendChild(buttonEl);
		bar.appendChild(rightSide);

		let div = document.createElement('div');
		div.classList.add('sd_fl_clr');
		bar.appendChild(div);

		leftSide = document.createElement('div');
		leftSide.classList.add('sd_fl_lft');

		div = document.createElement('div');
		div.classList.add('sd_pin_details');
		div.dataset.sdPinDetails = buttonId;
		let span = document.createElement('span');
		span.classList.add('sd_pin_title', 'sd_fl_lft', 'sd_txt_overflow');
		span.addEventListener('click', () => {

			if (button.itemInfo && button.itemInfo.dashboard && button.itemInfo.dashboard.boardType === constants.BOARD_TYPE.SALES) {
				this.sortdDataService.callIfLoggedIn(this.showDealForm.bind(this), [buttonId, eventOrigin, event.target]);
			} else {
				this.sortdDataService.callIfLoggedIn(this.showTaskForm.bind(this), [buttonId, eventOrigin, event.target]);
			}
		});
		div.appendChild(span);

		span = document.createElement('span');
		span.classList.add('sd_ns_title', 'sd_fl_lft', 'sd_txt_overflow');
		div.appendChild(span);
		leftSide.appendChild(div);
		bar.appendChild(leftSide);

		const clear = document.createElement('div');
		clear.classList.add('sd_fl_clr');
		bar.appendChild(clear);

		return bar;
	}

	pinClickHandler(event, buttonId, eventOrigin) {

		if (!buttonId) {
			return;
		}

		// Don't allow changing of pin it this is from Compose from task preview
		const button = this.buttons[buttonId];
		if (!button || button.addThreadToTask) {
			return;
		}

		const isActiveButton = (this.activeButtonId === buttonId);
		
		this.sortdDataService.getListsByDashboard()
			.then((boards) => {

				let result;
				let headingEl;
				let ul;
				let li;
				let activeBoardEl;

				const el = document.createElement('div');
				const positionToElement = event.target;
				const positionToCoord = positionToElement ? null : event.position;
				const hasExistingPinTo = !!(button.itemInfo && button.itemInfo.list && button.itemInfo.list.listId);

				boards.forEach((board) => {

					headingEl = document.createElement('div');
					headingEl.textContent = board.name;
					headingEl.classList.add('sortdHeader');
					el.appendChild(headingEl);
					el.appendChild(this.gmailApi.Dom.createHorzSeparatorElement());

					ul = document.createElement('ul');
					ul.classList.add('sd_with_hdr');
					board.lists.forEach((list) => {

						li = document.createElement('li');
						li.textContent = list.name;
						li.dataset.dashboardId = board.id;
						li.dataset.boardType = board.type;
						li.dataset.listId = list.id;
						if (hasExistingPinTo && button.itemInfo.list.listId === list.id) {
							li.classList.add('selected');
							activeBoardEl = headingEl;
						}
						ul.appendChild(li);
					});
					el.appendChild(ul);
					el.appendChild(this.gmailApi.Dom.createHorzSeparatorElement());
				});

				if (hasExistingPinTo) {

					ul = document.createElement('ul');
					ul.classList.add('sd_with_hdr');
					li = document.createElement('li');
					li.textContent = 'Rename task';
					li.dataset.sortdAction = 'rename';
					li.classList.add('sd_clear_lnk');
					ul.appendChild(li);

					li = document.createElement('li');
					li.textContent = 'Remove from board';
					li.dataset.sortdAction = 'remove';
					li.classList.add('sd_clear_lnk');
					ul.appendChild(li);
					el.appendChild(ul);
					el.appendChild(this.gmailApi.Dom.createHorzSeparatorElement());
				}

				ul = document.createElement('ul');
				ul.classList.add('sd_with_hdr');
				li = document.createElement('li');
				const a = document.createElement('a');
				a.href = this.config.chrome.sortd.customizeListsURL;
				a.textContent = 'Learn how to customize your lists?';
				li.appendChild(a);
				li.classList.add('sd_help_lnk');
				ul.appendChild(li);
				el.appendChild(ul);

				el.addEventListener('click', (evt) => {

					const clickedEl = evt.target;
					if (clickedEl.dataset.dashboardId) {
						const dashboardId = clickedEl.dataset.dashboardId;
						const listId = clickedEl.dataset.listId;
						const boardType = clickedEl.dataset.boardType;

						const listsEl = clickedEl.closest('ul');
						const headerEl = listsEl.previousSibling;

						const dashboard = {
							dashboardId,
							boardType,
							title: headerEl.textContent
						};

						const list = {
							listId,
							title: clickedEl.textContent
						};

						if (result && result.destroy) {
							result.destroy();
						}

						if (hasExistingPinTo || this.displayMode === constants.PANEL_DISPLAY_MODES.FULLSCREEN) {
							this.pinToSelectedList(buttonId, eventOrigin, dashboard, list);
						} else if (positionToElement) {
							if (boardType === constants.BOARD_TYPE.SALES) {
								this.showDealForm(buttonId, eventOrigin, positionToElement, dashboard, list);
							} else {
								this.showTaskForm(buttonId, eventOrigin, positionToElement, dashboard, list);
							}
						}
					} else if (clickedEl.dataset.sortdAction) {

						if (result && result.destroy) {
							result.destroy();
						}

						switch (clickedEl.dataset.sortdAction) {
							case 'remove': 
								this.pinToSelectedList(buttonId, eventOrigin);
								break;
							case 'rename':
								if (positionToElement) {
									if (button.itemInfo && button.itemInfo.dashboard && button.itemInfo.dashboard.boardType === constants.BOARD_TYPE.SALES) {
										this.showDealForm(buttonId, eventOrigin, positionToElement);
									} else {
										this.showTaskForm(buttonId, eventOrigin, positionToElement);
									}
								}
								break;
						}
					}
				});

				if (isActiveButton) {
					this.dropDownActive.emit(true);
				}

				this.lastPinHandlerResult = result = this.gmailApi.Dom.showDropDown({
					el,
					positionToElement,
					positionToCoord,
					centerToElement: isActiveButton,
					destroyedCallback: () => {
						if (isActiveButton) {
							this.dropDownActive.emit(false);
						}

						this.lastPinHandlerResult = null;
					}
				});

				if (activeBoardEl) {
					setTimeout(() => {
						activeBoardEl.scrollIntoView({ behavior: 'smooth' });
					}, 500);
				}
			})
			.catch((ex) => this.handleError(ex))
	}

	/**
	 * Update button properties and update DOM to show selected item
	 * @param {string} buttonId - The id of the button to update
	 * @param {string} eventOrigin - The origin of the update for tracking
	 * @param {Object} [dashboard] - The dashboard to pin to
	 * @param {Object} [list] - The list to pin to
	 * @param {boolean} [save] - Defaults to true.  False not to save immediately
	 * @private
	 */
	pinToSelectedList(buttonId, eventOrigin, dashboard, list, save = true) {

		const button = this.buttons[buttonId];
		if (!button) {
			return;
		}

		let pinButtons
		let moreButtons;
		if (buttonId === this.activeButtonId) {
			pinButtons = document.body.querySelectorAll(`[data-sd-pin-btn="bar"], [data-sd-pin-btn="${buttonId}"]`);
			moreButtons = document.body.querySelectorAll(`[data-sd-more-btn="bar"], [data-sd-more-btn="${buttonId}"]`);
		} else {
			pinButtons = document.body.querySelectorAll(`[data-sd-pin-btn="${buttonId}"]`);
			moreButtons = document.body.querySelectorAll(`[data-sd-more-btn="${buttonId}"]`);
		}

		if (moreButtons && moreButtons.length > 0) {
			moreButtons.forEach((moreButton) => {
				moreButton.style.display = 'none';
			});
		}

		if (list && list.listId) {
			button.itemInfo.dashboard = dashboard;
			button.itemInfo.list = list;
			if (pinButtons && pinButtons.length > 0) {
				pinButtons.forEach((pinButton) => {
					pinButton.classList.add('selected');
					pinButton.textContent = list.title;
				});
			}
			if (moreButtons && moreButtons.length > 0) {
				moreButtons.forEach((moreButton) => {
					moreButton.style.display = 'inline';
				});
			}
		} else {
			button.itemInfo.dashboard = {};
			button.itemInfo.list = {};
			if (pinButtons) {
				pinButtons.forEach((pinButton) => {
					pinButton.classList.remove('selected');
					pinButton.textContent = pinButton.dataset.defaultText;
				});
			}
		}

		if (eventOrigin) {
			button.eventOrigins.pinTo = eventOrigin;
		}

		if (save && button.openedFromThread) {
			this.saveButtonInfo(buttonId);
		}
	}

	datePlanClickHandler(event, buttonId, eventOrigin) {

		if (!buttonId) {
			return;
		}

		const isActiveButton = (this.activeButtonId === buttonId);
		const positionToElement = event.target;
		const button = this.buttons[buttonId];
		if (!button) {
			return;
		}

		let selectedDate = null;
		if (button.itemInfo && button.itemInfo.item && button.itemInfo.item.plannedDate) {
			selectedDate = button.itemInfo.item.plannedDate;
		}

		this.gmailApi.Dom.showDateTimeDropDown({
			dateIds: constants.DATE_PLAN_DATE_IDS,
			dateOnly: true,
			heading: 'Add to My Day plan for...',
			selectedDate,
			centerToElement: isActiveButton,
			positionToElement,
			showSetDueDate: true,
			dropDownActiveStream: isActiveButton && this.dropDownActive,
			inlinePicker: false,
			minDate: 0,
			onDateSelected: (date, setDueDateToPlannedDate) => this.setPlannedDate(buttonId, eventOrigin, date, 0, true, setDueDateToPlannedDate)
		});
	}

	/**
	 * Update button properties and update DOM to show selected planned date
	 * @param {string} buttonId - The id of the button to update
	 * @param {string} eventOrigin - The origin of the update for tracking
	 * @param {Date} [date] - The planned date
	 * @param {number} [dateOrder] - The order on the date plan
	 * @param {boolean} [save] - Defaults to true.  False not to save immediately
	 * @param {boolean} [setDueDateToPlannedDate] - Defaults to false.  True to also update the due date
	 * @private
	 */
	setPlannedDate(buttonId, eventOrigin, date, dateOrder, save = true, setDueDateToPlannedDate = false) {

		const button = this.buttons[buttonId];
		if (!button) {
			return;
		}

		let datePlanButtons;
		if (buttonId === this.activeButtonId) {
			datePlanButtons = document.body.querySelectorAll(`[data-sd-date-plan-btn="bar"], [data-sd-date-plan-btn="${buttonId}"]`);
		} else {
			datePlanButtons = document.body.querySelectorAll(`[data-sd-date-plan-btn="${buttonId}"]`);
		}

		if (date !== '' && date !== undefined && date !== null) {
			date.setHours(0, 0, 0, 0);
			button.itemInfo.item = button.itemInfo.item || {};
			button.itemInfo.item.plannedDate = date;
			button.itemInfo.item.plannedDateOrder = dateOrder;
			button.itemInfo.item.setDueDateToPlannedDate = setDueDateToPlannedDate;
			if (datePlanButtons) {
				datePlanButtons.forEach((datePlanButton) => {
					datePlanButton.classList.add('selected');
					datePlanButton.textContent = SortdCommon.formatShortDate(date);
				});
			}
		} else {
			if (button.itemInfo.item) {
				button.itemInfo.item.plannedDate = null;
				button.itemInfo.item.plannedDateOrder = null;
				button.itemInfo.item.setDueDateToPlannedDate = setDueDateToPlannedDate;
			}
			if (datePlanButtons) {
				datePlanButtons.forEach((datePlanButton) => {
					datePlanButton.classList.remove('selected');
					datePlanButton.textContent = datePlanButton.dataset.defaultText;
				});
			}
		}

		if (eventOrigin) {
			button.eventOrigins.plannedDate = eventOrigin;
		}

		if (save && button.openedFromThread) {
			this.saveButtonInfo(buttonId);
		}
	}

	reminderAndNextStepHandler(event, buttonId, eventOrigin, originNextStep) {

		if (!buttonId) {
			return;
		}

		const isActiveButton = (this.activeButtonId === buttonId);
		const positionToElement = event.target;
		const button = this.buttons[buttonId];
		if (!button) {
			return;
		}

		let selectedDate = null;
		if (button.itemInfo && button.itemInfo.reminderDate) {
			selectedDate = button.itemInfo.reminderDate;
		}
		let currentNextStep = null;
		if (button.itemInfo && button.itemInfo.item && button.itemInfo.item.nextStep) {
			currentNextStep = button.itemInfo.item.nextStep;
		}

		this.gmailApi.Dom.showReminderTimeDropDown({
			dateIds: constants.REMINDER_DATE_IDS,
			dateOnly: false,
			originNextStep,
			heading: originNextStep ? 'Status and next step...' : 'Reminder...',
			selectedDate,
			nextStep: currentNextStep,
			centerToElement: isActiveButton,
			positionToElement,
			inlinePicker: false,
			dropDownActiveStream: isActiveButton && this.dropDownActive,
			onDateSelected: (date, addToDatePlan, nextStep) => {

				this.setReminder(buttonId, eventOrigin, date, false);
				this.setNextStep(buttonId, eventOrigin, nextStep, null, false);

				if (addToDatePlan) {
					let dateOnlyDate = null;
					if (date) {
						dateOnlyDate = new Date(date);
						dateOnlyDate.setHours(0, 0, 0, 0);
					}
					this.setPlannedDate(buttonId, eventOrigin, dateOnlyDate, 0, false, false);
				}

				if (button.openedFromThread) {
					this.saveButtonInfo(buttonId);
				}
			}
		});
	}

	reminderClickHandler(event, buttonId, eventOrigin) {
		this.reminderAndNextStepHandler(event, buttonId, eventOrigin, false);
	}

	/**
	 * Update button properties and update DOM to show selected reminder
	 * @param {string} buttonId - The id of the button to update
	 * @param {string} eventOrigin - The origin of the update for tracking
	 * @param {Object} [date] - The date for the reminder
	 * @param {boolean} [save] - Defaults to true.  False not to save immediately
	 * @private
	 */
	setReminder(buttonId, eventOrigin, date, save = true) {

		const button = this.buttons[buttonId];
		if (!button) {
			return;
		}

		let remindButtons;
		if (buttonId === this.activeButtonId) {
			remindButtons = document.body.querySelectorAll(`[data-sd-remind-btn="bar"], [data-sd-remind-btn="${buttonId}"]`);
		} else {
			remindButtons = document.body.querySelectorAll(`[data-sd-remind-btn="${buttonId}"]`);
		}

		if (date !== '' && date !== undefined && date !== null) {
			button.itemInfo.reminderDate = date;
			if (remindButtons) {
				remindButtons.forEach((remindButton) => {
					remindButton.classList.add('selected');
					remindButton.textContent = SortdCommon.formatShortDateTime(date);
				});
			}
		} else {
			button.itemInfo.reminderDate = null;
			if (remindButtons) {
				remindButtons.forEach((remindButton) => {
					remindButton.classList.remove('selected');
					remindButton.textContent = remindButton.dataset.defaultText;
				});
			}
		}

		if (eventOrigin) {
			button.eventOrigins.reminder = eventOrigin;
		}

		if (save && button.openedFromThread) {
			this.saveButtonInfo(buttonId);
		}
	}

	addNextStepClickHandler(event, buttonId, eventOrigin) {
		this.reminderAndNextStepHandler(event, buttonId, eventOrigin, true);
	}

	toggleSelectedNextStepTime(container, time) {

		container.querySelectorAll('[data-sortd-date]').forEach((dateEl) => dateEl.classList.remove('selected', 'animate'));

		let el = container.querySelector(`[data-sortd-date="${time}"]`);
		if (!el) {
			el = container.querySelector('[data-sortd-date="custom"]');
		}

		el.classList.add('selected');
	}

	/**
	 * Update button properties and update DOM to show selected next step
	 * @param {string} buttonId - The id of the button to update
	 * @param {string} eventOrigin - The origin of the update for tracking
	 * @param {Object} [nextStep] - The next step object to set
	 * @param {string} [reminder] - The reminder date in ISO format
	 * @param {boolean} [save] - Defaults to true.  False not to save immediately
	 * @private
	 */
	setNextStep(buttonId, eventOrigin, nextStep, reminder, save = true) {
		
		const button = this.buttons[buttonId];
		if (!button) {
			return;
		}

		if (buttonId === this.activeButtonId) {
			this.setNextStepForBar(buttonId, nextStep);
		}

		const nextStepButtons = document.body.querySelectorAll(`[data-sd-next-step-btn="${buttonId}"]`);

		if (nextStep) {
			button.itemInfo.item = button.itemInfo.item || {};
			button.itemInfo.item.nextStep = nextStep;
			if (nextStepButtons) {
				nextStepButtons.forEach((nextStepButton) => {
					nextStepButton.classList.add('selected');
					nextStepButton.textContent = 'Edit next step';
				});
			}
		} else {
			if (button.itemInfo.item) {
				button.itemInfo.item.nextStep = null;
			}
			if (nextStepButtons) {
				nextStepButtons.forEach((nextStepButton) => {
					nextStepButton.classList.remove('selected');
					nextStepButton.textContent = nextStepButton.dataset.defaultText;
				});
			}
		}
		this.updateNextStepAndTaskDetailUi(buttonId);

		if (eventOrigin) {
			button.eventOrigins.nextStep = eventOrigin;
		}

		if (reminder) {
			this.setReminder(buttonId, eventOrigin, new Date(reminder), save);
		} else if (save && button.openedFromThread) {
			this.saveButtonInfo(buttonId);
		}
	}
	
	setNextStepForBar(buttonId, nextStep) {
		
		const button = this.buttons[buttonId];
		if (!button) {
			return;
		}

		const nextStepEl = document.body.querySelector('.sd_ns_text_bar');
		if (nextStep) {
			button.itemInfo.item = button.itemInfo.item || {};
			button.itemInfo.item.nextStep = nextStep;
			if (nextStepEl) {
				nextStepEl.value = nextStep;
				nextStepEl.classList.add('selected', 'animate');
				setTimeout(() => nextStepEl.classList.remove('animate'), 1500);
			}
		} else {
			if (button.itemInfo.item) {
				button.itemInfo.item.nextStep = null;
			}
			if (nextStepEl) {
				nextStepEl.value = '';
				nextStepEl.classList.remove('selected', 'animate');
			}
		}
	}

	updateWorkflowStatus(buttonId, eventOrigin, workflowStatusId) {

		const button = this.buttons[buttonId];

		if (!(button && button.itemInfo && button.itemInfo.item && button.itemInfo.list)) {
			return;
		}

		this.messages.sendMessageToFrame('updateTaskWorkflowStatus', {
			taskId: button.itemInfo.item.itemId,
			workflowStatusId,
			eventOrigin
		});

		let moreButtons;
		if (buttonId === this.activeButtonId) {
			moreButtons = document.body.querySelectorAll(`[data-sd-more-btn="bar"], [data-sd-more-btn="${buttonId}"]`);
		} else {
			moreButtons = document.body.querySelectorAll(`[data-sd-more-btn="${buttonId}"]`);
		}

		if (moreButtons && moreButtons.length > 0) {
			moreButtons.forEach((moreButton) => {
				moreButton.style.display = 'none';
			});
		}
	}
		
	moreOptionsClickHandler(event, buttonId, eventOrigin) {

		const isActiveButton = (this.activeButtonId === buttonId);
		const positionToElement = event.target;
		const button = this.buttons[buttonId];

		let result;
		let li;

		const el = document.createElement('div');
		const ul = document.createElement('ul');

		if (button.itemInfo.dashboard && button.itemInfo.dashboard.boardType === constants.BOARD_TYPE.SALES) {
			li = document.createElement('li');
			li.textContent = 'Won';
			li.dataset.sdCompleted = constants.TASK_WORKFLOW_STATUS.WON;
			ul.appendChild(li);

			li = document.createElement('li');
			li.textContent = 'Lost';
			li.dataset.sdCompleted = constants.TASK_WORKFLOW_STATUS.LOST;
			ul.appendChild(li);

			li = document.createElement('li');
			li.textContent = 'Abandoned';
			li.dataset.sdCompleted = constants.TASK_WORKFLOW_STATUS.ABANDONED;
			ul.appendChild(li);
		} else {
			li = document.createElement('li');
			li.textContent = 'Mark complete';
			li.dataset.sdCompleted = constants.TASK_WORKFLOW_STATUS.COMPLETED;
			ul.appendChild(li);
		}

		li = document.createElement('li');
		li.textContent = 'Open in sidebar';
		li.dataset.sdMode = constants.PANEL_DISPLAY_MODES.SIDEBAR;
		ul.appendChild(li);

		li = document.createElement('li');
		li.textContent = 'Show on board';
		li.dataset.sdMode = constants.PANEL_DISPLAY_MODES.FULLSCREEN;
		ul.appendChild(li);
		el.appendChild(ul);

		el.addEventListener('click', (evt) => {

			const clickedEl = evt.target;

			if (clickedEl.dataset.sdCompleted) {
				this.updateWorkflowStatus(buttonId, eventOrigin, clickedEl.dataset.sdCompleted);
			} else if (clickedEl.dataset.sdMode) {
				const action = (clickedEl.dataset.sdMode === constants.PANEL_DISPLAY_MODES.SIDEBAR) ?
					'openSidbarAndSendMessage' :
					'openPanelAndSendMessage';
				
				this.messages.sendMessageToFrame('closeTaskPreview');
				this.messages.sendMessageToExtension(action, {
					action: 'viewTask',
					taskId: button.itemInfo.item.itemId,
					dashboardId: button.itemInfo.dashboard.dashboardId
				});
			}

			if (result && result.destroy) {
				result.destroy();
			}
		});

		if (isActiveButton) {
			this.dropDownActive.emit(true);
		}

		result = this.gmailApi.Dom.showDropDown({
			el,
			positionToElement,
			destroyedCallback: () => {
				if (isActiveButton) {
					this.dropDownActive.emit(false);
				}
			}
		});
	}

	updateNextStepAndTaskDetailUi(buttonId) {

		const button = this.buttons[buttonId];
		if (!button) {
			return;
		}

		const pinDetailsEls = document.querySelectorAll(`[data-sd-pin-details="${buttonId}"]`);
		pinDetailsEls.forEach((pinDetailsEl) => {

			const pinEl = pinDetailsEl.querySelector('.sd_pin_title');
			const nsEl = pinDetailsEl.querySelector('.sd_ns_title');
			pinEl.classList.remove('sd_show_border');
			pinEl.style.display = 'none';
			nsEl.style.display = 'none';

			const pinned = button.itemInfo.list && button.itemInfo.list.listId &&
				button.itemInfo.item && button.itemInfo.item.title;

			const nextStep = button.itemInfo && button.itemInfo.item && button.itemInfo.item.nextStep;

			if (pinned || nextStep) {
				pinDetailsEl.style.display = 'block';
				if (pinned) {
					pinEl.textContent = button.itemInfo.item.title;
					pinEl.style.display = 'block';
				}
				if (nextStep) {
					nsEl.textContent = `Next: ${nextStep}`;
					nsEl.style.display = 'block';
				}
				if (pinned && nextStep) {
					pinEl.classList.add('sd_show_border');
				}
			} else {
				pinDetailsEl.style.display = 'none';
			}
		});
	}

	showTaskForm(buttonId, eventOrigin, positionToElement, dashboard, list) {

		if (!buttonId) {
			return;
		}

		const closeTaskForm = (saveTitle) => {

			if (saveTitle) {
				if (dashboard && list) {
					this.pinToSelectedList(buttonId, eventOrigin, dashboard, list, false);
				}
				this.setTaskDetails(buttonId, eventOrigin, taskName.value || '', null, true);
			}

			if (result && result.destroy) {
				result.destroy();
			}
		}
		
		let result;
		const activeButton = this.buttons[buttonId];
		const isActiveButton = (this.activeButtonId === buttonId);

		const el = document.createElement('div');
		el.classList.add('sd_pin_task_frm');
		const headingEl = document.createElement('div');
		headingEl.textContent = 'Add a title';
		headingEl.classList.add('sortdHeader');
		el.appendChild(headingEl);

		const form = document.createElement('form');
		let div = document.createElement('div');
		div.classList.add('form-group');
		const taskName = document.createElement('textarea');
		taskName.placeholder = 'Type here';
		taskName.textContent = activeButton && activeButton.itemInfo && activeButton.itemInfo && activeButton.itemInfo.item && activeButton.itemInfo.item.title || activeButton.getSubject() || '';
		taskName.addEventListener('keydown', (evt) => {

			if (evt && evt.keyCode === 13) {
				evt.stopPropagation();
				closeTaskForm(true);
			} else if (evt && evt.keyCode === 27) {
				evt.stopPropagation();
				closeTaskForm(false);
			}
		});
		div.appendChild(taskName);
		form.appendChild(div);

		div = document.createElement('div');
		div.classList.add('form-group', 'sd_txt_aln_rht');
		let button = this.gmailApi.Dom.createSortdButtonElement(
			'Cancel',
			() => closeTaskForm(false));
		div.appendChild(button);

		button = this.gmailApi.Dom.createSortdButtonElement(
			'OK',
			() => closeTaskForm(true),
			['primary']);
		div.appendChild(button);
		form.appendChild(div);
		el.appendChild(form);

		if (isActiveButton) {
			this.dropDownActive.emit(true);
		}

		result = this.gmailApi.Dom.showDropDown({
			el,
			positionToElement,
			centerToElement: isActiveButton,
			destroyedCallback: () => {
				if (isActiveButton) {
					this.dropDownActive.emit(false);
				}
			}
		});

		setTimeout(() => {
			taskName.focus();
			taskName.select();
		}, 500);
	}

	showDealForm(buttonId, eventOrigin, positionToElement, dashboard, list) {

		if (!buttonId) {
			return;
		}

		const closeDealForm = (saveTitle) => {

			if (saveTitle) {
				if (dashboard && list) {
					this.pinToSelectedList(buttonId, eventOrigin, dashboard, list, false);
				}
				this.setTaskDetails(buttonId, eventOrigin, dealName.value || '', null, true);
			}

			if (result && result.destroy) {
				result.destroy();
			}
		}
		
		let result;
		const activeButton = this.buttons[buttonId];
		const isActiveButton = (this.activeButtonId === buttonId);

		const el = document.createElement('div');
		el.classList.add('sd_pin_task_frm');
		const headingEl = document.createElement('div');
		headingEl.textContent = 'Add a title';
		headingEl.classList.add('sortdHeader');
		el.appendChild(headingEl);

		const form = document.createElement('form');
		
		let div = document.createElement('div');
		div.classList.add('form-group');
		const dealName = document.createElement('textarea');
		dealName.placeholder = 'Type here';
		dealName.textContent = activeButton && activeButton.itemInfo && activeButton.itemInfo.item && activeButton.itemInfo.item.title || activeButton.getSubject() || '';
		dealName.addEventListener('keydown', (evt) => {

			if (evt && evt.keyCode === 13) {
				evt.stopPropagation();
				closeDealForm(true);
			} else if (evt && evt.keyCode === 27) {
				evt.stopPropagation();
				closeTaskForm(false);
			}
		});

		div.appendChild(dealName);
		form.appendChild(div);

		// subEl = document.createElement('div');
		// subEl.classList.add('form-group');
		// subEl.textContent = 'Value';
		// form.appendChild(subEl);

		// div = document.createElement('div');
		// div.classList.add('form-group');
		// const dealValue = document.createElement('input');
		// dealValue.type = 'number';
		// dealValue.placeholder = '$ value';
		// dealValue.textContent = activeButton && activeButton.itemInfo && activeButton.itemInfo.item && activeButton.itemInfo.item.title || '';
		// div.appendChild(dealValue);
		// form.appendChild(div);

		div = document.createElement('div');
		div.classList.add('form-group', 'sd_txt_aln_rht');
		let button = this.gmailApi.Dom.createSortdButtonElement(
			'Cancel',
			() => closeDealForm(false));
		div.appendChild(button);

		button = this.gmailApi.Dom.createSortdButtonElement(
			'OK',
			() => closeDealForm(true),
			['primary']);
		div.appendChild(button);
		form.appendChild(div);
		el.appendChild(form);

		if (isActiveButton) {
			this.dropDownActive.emit(true);
		} 

		result = this.gmailApi.Dom.showDropDown({
			el,
			positionToElement,
			centerToElement: isActiveButton,
			destroyedCallback: () => {
				if (isActiveButton) {
					this.dropDownActive.emit(false);
				}
			}
		});

		setTimeout(() => {
			dealName.focus();
			dealName.select();
		}, 500);
	}

	/**
	 * Update the title & value for the task or deal
	 * @param {string} buttonId - The id of the button to update
	 * @param {string} eventOrigin - The origin of the update for tracking
	 * @param {string} title - The new task title
	 * @param {number} dealValue - The new deal value
	 * @param {boolean} [save] - Defaults to true.  False not to save immediately
	 */
	setTaskDetails(buttonId, eventOrigin, title, dealValue, save = true) {

		const button = this.buttons[buttonId];
		if (!button) {
			return;
		}

		if (button.itemInfo) {
			button.itemInfo.item = button.itemInfo.item || {};
			button.itemInfo.item.title = title;
			button.itemInfo.item.dealValue = dealValue;

			if (eventOrigin) {
				button.eventOrigins.taskDetails = eventOrigin;
			}

			if (save && button.openedFromThread) {
				this.saveButtonInfo(buttonId);
			}
		}

		this.updateNextStepAndTaskDetailUi(buttonId);
	}

	handleArchiveCompleteClicked(event) {

		event.stopPropagation();

		const threadId = this.openThread.threadId;
		this.clickArchiveButton();

		if (threadId) {
			this.listItems.completeAndArchiveByThreadId(threadId);
		}
	}

	addArchiveCompleteButton() {

		if (!this.openThread.buttonId) {
			return;
		}

		const button = this.buttons[this.openThread.buttonId];
		if (!button) {
			return;
		}

		if (!(button.itemInfo && button.itemInfo.item && button.itemInfo.item.currentWorkflowStatusId > constants.TASK_WORKFLOW_STATUS.COMPLETED)) {
			if (this.openThread.archiveCompleteAdded) {
				const currentButton = document.querySelector('.sd_complete_archive_btn');
				if (currentButton) {
					currentButton.remove();
				}
			}
			return;
		}

		if (this.openThread.archiveCompleteAdded) {
			return;
		}

		this.openThread.archiveCompleteAdded = true;

		const domMapArchive = SortdGmailApiDomMap.VIEW_MAIL[0].ARCHIVE;
		const selectors = domMapArchive && domMapArchive.SELECTORS;
		if (!selectors || !this.openThread.buttonId) {
			return;
		}

		const elements = this.gmailApi.Dom.getElementsFromSelector(selectors);
		if (!elements) {
			return;
		}

		const archiveButton = elements[0].parentElement;
		let completeArchiveBtn = document.createElement('div');
		
		let regVal = `\b${domMapArchive.REPLACE_CLASS}\b`.replace(/[\^\$\.\|\?\*\+\(\)\[\{]/gi, (i) => `\\${i}`);
		let regX = new RegExp(regVal, 'g');

		completeArchiveBtn.innerHTML = archiveButton.outerHTML.replace(regX, 'sd_complete_archive_btn');
		completeArchiveBtn = completeArchiveBtn.firstChild;
		completeArchiveBtn.setAttribute('title', 'Archive & Complete');
		completeArchiveBtn.id = 'sortdCompleteArchiveThread';
		completeArchiveBtn.firstChild.classList.add('sdCompleteArchive');
		const icons = completeArchiveBtn.querySelectorAll(domMapArchive.ICON.SELECTORS);
		if (icons) {
			icons.forEach((icon) => {
				icon.style.backgroundImage = `url(${chrome.extension.getURL('img/date-plan/archive_blue.png')})`;
			});
		}

		completeArchiveBtn.addEventListener('click', (event) => this.handleArchiveCompleteClicked(event));
		
		const parent = archiveButton.parentElement;
		parent.insertBefore(completeArchiveBtn, parent.firstChild.nextSibling);
	}

	registerOverrideElementCallback(dateOverride, selectedDate, resultCallback, captureMouseClick, validationFn, elements) {

		elements.forEach((el) => {

			el.parentElement.classList.add('sortd_dateplan_override');

			const captureFn = captureMouseClick ? this.gmailApi.Dom.captureMouseClickEvent : this.gmailApi.Dom.captureMouseUpEvent;
			captureFn.call(this.gmailApi.Dom, el, (evt, propagateOriginalEvtCallback) => {

				let message;
				let propagateOriginalEvtOnValue = true;
				let propagateOriginalEvtOnCancel = false;

				if (dateOverride === constants.DATE_OVERRIDES.MARK_AS_STAR) {
					setTimeout(propagateOriginalEvtCallback, 10);
					propagateOriginalEvtOnValue = false;
					message = 'Email starred.';
				} else if (dateOverride === constants.DATE_OVERRIDES.MARK_AS_UNREAD) {
					propagateOriginalEvtOnCancel = true;
					message = 'Email marked as unread';
				}

				if (message) {
					// This needs to be in a timeout as we'll be clicking the button just now
					// and a document.click hides the notification again
					setTimeout(() => {
						this.messages.sendMessageToExtension('displayInformationNotification', { message });
					}, 100);
				}

				this.gmailApi.Dom.showActionOverrideDateTimeDropDown({
					dateIds: constants.DATE_PLAN_DATE_IDS,
					dateOnly: true,
					selectedDate,
					inlinePicker: false,
					minDate: 0,
					positionToElement: evt.target,
					onDateSelected: (date) => {

						if (propagateOriginalEvtOnValue) {
							propagateOriginalEvtCallback();
							propagateOriginalEvtOnValue = false;
							propagateOriginalEvtOnCancel = false;
						}

						resultCallback(date);
					},
					onClose: (selectedClose) => {

						if ((selectedClose && propagateOriginalEvtOnValue) ||
							(!selectedClose && propagateOriginalEvtOnCancel)
						) {
							propagateOriginalEvtCallback();
							propagateOriginalEvtOnValue = false;
							propagateOriginalEvtOnCancel = false;
						}
					}
				});
			}, validationFn);
		});
	}

	registerDatePlanOverridesForSelectors(dateOverride, eventOrigin, closeMailUpdate, selectors, baseElement, captureMouseClick, validationFn) {

		if (!selectors || !this.openThread.buttonId) {
			return;
		}

		const button = this.buttons[this.openThread.buttonId];
		if (!button) {
			return;
		}

		const elements = this.gmailApi.Dom.getElementsFromSelector(selectors, baseElement);
		if (!elements) {
			return;
		}

		let selectedDate = null;
		if (button.itemInfo && button.itemInfo.item && button.itemInfo.item.plannedDate) {
			selectedDate = button.itemInfo.item.plannedDate;
		}

		this.registerOverrideElementCallback(dateOverride, selectedDate, (date) => {

			if (date) {
				date.setHours(0, 0, 0, 0);
			}

			this.setPlannedDate(button.buttonId, eventOrigin, date, 0);
			if (closeMailUpdate) {
				this.inboxSdk.Router.goto(this.inboxSdk.Router.NativeRouteIDs.INBOX);
			}
		}, captureMouseClick, validationFn, elements);
	}

	registerDatePlanOverridesMarkAsUnreadForSelectors(selectors, baseElement, captureMouseClick, validationFn) {

		this.registerDatePlanOverridesForSelectors(
			constants.DATE_OVERRIDES.MARK_AS_UNREAD,
			constants.TRACKER.EVENT_ORIGINS.MARK_AS_UNREAD,
			true,
			selectors,
			baseElement,
			captureMouseClick,
			validationFn);
	}

	registerDatePlanOverridesStarForSelectors(selectors, baseElement, captureMouseClick, validationFn) {

		this.registerDatePlanOverridesForSelectors(
			constants.DATE_OVERRIDES.MARK_AS_STAR,
			constants.TRACKER.EVENT_ORIGINS.MARK_AS_STAR,
			false,
			selectors,
			baseElement,
			captureMouseClick,
			validationFn);
	}

	registerMarkAsUnreadOverride() {

		const selectors = SortdGmailApiDomMap.VIEW_MAIL[0].MARK_AS_UNREAD && SortdGmailApiDomMap.VIEW_MAIL[0].MARK_AS_UNREAD.SELECTORS;
		this.registerDatePlanOverridesMarkAsUnreadForSelectors(selectors, null, false);
	}

	registerStarOverridesForThread() {

		if (!this.openThread.bodyEl) {
			return;
		}

		const selectors = SortdGmailApiDomMap.VIEW_MAIL[0].MARK_AS_STAR && SortdGmailApiDomMap.VIEW_MAIL[0].MARK_AS_STAR.SELECTORS;
		this.registerDatePlanOverridesStarForSelectors(selectors, this.openThread.bodyEl, true, (evt) => {
			const isChecked = evt.target.parentElement.getAttribute('aria-checked');
			return isChecked === 'false' || isChecked === false;
		});
	}

	registerStarOverride() {

		if (!(this.openThread.buttonId && this.openThread.bodyEl)) {
			return;
		}

		const mutationSub = this.gmailApi.Dom
			.observeMutations(this.openThread.bodyEl, { childList: true, subtree: true })
			.filter((mutations) => mutations.some((m) => m.addedNodes.length > 0))
			.observe((mutations) => {
				if (mutations && mutations.length > 0) {
					this.registerStarOverridesForThread();
				}
			});

		this.registerStarOverridesForThread();

		return () => mutationSub.unsubscribe();
	}

	registerMainMenuMarkAsUnreadOverrides(popupEl) {

		const selectors = SortdGmailApiDomMap.VIEW_MAIL[0].MORE_MENU.POP_UP.MARK_AS_UNREAD.SELECTORS;
		this.registerDatePlanOverridesMarkAsUnreadForSelectors(selectors, popupEl, false);
	}

	registerMainMenuMarkAsStarOverrides(popupEl) {

		const selectors = SortdGmailApiDomMap.VIEW_MAIL[0].MORE_MENU.POP_UP.MARK_AS_STAR.SELECTORS;
		this.registerDatePlanOverridesStarForSelectors(selectors, popupEl, false);
	}

	registerMessageMenuMarkAsUnreadOverrides(popupEl) {

		const selectors = SortdGmailApiDomMap.VIEW_MAIL[0].MESSAGE_MORE_MENU.POP_UP.MARK_AS_UNREAD.SELECTORS;
		this.registerDatePlanOverridesMarkAsUnreadForSelectors(selectors, popupEl, false);
	}

	registerMainMoreMenuOverrides() {

		if (!this.openThread.buttonId) {
			return;
		}

		const domViewMail = SortdGmailApiDomMap.VIEW_MAIL[0];
		const selectors = domViewMail.MORE_MENU && domViewMail.MORE_MENU.BUTTON && domViewMail.MORE_MENU.BUTTON.SELECTORS;
		let moreButton = this.gmailApi.Dom.getElementsFromSelector(selectors);
		moreButton = moreButton && moreButton[0];
		if (!moreButton) {
			return;
		}

		const mutationSub = this.gmailApi.Dom
			.observeMutations(moreButton, { attributes: true, attributeFilter: ['aria-expanded'] })
			.observe((mutations) => {

				if (mutations && mutations.length > 0) {

					if (!this.openThread.buttonId) {
						return;
					}
			
					const button = this.buttons[this.openThread.buttonId];
					if (!button) {
						return;
					}

					const selectors = domViewMail.MORE_MENU.POP_UP.SELECTORS;
					let morePopup = this.gmailApi.Dom.getElementsFromSelector(selectors);
					morePopup = morePopup && morePopup[0];

					if (morePopup) {
						this.registerMainMenuMarkAsUnreadOverrides(morePopup);
						this.registerMainMenuMarkAsStarOverrides(morePopup);
					}
				}
			});

		return () => mutationSub.unsubscribe();
	}

	registerMessageMoreMenuOverrides() {

		if (!(this.openThread.buttonId && this.openThread.bodyEl)) {
			return;
		}

		const domViewMail = SortdGmailApiDomMap.VIEW_MAIL[0];
		const selectors = domViewMail.MESSAGE_MORE_MENU && domViewMail.MESSAGE_MORE_MENU.BUTTON && domViewMail.MESSAGE_MORE_MENU.BUTTON.SELECTORS;
		let moreButton = this.gmailApi.Dom.getElementsFromSelector(selectors, this.openThread.bodyEl);
		moreButton = moreButton && moreButton[0];
		if (!moreButton) {
			return;
		}

		const mutationSub = this.gmailApi.Dom
			.observeMutations(moreButton, { attributes: true, attributeFilter: ['aria-expanded'] })
			.observe((mutations) => {

				if (mutations && mutations.length > 0) {

					if (!(this.openThread.buttonId && this.openThread.bodyEl)) {
						return;
					}
			
					const button = this.buttons[this.openThread.buttonId];
					if (!button) {
						return;
					}

					const selectors = domViewMail.MESSAGE_MORE_MENU.POP_UP.SELECTORS;
					let morePopup = this.gmailApi.Dom.getElementsFromSelector(selectors);
					morePopup = morePopup && morePopup[0];

					if (morePopup) {
						this.registerMessageMenuMarkAsUnreadOverrides(morePopup);
					}
				}
			});

		return () => mutationSub.unsubscribe();
	}
	
	/**
	 * Save the info for the button by sending it to Sortd
	 * @param {string} buttonId - The id of the button
	 */
	saveButtonInfo(buttonId) {

		try {
			if (buttonId) {
				const button = this.buttons[buttonId];

				if (!button || button.cancelled) {
					return;
				}

				if (button.threadId && button.itemInfo) {

					const itemInfo = button.itemInfo;
					const currentInfo = button.currentItemInfo;
					const currentItemId = currentInfo.item && currentInfo.item.itemId;
					const itemId = currentItemId || ObjectId.generate();

					if (!itemInfo.item) {
						itemInfo.item = {};
					}

					if (!itemInfo.item.itemId) {
						itemInfo.item.itemId = itemId;
					}

					const requestData = {
						eventOrigins: button.eventOrigins,
						messageId: button.messageId,
						threadId: button.threadId,
						pinTo: {
							itemId,
							title: itemInfo.item.title || button.getSubject() || '(no subject)',
							listId: (itemInfo.list && itemInfo.list.listId) || null,
							dashboardId: (itemInfo.dashboard && itemInfo.dashboard.dashboardId) || null,
							addThreadToTask: button.addThreadToTask || false,
							mailContentItem: button.mailContentItem || null,
							dealValue: itemInfo.item.dealValue || null,
							nextStep: itemInfo.item.nextStep || null,
							plannedDate: itemInfo.item.plannedDate || null,
							plannedDateOrder: itemInfo.item.plannedDateOrder || 0,
							setDueDateToPlannedDate: itemInfo.item.setDueDateToPlannedDate || false
						},
						reminder: itemInfo.reminderDate || null,
						currentReminder: currentInfo.reminderDate || null,
						currentReminderId: currentInfo.reminderId || null
					};

					const listId = currentInfo.list && currentInfo.list.listId;
					const dashboardId = currentInfo.dashboard && currentInfo.dashboard.dashboardId;
					if (currentItemId) {
						requestData.currentPinTo = {
							itemId: currentInfo.item.itemId,
							listId,
							dashboardId,
							title: currentInfo.item && currentInfo.item.title || null,
							dealValue: (currentInfo.item && currentInfo.item.dealValue) || null,
							nextStep: currentInfo.item && currentInfo.item.nextStep || null,
							plannedDate: currentInfo.item && currentInfo.item.plannedDate || null,
							plannedDateOrder: currentInfo.item && currentInfo.item.plannedDateOrder || 0
						};
					}

					this.messages.sendMessageToFrame('updateTaskFromGmailMode', requestData, (eventData) => {

						if (eventData && eventData.errorStatus) {

							this.activeNotification = this.gmailApi.Notifications.showNotification({
								text: itemId ? 'Could not update details in Sortd' : 'Could not add details to Sortd'
							});
						} else {
							// Update the current button to be the same as the data that was posted 
							button.currentItemInfo.dashboard = Object.assign({}, button.itemInfo.dashboard);
							button.currentItemInfo.list = Object.assign({}, button.itemInfo.list);
							button.currentItemInfo.item = Object.assign({}, button.itemInfo.item);
							button.currentItemInfo.reminderDate = button.itemInfo.reminderDate;

							this.listenForListItemChanges(buttonId);
						}
					});
				}
			}

		} catch (ex) {
			this.handleError(ex);
		}
	}

	clickButtonForSelectors(selectors, baseElement) {

		if (!selectors) {
			return;
		}

		selectors = selectors.map((selector) => selector.replace(':not(.sortd_dateplan_override)', '.sortd_dateplan_override'));

		const elements = this.gmailApi.Dom.getElementsFromSelector(selectors, baseElement);
		if (!elements || elements.length === 0) {
			return;
		}

		SortdCommon.simulateRawClick($(elements[0]));
	}

	clickArchiveButton() {
		const selectors = SortdGmailApiDomMap.VIEW_MAIL[0].ARCHIVE && SortdGmailApiDomMap.VIEW_MAIL[0].ARCHIVE.SELECTORS;
		this.clickButtonForSelectors(selectors);
	}

	clickMarkAsUnreadButton() {
		const selectors = SortdGmailApiDomMap.VIEW_MAIL[0].MARK_AS_UNREAD && SortdGmailApiDomMap.VIEW_MAIL[0].MARK_AS_UNREAD.SELECTORS;
		this.clickButtonForSelectors(selectors);
	}

	clickMarkAsStar() {
		const selectors = SortdGmailApiDomMap.VIEW_MAIL[0].MARK_AS_STAR && SortdGmailApiDomMap.VIEW_MAIL[0].MARK_AS_STAR.SELECTORS;
		this.clickButtonForSelectors(selectors);
	}
}
