﻿/**
 * Sortd
 * Copyright (c) 2019, Sortd Inc
 * https://www.sortd.com
 *
 * @fileoverview The stuff going through the inbox-sdk
 * @name sortd-inboxsdk
 */

var SortdInboxSdk = (function () {

	'use strict';

	const mailPreviewRegex = /#inbox\/[a-zA-Z0-9]*$/;

	/**
	 * Initialize the class
	 * @class SortdInboxSdk
	 */
	function SortdInboxSdk(options) {

		this.__FILE_NAME = 'sortd-inboxsdk.js';

		let myOverrides = {};

		if (typeof overrides !== 'undefined') {
			myOverrides = overrides;
		}

		this._config = new config(myOverrides);
		this._sdk = null;
		this._callbacksRegistered = false;
		this._dayPlanLinks = {
			shouldRegister: false,
			registered: false
		};

		this.isLoggedIn = false;
		this.hashedUser = null;
		this.listView = null;
		this.errorHandler = options.errorHandler;
		this.tracker = options.tracker;
		this.gmailApi = options.gmailApi;
		this.messages = options.messages;
		this.threadAddedToList = {
			emitter: null,
			sendMessage: true
		};
		this.showDayPlanLinksEnabled = null;
		this.composeData = null;
		this.sortdDataService = SortdDataService.getInstance({ messages: this.messages });
		this.shortcutHelper = SortdShortcutHelper.getInstance({
			config: this._config,
			messages: this.messages,
			errorHandler: this.errorHandler
		});

		this.changePreviewToAllMail = false;

		Kefir.stream((emitter) => this.threadAddedToList.emitter = emitter)
			.debounce(500)
			.onValue(() => this.checkSendNewThreadNotification());

		this.messages.addListener('snoozeMailThreadComplete', this.snoozeComplete.bind(this));
		this.messages.addListener('isLoggedIn', this.handleUserLogin.bind(this));
		this.messages.addListener('showDayPlanLinks', this.handleShowDayPlanLinks.bind(this));
		this.messages.addListener('setSortdUser', this.handleSetSortdUserReceived.bind(this));
		this.messages.addListener('openMailPreview', this.handleOpenMailPreview.bind(this));
		this.messages.addListener('isConversationViewDisabled', this.handleIsConversationViewDisabled.bind(this));
		this.messages.addListener('composeMail', this.handleComposeMail.bind(this));
		this.messages.addListener('getAccountSwitcherContacts', this.handleGetAccountSwitcherContacts.bind(this));

		this.shortcutHelper.addListener(
			constants.SHORTCUTS.KEY_CODES.CTRL + '_' + constants.SHORTCUTS.KEY_CODES.SPACE,
			this.handleShowDatePlan.bind(this));

		window.onhashchange = (event) => {

			if (this.changePreviewToAllMail) {
				
				if (mailPreviewRegex.test(event.newURL)) {
					window.location.hash = window.location.hash.replace('#inbox/', '#all/');
				}

				this.changePreviewToAllMail = false;
			}
		};
	}

	SortdInboxSdk.prototype.handleError = function (error) {

		if (this.errorHandler) {
			this.errorHandler.handle(error, this.__FILE_NAME);
		}
	};

	SortdInboxSdk.prototype.postMessage = function (message, data) {
		this.messages.sendMessageToExtension(message, data);
	};

	SortdInboxSdk.prototype.getAccountSwitcherContactList = function() {

		if (this._sdk) {

			// Remove the text "(default)" in various languages from the end of the address
			// but don't replace (***) inside the address
			return this._sdk.User.getAccountSwitcherContactList()
				.filter((c) => c.emailAddress && c.emailAddress.indexOf('@') !== -1)
				.map((c) => ({
					name: c.name,
					emailAddress: c.emailAddress.replace(/\(([^\)]*)\)[ \t]*$/, '').trim().toLowerCase()
				}));
		}
	}

	SortdInboxSdk.prototype.handleUserLogin = function (message) {

		if (message && message.data) {
			this.isLoggedIn = message.data.loggedIn || false;
		}
	};

	SortdInboxSdk.prototype.handleShowDayPlanLinks = function (message) {

		if (this.showDayPlanLinksEnabled === null) {
			this.showDayPlanLinksEnabled = message && message.data && message.data.enabled || false;
			if (this.showDayPlanLinksEnabled) {

				this._dayPlanLinks.shouldRegister = true;
				if (this._sdk) {
					this.registerDayPlanLinks();
				}
			}
		}
	};

	SortdInboxSdk.prototype.handleSetSortdUserReceived = function (message) {

		try {
			if (message && message.data && message.data.hashedUser) {
				this.hashedUser = message.data.hashedUser;
			}
		} catch (ex) {
			this.handleError(ex);
		}
	};

	SortdInboxSdk.prototype.handleOpenMailPreview = function (message) {

		if (this._sdk && message && message.data && message.data.hexThreadId) {
			this._sdk.Router.goto(this._sdk.Router.NativeRouteIDs.THREAD, { threadID: message.data.hexThreadId })
				.then(() => {
					if (message.data.isFullscreen) {
						this.changePreviewToAllMail = true;
					}
				})
				.catch((ex) => {
					this.messages.sendMessageToFrame('closeMail', {
						gmailErrorMessage: 'The linked email could not be opened.  It may have been deleted from Gmail.',
						originalError: ex
					});
				});
		}
	};

	SortdInboxSdk.prototype.handleIsConversationViewDisabled = function (message, callback) {

		if (callback) {

			this._waitForSDK((sdk) => {

				try {
					const isConversationViewDisabled = sdk.User.isConversationViewDisabled();
					// Docs say it will return a boolean, but it was returning a Promise, so making provision for both
					if (typeof isConversationViewDisabled === 'boolean') {
						callback(null, { isConversationViewDisabled });
					} else if (isConversationViewDisabled && typeof isConversationViewDisabled.then === 'function') {
						isConversationViewDisabled
							.then((result) => {
								callback(null, { isConversationViewDisabled: result });
							})
							.catch((err) => {
								this.handleError(err);
								callback(err);
							})
					}
				} catch (ex) {
					this.handleError(ex);
					callback(ex);
				}
			});
		}
	};

	SortdInboxSdk.prototype.handleComposeMail = function (message) {

		if (this._sdk) {

			if (message && message.data) {
				this.composeData = message.data;
			}

			this._sdk.Compose.openNewComposeView()
				.catch((err) => this.handleError(err));
		}
	};

	SortdInboxSdk.prototype.handleShowDatePlan = function () {
		this.messages.sendMessageToExtension('openPanelAndSendMessage', { action: 'showWeekPlan' });
	};

	SortdInboxSdk.prototype._waitForSDK = function (next) {

		if (this._sdk) {
			next(this._sdk);
		} else {
			setTimeout(() => this._waitForSDK(next), 1000);
		}
	};

	SortdInboxSdk.prototype.handleGetAccountSwitcherContacts = function(message, next) {
		next(null, this.getAccountSwitcherContactList());
	};

	SortdInboxSdk.prototype.postThreadMessage = function (message, thread, target, customTime) {

		thread.getThreadIDAsync().then((threadId) => {

			switch (message) {
				case 'addMailThreadToSortdList':

					this.messages.sendMessageToFrame('updateTaskFromGmailMode', {
						threadId,
						pinTo: {
							listId: target.getAttribute('listId'),
							followUp: false,
							title: thread.getSubject(),
						},
						eventOrigins: { pinTo: constants.TRACKER.EVENT_ORIGINS.PIN_TO_SORTD }

					}, (data) => {
						if (!(data && data.errorStatus)) {
							this.gmailApi.Notifications.showNotification({
								text: 'Your message(s) have been added successfully'
							});
						}
					});
					break;
				case 'setReminder':

					const reminder = customTime ? customTime : SortdCommon.createReminderFromTimeLabel(target.getAttribute('reminder'));
					this.messages.sendMessageToFrame('updateTaskFromGmailMode', {
						threadId,
						pinTo: {
							title: thread.getSubject(),
							followUp: false
						},
						reminder,
						eventOrigins: { reminder: constants.TRACKER.EVENT_ORIGINS.GMAIL_TOOLBAR }
					}, (data) => {
						if (!(data && data.errorStatus)) {
							this.gmailApi.Notifications.showNotification({
								text: 'Your reminder(s) have been added successfully'
							});
						}
					});
					break;
			}
		});
	};

	/**
	 * This callback is used by the {@link SortdInboxSdk#showDatePicker} method.
	 * @callback SortdInboxSdk~showDatePicker
	 * @param {Error} - The error, if any, null if no error occurred.
	 * @param {Date} - The selected date, null if not selected
	 */
	/**
	 * Show the datepicker
	 * @param {JQueryElement} element - The element to attach the picker to
	 * @param {Function} next - The result callback
	 */
	SortdInboxSdk.prototype.showDatePicker = function (element, next) {

		var picker,
			defaultText = element.find('.default-text');

		// If the click target is not the same as the date picker element the picker closes immediately
		if (defaultText && defaultText.length > 0) {
			element = defaultText;
		}

		var bodyRect = document.body.getBoundingClientRect(),
			elemRect = element[0].getBoundingClientRect();

		// If we're closer to the bottom than the picker's height, then align to bottom instead of top
		if (bodyRect.bottom - elemRect.bottom < 330) {
			element.addClass('align-bottom');
		}

		const today = new Date();
		today.setHours(0, 0, 0, 0);

		var options = {
			dateOnly: false,
			container: element,
			closeOnSelect: false,
			min: today,
			max: 200,
			onClose: function () {

				try {
					var selected = this.get('select'),
						highlighted = this.get('highlight');

					if (this.get('cancelled')) {
						next(null, null);
					} else {
						if (this.component && this.component.item && this.component.item.clear === true) {
							next(null, null);
						} else if (selected && highlighted && highlighted.pick) {
							var date = new Date(highlighted.pick);
							if (options.dateOnly) {
								date.setHours(0, 0, 0, 0);
							}
							next(null, date);
						} else {
							next(null, null);
						}
					}

					if (this.get('start')) {
						this.stop();
					}
				} catch (ex) {
					next(ex);
				}
			}
		};

		element.pickadate(options);
		picker = element.pickadate('picker');
		picker.stop(true).start().set('select', Date.now()).open();
	};

	SortdInboxSdk.prototype.addMultipleThreadRowViewClickCallback = function (dropDownEvent, listItem, message, customTime) {

		let rowIndex;

		for (rowIndex = 0; rowIndex < dropDownEvent.selectedThreadViews.length; rowIndex++) {
			const threadView = dropDownEvent.selectedThreadViews[rowIndex];
			this.postThreadMessage(message, threadView, listItem, customTime);
		}

		for (rowIndex = 0; rowIndex < dropDownEvent.selectedThreadRowViews.length; rowIndex++) {
			const threadRowView = dropDownEvent.selectedThreadRowViews[rowIndex];
			this.postThreadMessage(message, threadRowView, listItem, customTime);
		}

		if (dropDownEvent.selectedThreadViews.length > 0) {
			// close mail
			const routeView = (this.Router && this.Router.getCurrentRouteView());
			if (routeView && routeView.getRouteType() === 'THREAD' && message === 'snoozeMailThread') {
				//TODO: this is not perfect, it goes to inbox even if you opened the mail from another folder/search
				this.Router.goto(this.Router.NativeRouteIDs.INBOX);
			}
		}

		dropDownEvent.dropdown.close();
	};

	//click handler in selected items in list
	SortdInboxSdk.prototype.addMultipleThreadRowViewClick = function (self, dropDownEvent, listItem, message) {

		listItem.addEventListener('click', function (evt) {

			if (listItem.getAttribute('reminder') === 'datepicker') {
				self.showDatePicker($(listItem), function (err, date) {
					if (!err && date) {
						self.addMultipleThreadRowViewClickCallback(dropDownEvent, listItem, message, date);
					}
				});
			} else {
				self.addMultipleThreadRowViewClickCallback(dropDownEvent, listItem, message);
			}
		});
	};

	SortdInboxSdk.prototype.showDropdown = function (addSubClick, message, htmlOptions, event) {

		var self = this,
			index,
			lis;

		if (self.isLoggedIn) {

			switch (message) {
				case 'addMailThreadToSortdList':
					self.sortdDataService.getListsByDashboard().then((sortdBoards) => {

						let menuItems = '';

						sortdBoards.forEach((board) => {

							menuItems += '<li class="heading">' + SortdCommon.escapeHtml(board.name) + '</li><li class="sordLists"><ul>';
							board.lists.forEach((list) => {
								menuItems += '<li title="' + SortdCommon.escapeHtml(list.name) + '" listId="' + SortdCommon.escapeHtml(list.id) + '">' + SortdCommon.escapeHtml(list.name) + '</li>';
							});
							menuItems += '</ul></li>';
						});

						event.dropdown.el.innerHTML = htmlOptions
							.replace('{LIST_OPTIONS}', menuItems)
							.replace('TOOL_TIP_ICON_URL', chrome.extension.getURL('img/loader_i.png'));

						var lis = event.dropdown.el.querySelectorAll('ul li ul li');
						var index;
						for (index = 0; index < lis.length; index++) {
							addSubClick(self, event, lis[index], message);
						}
					});
					break;
				default:

					event.dropdown.el.innerHTML = htmlOptions;
					lis = event.dropdown.el.querySelectorAll('li');

					for (index = 0; index < lis.length; index++) {
						// If there is a data-subscriptionfeature on the li check if it is enabled
						// before adding a click handler to it, otherwise add disabled class
						var feature = lis[index].dataset.subscriptionfeature;
						if (feature) {
							self.messages.sendMessageToFrame('checkSubscriptionFeatureEnabled',
								{ feature: feature },
								function (data) {

									if (data && data.enabled === false) {
										self.classList.add('disabled');
									} else {
										addSubClick(self, event, this, message);
									}
								}.bind(lis[index]));
						} else {
							addSubClick(self, event, lis[index], message);
						}
					}
			}

		} else {
			event.dropdown.el.innerHTML = constants.HTML.USER_NOT_LOGGED_IN;
			event.dropdown.el.addEventListener('click', function (evt) {
				event.dropdown.close();
			});

			event.dropdown.el.querySelector('#not-loggedin-notification-signin').addEventListener('click', function (evt) {
				self.postMessage('showLogin', { fullsreen: false });
				event.dropdown.close();
			});

		}
	};

	SortdInboxSdk.prototype.getContactsFromMessageView = function (messageView) {

		var contacts,
			contact,
			exists,
			i;

		contacts = [messageView.getSender()];

		return messageView.getRecipientsFull()
			.then(function (recipients) {
				for (i = 0; i < recipients.length; i++) {
					contact = recipients[i];
					exists = contacts.find(function (c) {
						return c.emailAddress === contact.emailAddress;
					});
					if (!exists) {
						contacts.push(contact);
					}
				}

				return contacts;
			});
	};

	/**
	 * Add all contacts from all messages in an array, but ensure that the sender of the currently opened message
	 * is the first contact in the array, and the recipients of the first message are after that, followed by other messages
	 * In the process also don't add duplicate contacts
	 */
	SortdInboxSdk.prototype.getContactsFromMessageViews = function (messageViews, MessageViewViewStates) {

		var i,
			contacts = [],
			promises = [];

		for (i = 0; i < messageViews.length; i++) {
			promises.push(this.getContactsFromMessageView(messageViews[i]));
		}

		return Promise.all(promises).then(function (results) {

			var merged = [].concat.apply([], results);
			merged.forEach(function (contact) {
				if (!contacts.includes(contact)) {
					contacts.push(contact);
				}
			});

			return contacts;
		});
	};

	SortdInboxSdk.prototype.handleOpenedThreadView = function (threadView, MessageViewViewStates) {

		threadView.getThreadIDAsync()
			.then((threadId) => {

				const messageViews = threadView.getMessageViews();

				Promise.all([
					this.getContactsFromMessageViews(messageViews, MessageViewViewStates),
					messageViews[0].getMessageIDAsync()
				])
					.then((result) => {

						const contacts = result[0];
						const messageId = result[1];

						const viewTaskParams = this.linkInterceptor.getViewTaskLinkParams(messageViews[0].getBodyElement());

						this.postMessage('gmailThreadOpened', {
							messageId,
							threadId,
							taskId: viewTaskParams && viewTaskParams.taskId,
							dashboardId: viewTaskParams && viewTaskParams.dashboardId,
							notificationType: viewTaskParams && viewTaskParams.notificationType,
							threadSubject: threadView.getSubject(),
							contacts
						});
					});

				threadView.on('destroy', () => {
					this.postMessage('gmailThreadClosed');
					this.newThreadNotification(false);
				});
			});
	};

	SortdInboxSdk.prototype.handleOpenedComposeView = function (composeView) {

		this.postMessage('gmailComposeOpened', {
			threadId: composeView.getThreadID(),
			subject: composeView.getSubject()
		});
	};

	SortdInboxSdk.prototype.init = function () {

		return InboxSDK.load('1.0', 'sdk_sortd4gmail_f0a3ebaa3e').then((sdk) => {

			this.listItems = new SortdListItems({
				messages: this.messages,
				errorHandler: this.errorHandler,
				gmailApi: this.gmailApi
			});

			this.readReceipts = new SortdReadReceipts({
				config: this._config,
				messages: this.messages,
				errorHandler: this.errorHandler,
				hashedUser: this.hashedUser,
				tracker: this.tracker
			});

			this.gmailSendService = new SortdGmailSendService({
				config: this._config,
				messages: this.messages,
				errorHandler: this.errorHandler,
				gmailApi: this.gmailApi,
				listItems: this.listItems,
				inboxSdk: sdk
			});

			this.linkInterceptor = new SortdLinkInterceptor({
				config: this._config,
				messages: this.messages,
				errorHandler: this.errorHandler,
				inboxSdk: sdk
			});

			this._sdk = sdk;
			const linkedEmails = this.getAccountSwitcherContactList().map((c) => c.emailAddress);
			this.tracker.updateEmailAddresses(sdk.User.getEmailAddress(), linkedEmails);

			this.registerCallbacks();

			if (this._dayPlanLinks.shouldRegister && !this._dayPlanLinks.registered) {
				this.registerDayPlanLinks();
			}

			return sdk;
		});
	};

	SortdInboxSdk.prototype.registerCallbacks = function () {

		var self = this;

		if (self._callbacksRegistered) {
			return;
		}

		if (!self._sdk) {
			return;
		}

		self._sdk.Router.handleListRoute(self._sdk.Router.NativeListRouteIDs.ANY_LIST, function (routeView) {
			self.listView = routeView;
		});

		self.Router = self._sdk.Router;

		self._sdk.Toolbars.registerThreadButton({
			title: 'Remind me',
			iconUrl: chrome.extension.getURL('img/send-and-pin/bell_ringing_grey_multiselect.png'),
			positions: ['LIST'],
			threadSection: 'INBOX_STATE',
			listSection: 'INBOX_STATE',
			onClick: self.showDropdown.bind(self, self.addMultipleThreadRowViewClick, 'setReminder', constants.HTML.REMINDER_OPTIONS),
			hasDropdown: true
		});

		self._sdk.Toolbars.registerThreadButton({
			title: 'Add to board',
			iconUrl: chrome.extension.getURL('img/send-and-pin/board_grey.png'),
			positions: ['LIST'],
			threadSection: 'INBOX_STATE',
			listSection: 'INBOX_STATE',
			onClick: self.showDropdown.bind(self, self.addMultipleThreadRowViewClick, 'addMailThreadToSortdList', constants.HTML.PIN_OPTIONS),
			hasDropdown: true
		});

		self._sdk.Toolbars.registerThreadButton({
			title: 'Archive & Complete',
			iconUrl: chrome.extension.getURL('img/date-plan/archive_blue.png'),
			positions: ['LIST'],
			threadSection: 'INBOX_STATE',
			listSection: 'INBOX_STATE',
			onClick: self.listItems.handleArchiveCompleteClicked.bind(self.listItems),
			hasDropdown: false
		});

		self._sdk.Toolbars.addToolbarButtonForApp({
			onClick: self.showAppDropDown.bind(self),
			iconUrl: chrome.extension.getURL('img/icons/sortd_icon_48x48.png'),
			iconClass: 'openSortdIcon',
			title: 'Open',
			titleClass: 'openSortdButton'
		});

		//each list row
		self._sdk.Lists.registerThreadRowViewHandler((threadRowView) => {

			self.readReceipts.handleThreadRowViewAdded(threadRowView);

			threadRowView.addButton({
				title: 'Email tracking',
				iconUrl: chrome.extension.getURL('img/readreceipts/blackticks.svg'),
				onClick: self.readReceipts.handleThreadRowClicked.bind(self.readReceipts)
			});

			self.newThreadNotification(true);
		});

		self._sdk.Conversations.registerThreadViewHandler(function (threadView) {
			self.handleOpenedThreadView(threadView, self._sdk.Conversations.MessageViewViewStates);
			self.gmailSendService.handleOpenedThreadView(threadView);
		});

		self._sdk.Conversations.registerMessageViewHandler(function (messageView) {

			self.readReceipts.handleOpenedMessageView(messageView);
			self.linkInterceptor.handleOpenedMessageView(messageView);
		});

		self._sdk.Compose.registerComposeViewHandler(function (composeView) {

			self.handleOpenedComposeView(composeView);
			self.readReceipts.handleOpenedComposeView(composeView);
			self.gmailSendService.handleOpenedComposeView(composeView, self.composeData);

			self.composeData = null;
		});

		this._sdk.Router.handleListRoute(this._sdk.Router.NativeRouteIDs.INBOX, () => {
			this.postMessage('tabsEnabled');
		});

		this.registerDatePlanHandlers();

		this._callbacksRegistered = true;
	};

	SortdInboxSdk.prototype.snoozeComplete = function () {
		if (this.listView) this.listView.refresh();
	};

	SortdInboxSdk.prototype.newThreadNotification = function (sendMessage) {

		if (sendMessage) {
			this.threadAddedToList.emitter.emit();
		} else {
			this.threadAddedToList.sendMessage = false;
		}
	};

	SortdInboxSdk.prototype.checkSendNewThreadNotification = function () {

		if (this.threadAddedToList.sendMessage) {
			// Only send messages in fullscreen mode
			this.messages.sendMessageToExtension('getDisplayMode', {}, ({ displayMode }) => {

				if (displayMode === constants.PANEL_DISPLAY_MODES.FULLSCREEN) {
					this.messages.sendMessageToFrame('newMailThreads', {});
				}
			});
		} else {
			this.threadAddedToList.sendMessage = true;
		}
	};

	SortdInboxSdk.prototype.showAppDropDown = function (event) {

		const dropdown = event.dropdown;
		dropdown.el.innerHTML = constants.HTML.SORTD_MENU;
		dropdown.el.classList.add('sortdTopMenuContainer');

		document.getElementById('sortdMenuFS').addEventListener('click', () => {
			this.messages.sendMessageToExtension('openPanel', { openIfLoggedOut: true });
			dropdown.close();
		});
		document.getElementById('sortdMenuBoard').addEventListener('click', () => {
			this.messages.sendMessageToExtension('openPanelAndSendMessage', {
				action: 'openBoardSwitcher',
				popup: true,
				openIfLoggedOut: true
			});
			dropdown.close();
		});
		document.getElementById('sortdMenuSettings').addEventListener('click', () => {
			this.messages.sendMessageToExtension('openUserSettings', { openIfLoggedOut: true });
			dropdown.close();
		});
		document.getElementById('sortdMenuSupport').addEventListener('click', () => {
			window.open(this._config.chrome.sortd.supportURL);
			dropdown.close();
		});
		document.getElementById('sortdMenuWebsite').addEventListener('click', () => {
			window.open(this._config.chrome.sortd.websiteURL);
			dropdown.close();
		});
	};

	SortdInboxSdk.prototype.registerDayPlanLinks = function () {

		this._dayPlanLinks.registered = true;
		this.messages.addListener('showDatePlan', this.handleShowDatePlan.bind(this));

		this._sdk.NavMenu.addNavItem({
			name: 'My Day plan',
			iconClass: 'sd_date_plan',
			orderHint: 0,
			onClick: (evt) => {
				evt.preventDefault();
				this.handleShowDatePlan();
			}
		});

		this.moveInboxSDKMenuItems();
	};

	SortdInboxSdk.prototype.registerDatePlanHandlers = function () {

		this._sdk.Lists.registerThreadRowViewHandler((threadRowView) => {

			this.listItems.handleThreadRowViewAdded(threadRowView);

			threadRowView.addButton({
				title: 'Add to date plan',
				iconUrl: chrome.extension.getURL('img/date-plan/dateplan_black.png'),
				onClick: this.listItems.handleThreadRowAddToDatePlanClicked.bind(this.listItems)
			});
		});
	};

	SortdInboxSdk.prototype.moveInboxSDKMenuItems = function () {

		this.gmailApi.Dom.observeMutations(document.body, { childList: true, subtree: true })
			.filter((mutations) => mutations.some((m) => {
				if (m.addedNodes.length > 0) {
					for (let i = 0; i < m.addedNodes.length; i++) {
						if (m.addedNodes[i].classList && m.addedNodes[i].classList.contains('inboxsdk__navMenu')) {
							return true;
						}
					}
				}

				return false;
			}))
			.take(1)
			.onValue(() => {

				const inboxsdk__navMenu = document.querySelector('.inboxsdk__navMenu');
				inboxsdk__navMenu.parentNode.prepend(inboxsdk__navMenu);

				const sd_date_plan = inboxsdk__navMenu.querySelector('.sd_date_plan');
				if (sd_date_plan) {
					let parent = sd_date_plan.closest('.inboxsdk__navItem');
					if (parent) {
						parent.firstChild.classList.add('sd_date_plan_nav_item', 'sd_' + constants.DATE_ID.TODAY);
					}

					if (SortdGmailApiDomMap.NAV_MENU) {
						const counterElements = this.gmailApi.Dom.getElementsFromSelector(SortdGmailApiDomMap.NAV_MENU.UNREAD_COUNTER.SELECTORS, parent);
						if (counterElements) {
							counterElements.forEach((el) => {
								el.classList.add('sd_dateplan_counter');
							});
						}
					}

					parent = parent.closest('.inboxsdk__navMenu');
					if (parent) {
						parent.classList.add('sd_date_plan_nav_menu');
					}
				}
			});
	};

	return SortdInboxSdk;
}());
