/**
 * Sortd
 * Copyright (c) 2019, Sortd Inc
 * https://www.sortd.com
 *
 * @fileoverview Chrome extension sortd link related functionality
 * @name sortd-link-interceptor
 */

const SortdLinkInterceptor = (function() {

	class SortdLinkInterceptor {

		constructor(options) {

			this.__FILE_NAME = 'sortd-link-interceptor.js';

			this.config = options.config;
			this.messages = options.messages;
			this.errorHandler = options.errorHandler;
			this.inboxSdk = options.inboxSdk;

			this.urls = this.config.chrome.host.mail3.linkUrls || [this.config.chrome.host.mail3.baseUrl];
		}

		handleOpenedMessageView(messageView) {

			try {
				let displayMode = constants.PANEL_DISPLAY_MODES.MINIMIZED;

				if (messageView.isLoaded()) {

					this.messages.sendMessageToExtension('getDisplayMode', {}, (result) => {
						displayMode = result.displayMode;
					});

					messageView.getBodyElement().addEventListener('click', (event) => {

						if (event.target.nodeName === 'A') {
							this._handleLinkClicked(event, displayMode);
						}
					});
				}
			} catch (ex) {
				this._handleError(ex);
			}
		}

		getViewTaskLinkParams(element) {

			try {
				const validLinkTypes = ['VIEW_TASK', 'VIEW_TASK_NTYPE'];
				for (const link of element.querySelectorAll('a')) {
					if (this._isSortdLink(link) && validLinkTypes.includes(link.dataset.sortdLinkKey)) {
						const appLink = constants.SORTD_APP_LINKS[link.dataset.sortdLinkKey];
						return this._getParams(link, appLink);
					}
				}
			} catch (ex) {
				this._handleError(ex);
			}
		}

		_handleError(error) {
			this.errorHandler.handle(error, this.__FILE_NAME);
		}

		_isSortdLink(link) {

			let isSortdLink = false;

			if (link.dataset.sortdProcessed) {
				if (link.dataset.sortdLinkKey) {
					isSortdLink = true;
				}
			} else {
				const keys = Object.keys(constants.SORTD_APP_LINKS);

				for (const key of keys) {
					if (this._checkIfSortdAppLink(link, key)) {
						isSortdLink = true;
						break;
					}
				}
			}

			return isSortdLink;
		}

		_getParams(link, appLink) {

			const params = {};
			for (let i = 0; i < appLink.PARAM_NAMES.length; i++) {
				params[appLink.PARAM_NAMES[i]] = link.dataset[`sortd${appLink.PARAM_NAMES[i]}`];
			}

			return params;
		}
		
		_handleLinkClicked(event, displayMode) {

			const link = event.target;

			if (this._isSortdLink(link)) {
				this._handleSortdLinkClicked(event, link, displayMode);
			}
		}
		
		_checkIfSortdAppLink(link, appLinkKey) {

			const appLink = constants.SORTD_APP_LINKS[appLinkKey];
			const matches = appLink.URL_REGEX.exec(link.href);

			link.dataset.sortdProcessed = true;

			if (matches && matches.length > (appLink.PARAM_NAMES.length + 1) && this.urls.includes(matches[1])) {
				link.dataset.sortdLinkKey = appLinkKey;

				const data = matches.splice(2, appLink.PARAM_NAMES.length);
				for (let i = 0; i < appLink.PARAM_NAMES.length; i++) {
					link.dataset[`sortd${appLink.PARAM_NAMES[i]}`] = data[i] || '';
				}
				return true;
			} else {
				return false;
			}
		}

		_handleSortdLinkClicked(event, link, displayMode) {

			try {
				event.preventDefault();

				const appLink = constants.SORTD_APP_LINKS[link.dataset.sortdLinkKey];
				if (displayMode === constants.PANEL_DISPLAY_MODES.FULLSCREEN && !appLink.ACTIVE_IN_FULLSCREEN) {
					return;
				}

				const params = this._getParams(link, appLink);
				if (appLink.OPEN_SORTD) {
					params.action = appLink.ACTION;
					this.messages.sendMessageToExtension('openPanelAndSendMessage', params);
				} else {
					this.messages.sendMessageToFrame(appLink.ACTION, params);
				}
			} catch (ex) {
				this._handleError(ex);
			}
		}
	}

	return SortdLinkInterceptor;
}());