/**
 * Sortd
 * Copyright (c) 2019, Sortd Inc
 * https://www.sortd.com
 *
 * @fileoverview Defines the SortdListItems class
 * @name sortd-list-items
 */

class SortdListItems {

	constructor(options) {

		this.__FILE_NAME = 'sortd-list-items.js';

		this.dateViewHighlightAfterInactiveMinutes = 60;

		this.messages = options.messages;
		this.errorHandler = options.errorHandler;
		this.gmailApi = options.gmailApi;

		this.mousePos = { x: 0, y: 0 };
		this.emitTimer = null;
		this.counterTimeouts = {};
		this.dateView = {
			lastActive: Date.now(),
			highlight: false,
			interval: null
		};
		this.listedThreads = null;
		this.listedThreadsForceReload = false;
		this.listItemData = {
			listItems: new Map(),
			emitter: null,
			stream: null
		};
		this.datePlanCounters = {
			counters: {},
			emitter: null,
			stream: null
		};

		this.listItemData.stream = Kefir
			.stream((emitter) => {
				this.listItemData.emitter = emitter;
				this.listItemData.emitter.emit(this.listItemData.listItems);
			})
			.toProperty(() => 0);

		Kefir.stream((emitter) => this.listedThreads = emitter)
			.bufferWithTimeOrCount(1000, 50)
			.onValue((value) => {

				if (value && value.length > 0) {
					this.messages.sendMessageToFrame('setListedThreads', { threadIds: value, forceReload: this.listedThreadsForceReload });
					this.listedThreadsForceReload = false;
				}
			});

		this.datePlanCounters.stream = Kefir
			.stream((emitter) => {
				this.datePlanCounters.emitter = emitter;
				this.datePlanCounters.emitter.emit(this.datePlanCounters.counters);
			})
			.toProperty(() => 0);

		this.messages.addListener('listItemsWithRetries', this.handleListItemsReceived.bind(this));
		this.messages.addListener('datePlanCounters', this.handleDatePlanCounters.bind(this));
		this.messages.addListener('dateViewLastActiveWithRetries', this.handleDateViewLastActive.bind(this));
		this.messages.addListener('getListItemById', this.handleGetListItemById.bind(this));
		this.messages.addListener('getListItemByThreadId', this.handleGetListItemByThreadId.bind(this));

		document.addEventListener('mousemove', this.onMouseUpdate.bind(this), false);
		document.addEventListener('mouseenter', this.onMouseUpdate.bind(this), false);

		this.dateView.interval = setInterval(() => this.checkDateViewHighlight(), 60000);

		this.datePlanCounters.stream.observe((datePlanCounters) => {

			const counters = new Map();

			[constants.DATE_ID.TODAY,
				constants.DATE_ID.ALL_UPCOMING,
				constants.DATE_ID.ALL_OVERDUE
			].forEach((dateId) => {

				const entry = datePlanCounters[dateId];
				const count = (entry && entry.count) || 0;
				let hasMore = (entry && entry.hasMore) || false;

				let text;
				if (count > 20) {
					text = '20+';
				} else {
					text = count.toString();
					if (hasMore) {
						text = text + '+';
					}
				}

				counters.set(dateId, { count, hasMore, displayText: text });
				this.updateCounter(dateId, counters);
				this.updateCounterHighlights();
			});

			this.messages.sendMessageToExtension('datePlanCounters', { counters });
		});
	}

	updateCounter(dateId, counters) {

		const counter = counters.get(dateId);
		if (counter) {

			if (this.counterTimeouts[dateId]) {
				clearTimeout(this.counterTimeouts[dateId]);
			}

			const elements = document.querySelectorAll(`.sd_${dateId} .sd_dateplan_counter`);
			if (elements) {
				elements.forEach((el) => {

					if (el.innerText !== counter.displayText) {
						el.innerText = counter.displayText;

						if (el.dataset.count) {
							const currentCount = parseInt(el.dataset.count, 10);
							if (currentCount > counter.count) {
								el.classList.add('sd_decr');
							} else if (currentCount < counter.count) {
								el.classList.add('sd_incr');
							}
						}
					}
					el.dataset.count = counter.count;

					const parent = el.closest('.sd_date_plan_nav_item');
					if (parent) {
						if (counter.count && counter.count > 0) {
							parent.classList.add('sd_has_items');
						} else {
							parent.classList.remove('sd_has_items');
						}
					}
				});

				this.counterTimeouts[dateId] = setTimeout(() => {
					this.counterTimeouts[dateId] = null;
					elements.forEach((el) => el.classList.remove('sd_incr', 'sd_decr'));
				}, 2000);
			}
		}
	}

	updateCounterHighlights() {

		[constants.DATE_ID.TODAY,
			constants.DATE_ID.ALL_UPCOMING,
			constants.DATE_ID.ALL_OVERDUE
		].forEach((dateId) => {

			const elements = document.querySelectorAll(`.sd_${dateId} .sd_dateplan_counter`);
			if (elements) {
				elements.forEach((el) => {

					if (this.dateView.highlight) {
						el.closest(`.sd_${dateId}`).classList.add('sd_highlight');
					} else {
						el.closest(`.sd_${dateId}`).classList.remove('sd_highlight');
					}
				});
			}
		});
	}

	onMouseUpdate(event) {

		this.mousePos.x = event.pageX;
		this.mousePos.y = event.pageY;
	}

	handleError(error) {
		this.errorHandler.handle(error, this.__FILE_NAME);
	}

	getOverdueLabel(date, dateId) {

		return {
			title: date ? SortdCommon.formatShortDate(date, dateId) : 'Past',
			foregroundColor: '#FFFFFF',
			backgroundColor: '#0597ab'
		};
	}

	getYesterdayLabel() {

		return {
			title: 'Yesterday',
			foregroundColor: '#FFFFFF',
			backgroundColor: '#0597ab'
		};
	}

	getTodayLabel() {

		return {
			title: 'Today',
			foregroundColor: '#FFFFFF',
			backgroundColor: '#00bcd4'
		};		
	}

	getTomorrowLabel() {

		return {
			title: 'Tomorrow',
			foregroundColor: '#FFFFFF',
			backgroundColor: '#5bcedc'
		};		
	}

	getLaterLabel(date, dateId) {

		return {
			title: date ? SortdCommon.formatShortDate(date, dateId) : 'Later',
			foregroundColor: '#FFFFFF',
			backgroundColor: '#5bcedc'
		};		
	}

	getLabelForDate(date) {

		const dateId = SortdCommon.getDateIdFromDate(date);

		switch (dateId) {
			case constants.DATE_ID.ALL_OVERDUE:
				return this.getOverdueLabel(date, dateId);
			case constants.DATE_ID.TODAY:
				return this.getTodayLabel();
			case constants.DATE_ID.YESTERDAY:
				return this.getYesterdayLabel();
			case constants.DATE_ID.TOMORROW:
				return this.getTomorrowLabel();
			case constants.DATE_ID.ALL_UPCOMING:
				return this.getLaterLabel(date, dateId);
		}
	}

	setPlannedDate(eventOrigin, listItem, threadId, title, plannedDate, plannedDateOrder) {

		title = (listItem && listItem.title) || title || '(no subject)';

		const itemId = (listItem && listItem.itemId) || ObjectId.generate();
		let listId = null;
		if (listItem && listItem.list) {
			listId = listItem.list.listId;
		}
		const requestData = {
			eventOrigins: { plannedDate: eventOrigin },
			threadId,
			pinTo: {
				itemId,
				listId,
				title,
				addThreadToTask: false,
				followUp: false,
				plannedDate: plannedDate || null,
				plannedDateOrder: plannedDateOrder || 0
			}
		};

		if (listItem) {
			const userPlannedDate = listItem.content && listItem.content.userPlannedDate;
			requestData.currentPinTo = {
				itemId,
				listId,
				title: listItem.title || null,
				plannedDate: userPlannedDate && userPlannedDate.date || null,
				plannedDateOrder: userPlannedDate && userPlannedDate.displayOrder || 0
			};
		}

		this.messages.sendMessageToFrame('updateTaskFromGmailMode', requestData, (eventData) => {

			if (eventData && eventData.errorStatus) {

				this.activeNotification = this.gmailApi.Notifications.showNotification({
					text: 'Could not add to date plan'
				});
			}
		});
	}

	completeAndArchiveByThreadId(threadId) {

		const listItem = this.listItemData.listItems.get(threadId);
		if (listItem) {

			this.messages.sendMessageToFrame('updateTaskWorkflowStatus', {
				taskId: listItem.itemId,
				workflowStatusId: constants.TASK_WORKFLOW_STATUS.COMPLETED,
				eventOrigin: constants.TRACKER.EVENT_ORIGINS.COMPLETE_AND_ARCHIVE,
				threadIdsToArchive: [threadId]
			});
		}
	}

	handleArchiveCompleteClicked(event) {

		try {
			SortdGmailApiDomMap.THREAD_LIST.ACTION_ITEMS.ARCHIVE_BUTTON.SELECTORS.forEach((archiveSelector) => {
				const archiveButton = document.querySelector(archiveSelector);
				if (archiveButton) {
					SortdCommon.simulateRawClick($(archiveButton));
				}
			});
		} catch (ex) {
			this.handleError(ex);
		}

		try {
			if (event.selectedThreadRowViews && event.selectedThreadRowViews.length > 0) {
				event.selectedThreadRowViews.forEach((threadRowView) => {
					threadRowView.getThreadIDAsync().then((threadId) => this.completeAndArchiveByThreadId(threadId));
				});
			}
		} catch (ex) {
			this.handleError(ex);
		}
	}

	handleThreadRowArchiveCompleteClicked(event) {

		if (event.threadRowView) {
			event.threadRowView.getThreadIDAsync()
				.then((threadId) => {

					const el = document.querySelector(`.sd_thid_${threadId}`);
					if (el) {
						SortdGmailApiDomMap.THREAD_ROW.ACTION_ITEMS.SELECTORS.forEach((barSelector) => {

							const toolbar = el.closest(barSelector);
							if (toolbar) {
								SortdGmailApiDomMap.THREAD_ROW.ACTION_ITEMS.ARCHIVE_BUTTON.SELECTORS.forEach((archiveSelector) => {
									const archiveButton = toolbar.querySelector(archiveSelector);
									if (archiveButton && archiveButton.click) {
										archiveButton.click();
									}
								});
							}
						});
					}

					this.completeAndArchiveByThreadId(threadId);
				})
				.catch((ex) => this.handleError(ex));
		}
	}

	threadRowAdded(threadRowView, threadId) {

		try {
			this.listedThreads.emit(threadId);

			threadRowView.addLabel(
				this.listItemData.stream.map((listItems) => {
					const listItem = listItems.get(threadId);

					if (listItem && listItem.content && listItem.content.userPlannedDate &&
						listItem.content.userPlannedDate.date &&
						listItem.content.currentWorkflowStatusId > constants.TASK_WORKFLOW_STATUS.COMPLETED
					) {
						return this.getLabelForDate(listItem.content.userPlannedDate.date);
					}
				}));

			threadRowView.addButton(
				this.listItemData.stream.map((listItems) => {
					const listItem = listItems.get(threadId);
	
					if (listItem && listItem.content && listItem.content.currentWorkflowStatusId > constants.TASK_WORKFLOW_STATUS.COMPLETED) {
						return {
							title: 'Archive & Complete',
							iconUrl: chrome.extension.getURL('img/date-plan/archive_blue.png'),
							iconClass: `sd_thid_${threadId}`,
							onClick: this.handleThreadRowArchiveCompleteClicked.bind(this)
						};
					}
				}));
		} catch (ex) {
			this.handleError(ex);
		}
	}

	handleThreadRowViewAdded(threadRowView) {

		try {
			threadRowView.getThreadIDAsync()
				.then((threadId) => this.threadRowAdded(threadRowView, threadId))
				.catch((err) => this.handleError(err));
		} catch (ex) {
			this.handleError(ex);
		}
	}

	handleThreadRowAddToDatePlanClicked(event) {

		if (event.threadRowView) {

			event.threadRowView.getThreadIDAsync()
				.then((threadId) => {

					const listItem = this.listItemData.listItems.get(threadId);
					const userPlannedDate = listItem && listItem.content && listItem.content.plannedDate;

					this.gmailApi.Dom.showDateTimeDropDown({
						dateIds: constants.DATE_PLAN_DATE_IDS,
						dateOnly: true,
						heading: 'Add to My Day plan for...',
						selectedDate: userPlannedDate && userPlannedDate.date,
						positionToCoord: this.mousePos,
						inlinePicker: false,
						minDate: 0,
						onDateSelected: (date) => {

							if (date) {
								date.setHours(0, 0, 0, 0);
							}

							this.setPlannedDate(
								constants.TRACKER.EVENT_ORIGINS.GMAIL_THREAD_ROW,
								listItem,
								threadId,
								event.threadRowView.getSubject(),
								date,
								userPlannedDate && userPlannedDate.displayOrder || 0);
						}
					});
				})
				.catch((ex) => this.handleError(ex));
		}
	}

	handleListItemsReceived(message, next) {

		try {
			const listItems = message && message.data && message.data.listItems;

			if (listItems && listItems.size > 0) {
				this.listItemData.listItems = listItems;
			}

			if (this.listItemData.emitter) {
				this.listItemData.emitter.emit(this.listItemData.listItems);
			}
		} catch (ex) {
			this.handleError(ex);
		} finally {
			if (next) {
				next();
			}
		}
	}

	handleDatePlanCounters(message) {

		try {
			const counters = message && message.data && message.data.counters;

			this.datePlanCounters.counters = counters;

			if (this.datePlanCounters.emitter) {
				this.datePlanCounters.emitter.emit(this.datePlanCounters.counters);
			}
		} catch (ex) {
			this.handleError(ex);
		}
	}

	checkDateViewHighlight() {

		const highlightDateViewAfterInactiveMilliseconds = this.dateViewHighlightAfterInactiveMinutes * 60000;
		const highlight = (this.dateView.lastActive < (Date.now() - highlightDateViewAfterInactiveMilliseconds));

		if (this.dateView.highlight !== highlight) {
			this.dateView.highlight = highlight;
			this.updateCounterHighlights();
		}
	}

	handleDateViewLastActive(message, next) {

		try {
			this.dateView.lastActive = message && message.data && message.data.dateViewLastActive;
			this.checkDateViewHighlight();
		} catch (ex) {
			this.handleError(ex);
		} finally {
			if (next) {
				next();
			}
		}
	}

	handleGetListItemByThreadId(message, callback) {

		try {
			const threadId = message && message.data && message.data.threadId;
			callback(null, this.listItemData.listItems.get(threadId));
		} catch (ex) {
			this.handleError(ex);
		}
	}

	handleGetListItemById(message, callback) {

		try {
			const listItemId = message && message.data && message.data.listItemId;
			let listItem;
			for (let [threadId, item] of this.listItemData.listItems) {
				if (item.itemId === listItemId) {
					listItem = item;
					break;
				}
			}
			callback(null, listItem);
		} catch (ex) {
			this.handleError(ex);
		}
	}

	getUserPopupReminder(reminders) {

		return reminders.find((reminder) => {
	
			return (reminder.notificationTypes.indexOf(constants.TASK_REMINDER.NOTIFICATION_TYPES.POPUP) !== -1 &&
				reminder.resourceType === constants.TASK_REMINDER.RESOURCE_TYPES.USER && reminder.resourceId === SortdDataService.userId);
		});
	}

	getObsListItemByThreadId(threadId) {

		let firstRun = true;

		return this.listItemData.stream.map((items) => {

			const item = items.get(threadId);

			if (!item && this.listedThreads) {
				if (firstRun) {
					this.listedThreadsForceReload = true;
				}
			}

			this.listedThreads.emit(threadId);

			firstRun = false;

			return item;
		});
	}
}
