/**
 * Sortd
 * Copyright (c) 2019, Sortd Inc
 * https://www.sortd.com
 *
 * @fileoverview Waits for document ready event, and then loads Sortd extension
 * @name sortd-loader
 */

var SortdLoader = (function () {

	'use strict';

	const MAX_RETRIES = 10;

	/**
	 * Initialize the class
	 * @class SortdLoader
	 */
	function SortdLoader() {

		var self = this,
			myOverrides = {};

		if (typeof overrides !== 'undefined') {
			myOverrides = overrides;
		}

		self._initNext = null;
		self._initTimeout = null;

		self._scriptLoaded = false;
		self._loadFullScreen = false;
		self._emailAddress = null;
		self._loaderStyle = null;
		self._extensionSettings = {};

		self._config = new config(myOverrides);

		self.FILE_NAME = 'sortd-loader.js';

		self._errorHandler = new SortdErrorHandler({
			config: self._config
		});

		self._messages = SortdMessages.getInstance({
			config: self._config,
			errorHandler: self._errorHandler
		});

		self.sortdPermissions = new SortdPermissions({
			config: self._config,
			messages: self._messages,
			errorHandler: self._errorHandler
		});

		self._tracker = new SortdTracker({
			config: self._config,
			messages: self._messages,
			errorHandler: self._errorHandler
		});

		// Initialize shortcut helper here as GmailApi.DOM uses it, but can't initialize it
		self.shortcutHelper = SortdShortcutHelper.getInstance({
			config: self._config,
			messages: self._messages,
			errorHandler: self._errorHandler
		});

		GmailApi
			.load({
				config: self._config,
				errorHandler: self._errorHandler,
				messages: self._messages,
				tracker: self._tracker
			})
			.then((gmailApi) => {
				self._gmailApi = gmailApi;
			});

		self.timerEvents = {};

		self.loggedIn = false;
	}

	SortdLoader.prototype._setLoggedIn = function(isLoggedIn) {
		this.loggedIn = isLoggedIn;
	};

	/**
	 * Logs the error
	 * @param {Error} err - The error
	 * @private
	 */
	SortdLoader.prototype._createNonNotificationError = function (err) {

		err.nonNotification = true;
		return err;
	};

	/**
	 * Logs the error
	 * @param {Error} err - The error
	 * @private
	 */
	SortdLoader.prototype._handleError = function (err, warnOnly) {

		var self = this,
			options;

		if (self._loadFullScreen && !warnOnly) {
			self._cancelFullScreenLoad({ properties: { reason: 'Error', message: err.message } });
		}

		if (warnOnly) {
			options = {
				message: constants.ERROR_MESSAGE.EXTENSION_ERROR,
				technical: err.message,
				fileName: self.FILE_NAME,
				originalError: err
			};
			self._errorHandler.logMessage({ level: constants.LOG.TYPE.WARN, message: options.message, meta: options });
		} else {
			self._errorHandler.handle(err, self.FILE_NAME);
		}
	};

	/**
	 * Track an event
	 * @param {string} event - The event to track
	 * @param {object} properties - Key/values
	 * @param {boolean} [updateLastActivity] - False if the last activity should not be updated
	 * @param {boolean} [trackOnce] - Defaults to false. True if the action should only be tracked once
	 * @private
	 */
	SortdLoader.prototype._track = function (event, properties, updateLastActivity, trackOnce) {

		try {
			this._tracker.track(event, properties, updateLastActivity, trackOnce);
		} catch (ex) {
			this._errorHandler.handle(ex, this.FILE_NAME);
		}
	};

	/**
	 * Get extension settings and cache locally
	 * Save the domMap to local storage
	 * @param {Function} [next] - The results callback
	 * @private
	 */
	SortdLoader.prototype._updateExtensionSettings = function (options, next) {

		var self = this;

		function update(settings) {

			self._extensionSettings = settings;

			if (self._extensionSettings.domMap) {
				localStorage.setItem(constants.STORAGE.KEYS.DOM_MAP, JSON.stringify(self._extensionSettings.domMap));
			}
			if (self._extensionSettings.buildNo) {
				localStorage.setItem(constants.STORAGE.KEYS.BUILD_NO, self._extensionSettings.buildNo);
			}
			next && next(null);
		}

		try {
			if (options && options.settings) {
				update(options.settings);
			} else {

				self._messages.sendMessageToEventPage('getExtensionSettings', null, function (err, result) {

					try{

						if (!err && result) {
							update(result);
						} else {
							if (!err) {
								err = new Error('Can\'t read settings');
							}
							next && next(err);
						}
					} catch (ex) {
						self._handleError(ex);
						next && next(ex);
					}
				});
			}
		} catch (ex) {
			self._handleError(ex);
			next && next(ex);
		}
	};

	/**
	 * Load extensions settings from the server
	 * @param {Function} next - The results callback
	 * @private
	 */
	SortdLoader.prototype._loadSettings = function (next) {

		var self = this,
			http;

		try {

			http = new SortdHttp();
			http.get(
				self._config.chrome.host.mail3.baseUrl + '/extension/settings',
				null,
				function (err, result) {

					try {
						if (err) {
							err.technical = 'error in SortdLoader.prototype._loadSettings http.get';
							self._handleError(err);
							self._updateExtensionSettings(null, next);
						}
						else if (result) {
							self._messages.sendMessageToEventPage('saveExtensionSettings', result, function (err) { });
							self._updateExtensionSettings({ settings: result }, next);
						} else {
							self._updateExtensionSettings(null, next);
						}
					} catch (ex) {
						self._handleError(ex);
						next && next(ex);
					}
				});
		} catch (ex) {
			try {
				ex.technical = 'exception in SortdLoader.prototype._loadSettings';
				self._handleError(ex);
				self._updateExtensionSettings(null, next);
			} catch (ex) {
				self._handleError(ex);
				next && next(ex);
			}
		}
	};

	/**
	 * Check if we are in the top most frame
	 * @returns {boolean} - True if the current window is the top window
	 * @private
	 */
	SortdLoader.prototype._isTopDocument = function () {
		return top.document === window.document;
	};

	/**
	 * Get the current page url
	 * @returns {String} - The current url
	 * @private
	 */
	SortdLoader.prototype._getURL = function () {
		return document.location.href;
	};

	/**
	 * Check if the extension should be loaded.
	 * @returns {boolean} - True if the extension should be loaded
	 * @private
	 */
	SortdLoader.prototype._canLoad = function () {

		var self = this,
			canLoad = false,
			url = self._getURL();

		try {
			if (self._isTopDocument()) {

				canLoad = true;

				constants.DONT_LOAD_WHEN_WINDOW_LOC_CONTAINS.forEach(function (value) {

					if (url.indexOf(value) >= 0) {
						canLoad = false;
					}
				});

				// Check if another Sortd extension is already active
				// If not set this one as the active extension
				if (canLoad) {
					var head = (document.head || document.documentElement);
					var sortdId = head.getAttribute('sortd-ce-id');
					if (sortdId && sortdId !== chrome.runtime.id.toString()) {
						self._handleError(new Error('You have multiple Sortd extensions enabled!'));
						self._cancelFullScreenLoad({ properties: { reason: 'Error', message: 'Multiple extension enabled' }});
						canLoad = false;
					} else {
						head.setAttribute('sortd-ce-id', chrome.runtime.id);
					}
				}
			}
		} catch (ex) {
			self._handleError(ex);
		}
		return canLoad;
	};

	/**
	 * Get the current user's email address
	 * @param {function} next - The results callback
	 * @private
	 */
	SortdLoader.prototype._getEmailAddress = function (next) {

		try {
			this._gmailApi.User.getEmailAddress().then((emailAddress) => {

				sessionStorage.setItem(constants.STORAGE.KEYS.EMAIL, emailAddress.toLowerCase());
				next(null, emailAddress);
			}).catch((err) => {
				this._handleError(err);
				next(err);
			});
		} catch (ex) {
			this._handleError(ex);
			next(ex);
		}
	};

	/**
	 * Load the user's settings from the background page
	 * @param {string} emailAddress - The email adress of the user to retrieve settings for
	 * @param {function} next - The results callback
	 * @private
	 */
	SortdLoader.prototype._getUserSettings = function (emailAddress, next) {

		var self = this,
			data = { email: emailAddress };

		try {
			self._messages.sendMessageToEventPage('getUserSettings', data, function (err, settings) {

				try {
					if (!err) {
						self._settings = settings;
						if (settings) {
							const debugKey = constants.USER_SETTING_KEYS.DEBUG_MODE;
							if (settings && settings[debugKey] && typeof settings[debugKey].value === 'boolean') {
								self._config.debugMode = settings[debugKey].value;
							}
						}
					}

					next(err, settings);
				} catch (ex) {
					self._handleError(ex);
					next(ex);
				}
			});
		} catch (ex) {
			self._handleError(ex);
			next(ex);
		}
	};

	/**
	 * Check if Sortd should be loaded fullscreen or not
	 * @param {function} next - The results callback
	 * @private
	 */
	SortdLoader.prototype._checkLoadFullscreenSetting = function (next) {

		var self = this;

		try {
			self._getEmailAddress(function (err, email) {

				var error,
					key,
					fullscreenOverride;

				try {
					if (err) {
						self._handleError(err);
						next(err);
					} else if (email) {
						self._emailAddress = email.toLowerCase();
						if (self._messages) {
							self._messages.setEmailAddress(self._emailAddress);
							self._messages.setTracker(self._tracker);
						}

						// Check if the user has overridden their default fullscreen setting
						key = constants.STORAGE.KEYS.FULLSCREEN_OVERRIDE.replace(constants.STORAGE.EMAIL_PLACEHOLDER, self._emailAddress);
						fullscreenOverride = localStorage.getItem(key);

						var track = constants.TRACKER.EVENTS.INSTALLED_EXTENSION;
						self._track(track.event, {}, false, track.trackOnce);

						if (fullscreenOverride) {
							return next(null, JSON.parse(fullscreenOverride));
						}

						self._getUserSettings(self._emailAddress, function (err, settings) {

							try {
								var fullscreen = false,
									fullscreenKey = constants.USER_SETTING_KEYS.LOAD_FULLSCREEN,
									featureFlags = null,
									featureFlagsKey = constants.USER_SETTING_KEYS.FEATURE_FLAGS,
									fsKey = constants.STORAGE.KEYS.FULLSCREEN.replace(constants.STORAGE.EMAIL_PLACEHOLDER, self._emailAddress),
									ffKey = constants.STORAGE.KEYS.FEATURE_FLAGS .replace(constants.STORAGE.EMAIL_PLACEHOLDER, self._emailAddress);

								if (err) {
									next(err);
								} else {
									if (settings && settings[fullscreenKey] && settings[fullscreenKey].value) {
										fullscreen = settings[fullscreenKey].value;
									}

									if (settings && settings[featureFlagsKey] && settings[featureFlagsKey].value && settings[featureFlagsKey].enabled) {
										featureFlags = settings[featureFlagsKey].value;
									}

									localStorage.setItem(fsKey, fullscreen);
									localStorage.setItem(ffKey, JSON.stringify(featureFlags));

									next(null, fullscreen);
								}
							} catch (ex) {
								self._handleError(ex);
								next(ex);
							}
						});
					} else {
						error = new Error('Could not determine user\'s email address');
						self._errorHandler.logMessage({
							level: constants.LOG.TYPE.WARN,
							message: error.message,
							meta: error
						});
						next(error);
					}
				} catch (ex) {
					self._handleError(ex);
					next(ex);
				}
			});
		} catch (ex) {
			self._handleError(ex);
			next(ex);
		}
	};

	/**
	 * Cancel the fullscreen load
	 * @private
	 */
	SortdLoader.prototype._cancelFullScreenLoad = function (options) {

		var self = this,
			key;

		self._loadFullScreen = false;

		if (self._emailAddress) {
			key = constants.STORAGE.KEYS.FULLSCREEN.replace(constants.STORAGE.EMAIL_PLACEHOLDER, self._emailAddress);
			localStorage.setItem(key, self._loadFullScreen);
		}

		self._messages.sendMessageToExtension('cancelFullScreenLoad');

		setTimeout(function () {

			self._removeFSLoader();
		}, 500);
	};

	/**
	 * Make sure fullscreen is eventually cancelled
	 * @private
	 */
	SortdLoader.prototype._checkFSLoaderRemoved = function () {

		var self = this;

		window.setTimeout(function () {

			if (self._loadFullScreen && self._loaderStyle) {
				self._cancelFullScreenLoad({ properties: { reason: '20 Second timout' } });
			}
		}, 20000);
	};

	/**
	 * Save the fullscreen setting
	 * @param {Number} fullscreen - True if sortd mode was selected
	 * @private
	 */
	SortdLoader.prototype._saveFullScreenSetting = function (fullscreen) {

		var self = this,
			key = constants.STORAGE.KEYS.FULLSCREEN_OVERRIDE.replace(constants.STORAGE.EMAIL_PLACEHOLDER, self._emailAddress);

		localStorage.setItem(key, !!fullscreen);

		self._messages.sendMessageToExtension('checkFullScreenOverride');
	};

	/**
	 * Click handler when the fullscreen mode is changed
	 * @param {DOMElement} element - The dom element that was clicked
	 * @private
	 */
	SortdLoader.prototype._toggleSortdModeHandler = function (element) {

		var self = this,
			timeout,
			el = $(element),
			fullscreen = el.data('sortd_fullscreen');

		if (typeof fullscreen !== 'undefined') {

			$('.sortdLoaderOptionsMenu li').removeClass('selected');
			el.addClass('selected');

			if (fullscreen === 0) {
				self._cancelFullScreenLoad({ userCancelled: true });
			}

			$('#sortdLoaderRememberSetting').show();
			timeout = setTimeout(function () {

				$('#sortdLoaderRememberSetting').hide();
				timeout = null;
			}, 5000);

			$('#sortdLoaderRememberSetting .sortdRememberOption').click(function () {

				var el = $(this);

				if (el.data('sortd_save_setting') === 1) {
					self._saveFullScreenSetting(fullscreen);
				}
				if (timeout) {
					clearTimeout(timeout);
					timeout = null;
				}
				$('#sortdLoaderRememberSetting').hide();
			});
		}
	};

	/**
	 * Add sortd menu HTML to the loader
	 * @private
	 */
	SortdLoader.prototype._addLoaderMenuHTML = function () {

		var self = this;

		try {
			$(constants.loader.loadingContainer).append(constants.loader.loaderMenuHtml);
			$('body').append(constants.loader.loaderMenuRememberHtml);

			$('.sortdLoaderOptionsMenu li[data-sortd_fullscreen="' + (self._loadFullScreen ? 1 : 0) + '"]').addClass('selected');
			$('.sortdLoaderOptionsMenu li').click(function () {
				self._toggleSortdModeHandler(this);
			});
		} catch (ex) {
			self._handleError(ex);
		}
	};

	/**
	 * Add css for the Sortd menu in the gmail loader
	 * @private
	 */
	SortdLoader.prototype._addLoaderMenuCSS = function () {

		try {
			var style,
				css = constants.loader.loaderMenuCSS
					.replace('TICK_ICON_URL', chrome.extension.getURL('img/tick_small_grey.png'));

			style = document.createElement('style');

			style.type = 'text/css';
			if (style.styleSheet){
				style.styleSheet.cssText = css;
			} else {
				style.appendChild(document.createTextNode(css));
			}

			(document.head || document.documentElement).appendChild(style);
		} catch (ex) {
			this._handleError(ex);
		}
	};

	SortdLoader.prototype._retrieveImageBackgroundInfo = function (imgId) {

		return constants.loader.ImagesBackgroundInfos[imgId];
	};
	/**
	 * Add loading sortd text to the gmail loader
	 * @private
	 */
	SortdLoader.prototype._addFSLoaderHTML = function (imgId) {

		var self = this,
			imgBackgroundInfo;

		try {
			imgBackgroundInfo = self._retrieveImageBackgroundInfo(imgId);
			$('body').append(
				constants.loader.fullscreenLoaderHtml
					.replace('USER_EMAIL_ADDRESS', self._emailAddress || '')
					.replace('LOADER_PROGRESS_BAR_URL', chrome.extension.getURL('img/progressbar.gif'))
					.replace('CREDITS_BACKGROUND_INFO_TITLE', imgBackgroundInfo.title)
					.replace('CREDITS_BACKGROUND_INFO_SOURCE_AUTHOR', imgBackgroundInfo.sourceAuthor)
					.replace('CREDITS_BACKGROUND_INFO_SOURCE_LINK', imgBackgroundInfo.sourceLink));

			$('.sortdLoaderMenuItems').prepend('<div class="sortdInfo" data-sortd_menu="info"><img src="' + chrome.extension.getURL('img/loader_i.png') + '" alt="Info"></div>');
			$('.sortdLoaderMenuItems > div').click(function () {

				$('.sortdLoaderMenu [data-sortd_menu_option]').hide();
				$('.sortdLoaderMenu [data-sortd_menu_option="' +  $(this).data('sortd_menu') + '"]').show();
			});
		} catch (ex) {
			self._handleError(ex);
		}
	};

	/**
	 * Add css to keep the gmail loader visible
	 * And css for the fullscreen loader
	 * @private
	 */
	SortdLoader.prototype._addFSLoaderCSS = function () {

		try {
			var self = this,
				style,
				imgId = Math.floor(Math.random() * constants.loader.backgroundImgCount) + 1,
				css = constants.loader.fullscreenLoaderCSS
					.replace('SORTD_LOADER_LOGO_URL', chrome.extension.getURL('img/loader_sortd_logo.png'))
					.replace('SORTD_LOADER_BACKGROUND_URL', chrome.extension.getURL('img/loader-background/' + imgId + '.jpg'));

			if (!self._loaderStyle) {
				style = document.createElement('style');

				style.type = 'text/css';
				if (style.styleSheet){
					style.styleSheet.cssText = css;
				} else {
					style.appendChild(document.createTextNode(css));
				}

				(document.head || document.documentElement).appendChild(style);
				self._loaderStyle = style;
			}
			return imgId;
		} catch (ex) {
			console.log(ex);
		}
	};

	/**
	 * Remove the sortd loading, and replace with the original gmail html
	 * @private
	 */
	SortdLoader.prototype._removeFSLoader = function () {

		try {
			var self = this;

			$('#sortdLoader').remove();

			if (self._loaderStyle) {
				(document.head || document.documentElement).removeChild(self._loaderStyle);
				self._loaderStyle = null;
			}
		}
		catch (ex) {
			console.log(ex);
		}
	};

	/**
	 * record timer events
	 * @private
	 */
	SortdLoader.prototype._timerMessage = function (message) {

		try {
			var self = this,
				ms,
				eventTime;

			if (message && message.data && message.data.timerEvent) {

				if (message.data.timerEvent === 'timeout' || message.data.timerEvent === 'panelReady' || message.data.timerEvent === 'loginPanelReady' || message.data.timerEvent === 'noload') {
					window.clearTimeout(self.timoutTimer);
				}

				if (message.data.timerEvent === 'timeout') {

					if (!self.timerEvents['panelReady'] && !self.timerEvents['timeout'] && !self.timerEvents['loginPanelReady'] && !self.timerEvents['noload']) {

						eventTime = new Date();
						self.timerEvents[message.data.timerEvent] = eventTime;
						ms = (eventTime.getTime() - self.startTime.getTime());

						// no event recieved before timeout
						self._errorHandler.logMessage({ level: constants.LOG.TYPE.ERROR, message: 'timer Message', meta: { event: 'timeout', elapsedTimeMs: ms, startTime: self.startTime, eventTime: eventTime } });
					}

				} else if (!self.timerEvents[message.data.timerEvent]  && !self.timerEvents['panelReady'] && !self.timerEvents['loginPanelReady'] && !self.timerEvents['noload']) {

					eventTime = new Date();
					self.timerEvents[message.data.timerEvent] = eventTime;
					ms = (eventTime.getTime() - self.startTime.getTime());

					self._errorHandler.logMessage({ level: constants.LOG.TYPE.INFO, message: 'timer Message', meta: { event: message.data.timerEvent, elapsedTimeMs: ms, startTime: self.startTime, eventTime: eventTime } });
				}
				
					if (message.data.timerEvent === 'panelReady') {
						self._setLoggedIn(true);
					}
			}
		}
		catch (ex) {
			console.log(ex);
		}
	};

	/**
	 * Set loggedIn flag
	 * @private
	 */
	SortdLoader.prototype._logout = function () {
		this._setLoggedIn(false);
	};

	// Create the sortd iframe video wrapper
	SortdLoader.prototype.createVideoFrame = function() {

		try {
			const iframe = document.createElement('iframe');
			iframe.src = chrome.extension.getURL('templates/video_frame.htm');
			iframe.id = 'sortdVideoWrapper';
			iframe.className = 'sortd videoFrame';
			iframe.frameBorder = '0';
			iframe.scrolling = 'no';
			iframe.style.display = 'none';
			document.body.appendChild(iframe);
		} catch (ex) {
			this._errorHandler.handle(ex, this.FILE_NAME);
		}
	};

	/**
	 * Complete load, by loading css and jsce boostrapper
	 * @private
	 */
	SortdLoader.prototype._loadJSCE = function () {

		var self = this,
			style,
			script,
			documentHead,
			initNext;

		try {

			initNext = self._initNext;
			self._initNext = null;

			if (self._initTimeout) {
				clearTimeout(self._initTimeout);
				self._initTimeout = null;
			}

			if (initNext) {
		
				documentHead = (document.head || document.documentElement);

				// Inject the css file into the DOM
				style = document.createElement('link');
				style.href = chrome.extension.getURL('css/ce.css');
				style.type = 'text/css';
				style.rel = 'stylesheet';
				documentHead.appendChild(style);

				style = document.createElement('link');
				style.href = chrome.extension.getURL('css/sortd-overlay-frame.css');
				style.type = 'text/css';
				style.rel = 'stylesheet';
				documentHead.appendChild(style);

				// Inject the jsce files into the DOM using require
				script = document.createElement('script');
				if (self._config.useRequire) {
					script.src = chrome.extension.getURL('jsce/lib/require.js');
					script.dataset.main = chrome.extension.getURL('jsce/cebootstrapper.js');
				}
				else {
					script.src = chrome.extension.getURL('jsce/cebootstrapper.js');
				}
				documentHead.appendChild(script);
				self.createVideoFrame();

				self._timerMessage({ data: { timerEvent: '_loadJSCE' } });

				initNext(null);
				initNext = null;
			}
		} catch (ex) {
			if (initNext) {
				initNext(ex);
				initNext = null;
			}
		}
	};

	/**
	 * Initialize the loader and message listener
	 * @param {Function} next - results callback
	 */
	SortdLoader.prototype.init = function (retriesRemaining, next) {

		var self = this,
			iframe,
			documentHead,
			frameTimeout;

		try {
			if (document && document.body && document.body.appendChild) {

				documentHead = (document.head || document.documentElement);
				if (documentHead && documentHead.appendChild) {
					if (chrome && chrome.extension && chrome.extension.getURL) {

						iframe = document.getElementById('sortdWrapper');
						if (iframe) {
							iframe.src = chrome.extension.getURL('templates/frame.htm');
						} else {
							// Create the sortd iframe wrapper
							iframe = document.createElement('iframe');
							iframe.src = chrome.extension.getURL('templates/frame.htm');
							iframe.id = 'sortdWrapper';
							iframe.className = 'sortd frame';
							iframe.frameBorder = '0';
							iframe.scrolling = 'no';
							iframe.style.display = 'none';
							document.body.appendChild(iframe);

							self._timerMessage({ data: { timerEvent: 'WrapperFrameCreated' } });
						}

						self._timerMessage({ data: { timerEvent: 'WrapperFrameCreated' } });

						self._initNext = next;
						if (retriesRemaining >= MAX_RETRIES - 2) {
							frameTimeout = 300;
						} else if (retriesRemaining >= MAX_RETRIES - 3) {
							frameTimeout = 1000;
						} else if (retriesRemaining >= MAX_RETRIES - 4) {
							frameTimeout = 5000;
						} else {
							frameTimeout = 10000;
						}

						if (retriesRemaining > 0) {
							self._initTimeout = setTimeout(function () {

								self._initNext = null;
								next(self._createNonNotificationError(new Error('wrapper frame did not send init message')));
							}, frameTimeout);
						}
					} else {
						next(self._createNonNotificationError( new Error('chrome.extension.getUrl does not exist in SortdLoader.prototype.init ')));
					}
				} else {
					next(self._createNonNotificationError( new Error('documentHead.appendChild does not exist in SortdLoader.prototype.init ')));
				}
			} else {
				next(self._createNonNotificationError( new Error('document.body.appendChild does not exist in SortdLoader.prototype.init ')));
			}
		} catch (ex) {
			next(ex);
		}
	};

	/**
	 * Call init.  If an error occurs retry
	 */
	SortdLoader.prototype.initWithRetries = function (retriesRemaining) {

		var self = this;
		
		if (retriesRemaining > 0) {
			self.init(retriesRemaining, function (err) {
				
				if (err) {

					err.retriesRemaining = retriesRemaining - 1;
					self._handleError(err, true);

					if (!err.nonNotification) {
						self._errorHandler.notifyUser(true);
					}
					setTimeout(function () {

						self.initWithRetries.call(self, retriesRemaining - 1);
					}, constants.RETRY_DELAY);
				}
				else {

					self._errorHandler.removeNotification();
				}
			});
		} else {

			self._handleError(new Error('Init exceeded retries'));
			self._errorHandler.notifyUser(false);
		}
	};

	/**
	 * Start loader, check if needs to be in fullscreen mode
	 * Calls initWithRetries
	 */
	SortdLoader.prototype.startLoader = function () {

		var self = this;
		
		try {
			if (self._canLoad()) {

				// Clear this, otherwise email-address-helper in JSCE will probably run before we get an
				// email address from inboxSDK, and return the address for the wrong tab
				sessionStorage.setItem(constants.STORAGE.KEYS.EMAIL, '');

				self.startTime = new Date();
				self._errorHandler.logMessage({ level: constants.LOG.TYPE.INFO, message: 'Sortd loader started', meta: { startTime: self.startTime } });
				self._messages.createListeners(
					self._removeFSLoader.bind(self),
					self._timerMessage.bind(self),
					self._loadJSCE.bind(self),
					self._logout.bind(self));
				
				self.timoutTimer = window.setTimeout(function () {
					
					self._timerMessage({ data: { timerEvent: 'timeout' } });
				}, 1000 * 60 * 5); //five minutes

				SortdCommon
					.waitForResult(() => self._gmailApi)
					.take(1)
					.onValue(() => {
						if (self._gmailApi && self._gmailApi.init) {
							self._gmailApi.init();
						}
					});

				self._loadSettings(function (err) {

					var loadSettingsError = false;

					try {
						if (err && err.status === 502) {
							loadSettingsError = true;
						}

						self._checkLoadFullscreenSetting(function (err, loadFullscreen) {

							try {
								if (loadSettingsError || err) {
									loadFullscreen = false;
									self._cancelFullScreenLoad({ properties: {reason: 'Error', message: 'Load setting error'}});
								}

								self._loadFullScreen = loadFullscreen;

								// Potential issue with chrome.extension.getURL not being defing (as it is in init)
								// but it should not be an issue here, as this takes longer before executing
								if (self._loadFullScreen) {
									var imgId = self._addFSLoaderCSS();
									self._addFSLoaderHTML(imgId);
									self._checkFSLoaderRemoved();
								}
								self._addLoaderMenuCSS();
								self._addLoaderMenuHTML();
							} catch (ex) {
								self._handleError(ex);
							}
						});

						self.initWithRetries(MAX_RETRIES);
					} catch (ex) {
						self._handleError(ex);
					}
				});
			}
		} catch (ex) {
			self._handleError(ex);
		}
	};

	return SortdLoader;
}());

setTimeout(function () {
	new SortdLoader().startLoader();
}, 10);
