/**
 * Sortd
 * Copyright (c) 2019, Sortd Inc
 * https://www.sortd.com
 *
 * @fileoverview Handles messaging between frame, content scripts and background page
 * @name sortd-messages
 */

'use strict';

class SortdMessages {

	constructor(options) {

		this._config = options.config;
		this._errorHandler = options.errorHandler;

		this.FILE_NAME = 'sortd-messages.js';
		this.CALLBACK_TIMEOUT = 2000;
		this.CHECK_INTERVAL = 15000;

		this._callbacks = {};
		this._callbacks.length = 0;
		this._emailAddress = null;
		this._tracker = null;
		this._listeners = {};

		this.messageOriginExtension = 'chrome-extension://' + chrome.runtime.id;
		this._allowedMessageOrigins = [
			this._config.messageOriginGmail,
			this._config.chrome.host.mail3.baseUrl,
			this.messageOriginExtension
		];

		setTimeout(() => {
			this._checkCallbackTimeouts();
		}, this.CHECK_INTERVAL);
	}

	/**
	 * Logs the error
	 * @param {Error} err - The error
	 * @private
	 */
	_handleError(err) {
		this._errorHandler.handle(err, this.FILE_NAME);
	}

	/**
	 * Delete a callback from callbacks
	 * @param {String} hashKey - The hashkey of the object
	 */
	_deleteCallback(hashKey) {

		if (this._callbacks.hasOwnProperty(hashKey)) {
			delete this._callbacks[hashKey];
			this._callbacks.length--;
		}
	}

	/**
	 * Add a callback to callbacks and return the hash
	 * @param {Function} cb - The callback function
	 * @return {String} - The hash key
	 */
	_addCallback(cb) {

		var timestamp = new Date().getTime(),
			hashKey = timestamp + Math.random();

		this._callbacks[hashKey] = {
			callback: cb,
			timeout: (timestamp + this.CALLBACK_TIMEOUT)
		};
		this._callbacks.length++;

		return hashKey;
	}

	/**
	 * Find callbacks that have timed out and remove them
	 */
	_checkCallbackTimeouts() {

		const timestamp = new Date().getTime();

		if (this._callbacks.length > 0) {
			for (let hashKey in this._callbacks) {
				if (this._callbacks[hashKey].timeout < timestamp) {
					this._deleteCallback(hashKey);
				}
			}
		}
		window.setTimeout(() => {
			this._checkCallbackTimeouts();
		}, this.CHECK_INTERVAL);
	}

	/**
	 * Send a message to the parent frame
	 * @param {String} action - The action to perform
	 * @param {Object} [data] - The data to send to the action
	 * @param cb
	 * @param destination
	 * @private
	 */
	_sendMessageToExtensionOrFrame(action, data, cb, destination) {

		try {
			let target;
			let targetOrigin;

			const messageData = {
				action: action,
				data: data || {},
				owner: 'sortd',
				source: 'ce',
				destination: destination
			};

			if (cb) {
				messageData.consumerHash = this._addCallback(cb);
			}

			if (destination === 'frame') {
				const wrapper = document.getElementById('sortdWrapper');
				if (wrapper) {
					target = wrapper.contentWindow;
					targetOrigin = this.messageOriginExtension;
				}
			} else {
				target = window;
				targetOrigin = this._config.messageOriginGmail;
			}

			if (target && target.postMessage && typeof target.postMessage === 'function') {
				target.postMessage(messageData, targetOrigin);
			}
		} catch (ex) {
			this._handleError(ex);
		}
	}

	/**
	 *
	 * @param {Request} request - The request that was received
	 * @param {Sender} sender - The message sender
	 * @param {Function} sendResponse - The callback
	 * @private
	 */
	_receiveEventPageMessage(request, sender, sendResponse) {

		try {
			if (request.action) {
				this.sendMessageToExtension(request.action, request.data);
			}

			if (sendResponse && typeof sendResponse === 'function') {
				sendResponse(null);
			}
		}
		catch (ex) {
			this._handleError(ex);
		}
	}

	/**
	 * Reply to a message with the data
	 * @param {Object} message - The message to reply to
	 * @param {Object} data - The data to send with the reply
	 * @private
	 */
	_replyToMessage(message, data) {

		let target;
		let targetOrigin;
		let postMessageData = null;

		if (message.consumerHash) {
			if (message.source === 'frame') {
				var wrapper = document.getElementById('sortdWrapper');
				if (wrapper) {
					target = wrapper.contentWindow;
					targetOrigin = this.messageOriginExtension;
				}
			} else {
				target = window;
				targetOrigin = this._config.messageOriginGmail;
			}

			if (target && target.postMessage) {
				if (data) {
					postMessageData = this._errorHandler.createMetaObject(data);
				}
				target.postMessage({
					consumerHash: message.consumerHash,
					data: postMessageData,
					owner: 'sortd',
					source: 'ce',
					destination: message.source
				}, targetOrigin);
			}
		}
	}

	/**
	 * Receive a message from the window
	 * @param {Event} evt - The message event that was received
	 */
	_receiveWindowMessage(evt) {

		try {
			const message = evt.data;
			
			if (message && message.owner === 'sortd' && this._allowedMessageOrigins.indexOf(evt.origin) !== -1) {

				if (message.action && message.destination === 'bgpage') {

					this.sendMessageToEventPage(message.action, message.data, (err, result) => {

						try {
							this._replyToMessage(message, {
								err: err,
								result: result
							});
						} catch (ex) {
							this._handleError(ex);
						}
					});
				} else if (message.source !== 'ce' || message.target === 'ce') {
					if (message.consumerHash && this._callbacks[message.consumerHash]) {
						this._callbacks[message.consumerHash].callback(message.data);
						this._deleteCallback(message.consumerHash);
					} else if (message.destination === 'ce') {
						if (message.action === 'hideSortdLoader') {
							if (this._hideLoaderCb) {
								this._hideLoaderCb(message);
							}

							this._replyToMessage(message);
						} else if (message.action === 'getCeVersion') {
							this._replyToMessage(message, {
								version: this._config.version
							});
						} else if (message.action === 'getCeType') {
							this._replyToMessage(message, {
								ceType: this._config.ceType
							});
						} else if (message.action === 'getCeID') {
							this._replyToMessage(message, {
								id: chrome.runtime.id
							});
						} else if (message.action === 'timer') {
							if (this._timerMessageCB) {
								this._timerMessageCB(message);
							}

							this._replyToMessage(message);
						} else if (message.action === 'logout') {
							if (this._logoutCB) {
								this._logoutCB(message);
							}

							this._replyToMessage(message);
						} else if (message.action === 'frameWrapperLoaded') {
							if (this._frameLoadedCb) {
								this._frameLoadedCb(message);
							}
						} else if (message.action === 'getFullScreenSetting') {
							this._replyToMessage(message, { setting: this._getFullScreenSetting() });
						} else if (message.action === 'trackEvent') {
							if (this._tracker && message && message.data && message.data.event) {
								this._tracker.track(
									message.data.event,
									message.data.meta,
									message.data.updateLastActivity,
									message.data.trackOnce);
							}
						} else if (message.action === 'trackUserProperties') {
							if (this._tracker && message && message.data) {
								this._tracker.setUserProperty(message.data);
							}
						} else if (message.action === 'updateUniqueId') {
							if (this._tracker && message && message.data) {
								this._tracker.updateUniqueId(message.data);
							}
						} else if (message.action === 'debugMode') {
							if (message.data && typeof message.data.debugMode === 'boolean') {
								this._config = message.data.debugMode;
							}
							this._replyToMessage(message);
						} else {
							this._callListeners(message.action, message);
						}
					}
				}
			}
		} catch (ex) {
			this._handleError(ex);
		}
	}

	/**
	 * Send a message to the extension
	 * @param {String} action - The action to perform
	 * @param {Object} [data] - The data to send to the action
	 * @private
	 */
	sendMessageToExtension(action, data, cb) {
		this._sendMessageToExtensionOrFrame(action, data, cb, 'jsce');
	}

	/**
	 * Send a message to the extension
	 * @param {String} action - The action to perform
	 * @param {Object} [data] - The data to send to the action
	 * @private
	 */
	sendMessageToFrame(action, data, cb) {
		this._sendMessageToExtensionOrFrame(action, data, cb, 'frame');
	}

	/**
	 * Send a message to the parent frame
	 * @param {String} action - The action to perform
	 * @param {Object} [data] - The data to send to the action
	 * @param {Function} next - The results callback
	 * @private
	 */
	sendMessageToEventPage(action, data, next) {

		try {
			chrome.runtime.sendMessage({
					action: action,
					data: data
				},
				(response) => {

					if (chrome.extension.lastError) {
						next(chrome.extension.lastError);
					} else {
						next(null, response);
					}
				});
		} catch (ex) {

			if (ex.message && ex.message.toLowerCase().indexOf('extension context invalidated') !== -1) {
				this.sendMessageToExtension('extensionUpdated');
			}

			next(ex);
		}
	}

	/**
	 * Create listeners to pass messages between extension to event page
	 * @param {function} hideLoaderCb - The callback when the loader should be hidden
	 * @private
	 */
	createListeners(hideLoaderCb, timerMessageCB, frameLoadedCb, logoutCB) {

		this._hideLoaderCb = hideLoaderCb;
		this._timerMessageCB = timerMessageCB;
		this._frameLoadedCb = frameLoadedCb;
		this._logoutCB = logoutCB;

		chrome.runtime.onMessage.addListener(this._receiveEventPageMessage.bind(this));
		window.addEventListener('message', this._receiveWindowMessage.bind(this), false);
	}

	setEmailAddress(emailAddress) {
		this._emailAddress = emailAddress;
	}

	setTracker(tracker) {
		this._tracker = tracker;
	}

	addListener(name, cb) {

		if (!this._listeners[name]) {
			this._listeners[name] = [];
		}
		this._listeners[name].push(cb);
	}

	removeListener(name, cb) {

		if (this._listeners[name]) {
			const index = this._listeners[name].indexOf(cb);
			if (index !== -1) {
				this._listeners[name].splice(index, 1);
			}
		}
	}

	_callListeners(name, message) {

		if (this._listeners[name]) {
			this._listeners[name].forEach((cb) => {
				if (message.consumerHash) {
					cb(message, (err, result) => {
						this._replyToMessage(message, { err, result });
					});
				} else {
					cb(message);
				}
			});
		}
	}

	_getFullScreenSetting() {

		if (this._emailAddress) {
			return localStorage.getItem(
				constants.STORAGE.KEYS.FULLSCREEN_OVERRIDE.replace(constants.STORAGE.EMAIL_PLACEHOLDER, this._emailAddress));
		}
	}

	_clearFullScreenSetting() {

		if (this._emailAddress) {
			localStorage.removeItem(
				constants.STORAGE.KEYS.FULLSCREEN_OVERRIDE.replace(constants.STORAGE.EMAIL_PLACEHOLDER, this._emailAddress));
		}
	}
}

SortdMessages.instance = null;

SortdMessages.getInstance = function(options) {

	if (SortdMessages.instance === null) {
		if (!options) {
			throw new Error('SortdMessages has not been initialized correctly');
		}
		SortdMessages.instance = new SortdMessages(options);
	}

	return SortdMessages.instance;
}
