/**
 * Sortd
 * Copyright (c) 2019, Sortd Inc
 * https://www.sortd.com
 *
 * @fileoverview Chrome extension permission related functionality
 * @name sortd-permissions
 */

const SortdPermissions = (function() {

	const permissionConstants = {
		HTML: {
			BROWSER_POPUP: `
				<div id="sd_permission_popup">
					<div id="sdpp_overlay"></div>
					<div class="sd_permissions sd_popup">
						<div class="sd_popup_heading"><strong>{{feature}}</strong></div>
						<div class="sd_popup_close" title="Close"></div>
						<p class="sd_pr_col">
							Please click OK to give your browser<br>
							permission to enable this feature
						</p>
						<div><button class="sortdButton primary">OK</button></div>
						<p class="sd_detail">Sortd needs these permissions to:</p>
						<div class="sd_detail">
							{{featureDetail}}
						</div>
					</div>
				</div>
			`,
			SORTD_POPUP: `
				<div id="sd_permission_popup">
					<div id="sdpp_overlay"></div>
					<div class="sd_permissions sd_popup">
						<div class="sd_popup_heading"><strong>{{feature}}</strong></div>
						<div class="sd_popup_close" title="Close"></div>
						<p class="sd_pr_col">
							Please click OK to give Sortd additional<br>
							permission to activate this feature
						</p>
						<div><button class="sortdButton primary">OK</button></div>
						<p class="sd_detail">Sortd needs these permissions to:</p>
						<div class="sd_detail">
							{{featureDetail}}
						</div>
					</div>
				</div>
			`,
			EMAIL_SHARING_POPUP: `
				<div id="sd_permission_popup">
					<div id="sdpp_overlay"></div>
					<div class="sd_permissions sd_popup" data-full="FULL_PERMISSIONS">
						<div class="sd_popup_heading"><strong>{{feature}}</strong></div>
						<div class="sd_popup_close" title="Close"></div>
						<div class="sd_img_med"><img src="${chrome.extension.getURL('img/email_sharing_pop_up.png')}" alt="Email Sharing"></div>
						<p class="sd_pr_col" data-permissions="full">
							Please click OK to give Sortd additional<br>
							permission to activate this feature
						</p>
						<p class="sd_detail">
							Emails that you add to a shared board will be<br>
							visible to all members of the board,<br>
							including <strong>all replies, cc's and bcc's</strong>
						</p>
						<div><button class="sortdButton primary">OK</button></div>
						<p class="sd_detail" data-permissions="full">Sortd needs these permissions to:</p>
						<div class="sd_detail">
							<span data-permissions="full">Enable powerful new collaboration features<br></span>
							<span data-permissions="full">Share & sync emails between team members<br></span><br>
							For more information on email sharing, <a href="{{infoURL}}"><strong>click here</strong></a>
						</div>
					</div>
				</div>
			`
		}
	};

	class SortdPermissions {

		constructor(options) {

			this.__FILE_NAME = 'sortd-permissions.js';

			this.config = options.config;
			this.messages = options.messages;
			this.errorHandler = options.errorHandler;
			this.permissionsCallback = null;

			this.messages.addListener('showPermissionRequestDialog', this.handleShowPermissionRequestDialog.bind(this));
			this.messages.addListener('handlePermissionsResponse', this.handlePermissionsResponse.bind(this));
		}

		handleError(error) {
			this.errorHandler.handle(error, this.__FILE_NAME);
		};

		closeAndCallback(popupId, callback, err, result) {

			try {
				const popup = document.getElementById(popupId);
				if (popup) {
					popup.remove();
				}

				callback(err || null, result || false);
			} catch (ex) {
				this.handleError(ex);
			}
		}

		requestCePermissions({ permissions, origins }, next) {

			try {
				this.messages.sendMessageToEventPage(
					'requestCePermissions',
					{ permissions, origins },
					(err, result) => next(err, result));
			} catch (ex) {
				this.handleError(ex);
			}
		};

		handleShowPermissionRequestDialog(message, next) {

			try {
				if (message && message.data && message.data && message.data.feature && message.data.permissionType && message.data.permission) {

					const permissionType = message.data.permissionType;
					const permission = message.data.permission;
					const featureType = message.data.featureType;
					const origins = message.data.origins;

					const id = 'sd_permission_' + Math.floor(Math.random() * (90000));
					const close = this.closeAndCallback.bind(this, id, next);
					const div = document.createElement('div');
					let popupHtml;

					switch (permissionType) {
					case 'browser':
						popupHtml = permissionConstants.HTML.BROWSER_POPUP;
						break;
					case 'oauth':
					case 'none' :
						if (featureType && featureType === 'email.sharing') {
							popupHtml = permissionConstants.HTML.EMAIL_SHARING_POPUP;
						} else {
							popupHtml = permissionConstants.HTML.SORTD_POPUP;
						}
						break;
					default:
						popupHtml = permissionConstants.HTML.SORTD_POPUP;
						break;
					}

					popupHtml = popupHtml
						.replace(/{{feature}}/gi, message.data.feature)
						.replace(/{{featureDetail}}/gi, message.data.featureDetail || '')
						.replace(/{{infoURL}}/g, this.config.chrome.sortd.emailSharingPermissionsURL);

					if (permission && permission === 'gmail.full') {
						popupHtml = popupHtml.replace('FULL_PERMISSIONS', 'Y');
					}

					div.innerHTML = popupHtml;

					const popup = div.firstElementChild;
					popup.id = id;
					document.body.appendChild(popup);

					popup.querySelector(`.sd_popup_close`).addEventListener('click', () => close());
					popup.querySelector(`#sdpp_overlay`).addEventListener('click', () => close());

					popup.querySelector('button.sortdButton').addEventListener('click', (evt) => {

						if (evt) {
							evt.preventDefault();
							evt.stopPropagation();
						}

						switch (permissionType) {
							case 'browser':
								this.requestCePermissions({permission, origins}, close);
								break;
							case 'oauth':
								this.permissionsCallback = close;
								this.messages.sendMessageToExtension('requestAdditionalOAuthPermissions', permission);
								break;
							default:
								close(null, true);
								break;
						}
					});
				}
			} catch (ex) {
				this.handleError(ex);
			}
		}

		handlePermissionsResponse(message) {

			if (message && message.data && this.permissionsCallback) {
				if (message.data.err) {
					this.permissionsCallback(message.data.err);
				} else {
					const result = message.data.result ? message.data.result : true;
					this.permissionsCallback(null, result);
				}
			}
		}
	}

	return SortdPermissions;
}());