/**
 * Sortd
 * Copyright (c) 2019, Sortd Inc
 * https://www.sortd.com
 *
 * @fileoverview Read receipt related functionality
 * @name read-receipts
 */

const SortdReadReceipts = (function() {

	const readReceiptConstants = {
		DISPLAY_NOT_LOGGED_IN_NOTIFICATION_DURATION: 3000,
		VIEWS_LENGTH_LIMIT: 50,
		SENT_ITEMS_NAV: {
			MENU_SELECTOR: '.aeN',
			SENT_SELECTOR: '.aim a[href*="#sent"]',
			RETRIES: 20
		},
		CE: {
			PERMISSIONS: [
				'notifications'
			],
			ORIGINS: [
				'*://*.googleusercontent.com/*',
				'*://*.sortd.com/*'
			]
		},
		HTML: {
			COMPOSE_VIEW_DROP_DOWN: `
				<div class="sdrm_compose_dd" data-viewid="{{viewid}}">
					<div class="sdrm_heading">Track this email</div>
					<div class="sdrm_settings_link">
						<img src="${chrome.extension.getURL('img/readreceipts/settings.png')}">
					</div>
					<div class="sdrm_toggle_container">
						<label><input type="checkbox" name="trackOpens" {{trackOpens}}> Track opens</label>
					</div>
					<div class="sdrm_toggle_container">
						<label><input type="checkbox" name="notifyMe" {{notifyMe}}> Notify me when opened</label>
					</div>
					<hr>
					<a href="{{gdprTrackingURL}}" target="_blank">
						Special note for emailing EU citizens
						<img class="sdrm_info" src="${chrome.extension.getURL('img/readreceipts/info.svg')}">
					</a>
				</div>
			`,
			COMPOSE_VIEW_DROP_DOWN_NOT_LOGGED_IN: `
				<div class="sdrm_compose_dd">${constants.NOTIFICATION_MESSAGES.USER_NOT_LOGGED_IN}</div>
			`,
			COMPOSE_VIEW_DROP_DOWN_NO_SUBSCRIPTION: `
				<div class="sdrm_compose_dd">
					Email tracking is not available<br/>on your current plan <a class="sd_primary_upper" href="#"><strong>Upgrade</strong><a/>
				</div>
			`,
			MESSAGE_VIEWS: `
				<div id="sd_rr_message_views">
					<div id="sdrm_overlay"></div>
					<div id="sdrm_popup" class="sd_popup">
						<div class="sd_popup_heading">
							<img src="${chrome.extension.getURL('img/readreceipts/activity_primary.png')}">
							<div class="sd_inlblock">Activity Stream</div>
						</div>
						<div class="sd_popup_close" title="Close"></div>
						<div class="sd_tabs">
							<div data-tabid="threads" class="sd_tab_text">Your thread was opened {{threadViewCount}} time/s</div>
							<div data-tabid="messages" class="sd_tab_text">Your message was opened {{messageViewCount}} time/s</div>
							<div data-tabid="threads" class="sd_tab">Thread ({{threadViewCount}})</div>
							<div class="sd_separator">|</div>
							<div data-tabid="messages" class="sd_tab">Message ({{messageViewCount}})</div>
							<div class="sd_separator">&nbsp;</div>
						</div>
						<div data-tabid="threads" class="sd_tab_content">
							{{threadViews}}
						</div>
						<div data-tabid="messages" class="sd_tab_content">
							{{messageViews}}
						</div>
						<div class="sd_popup_footer">
							<hr>
							<div class="sd_not_in_current_plan">
								Email tracking is not available on your current plan. <span class="sd_clickable">Click here to <span class="sd_primary_upper" href="#"><strong>Upgrade</strong></span></span>
							</div>
							<div class="sdrm_toggle_container">
								<label><input type="checkbox" name="notifyMe" {{notifyMe}}> Notify me when this email is opened</label>
							</div>
						</div>
					</div>
				</div>
			`,
			MESSAGE_VIEWS_NOT_TRACKED: `
				<div id="sd_rr_message_views">
					<div id="sdrm_overlay"></div>
					<div id="sdrm_popup" class="sd_popup">
						<div class="sd_popup_heading">
							<img src="${chrome.extension.getURL('img/readreceipts/activity_primary.png')}">
							<div class="sd_inlblock">Activity Stream</div>
						</div>
						<div class="sd_popup_close" title="Close"></div>
						<p class="sd_popup_body">This message has not been tracked.</p>
						<div class="sd_popup_footer">
							<hr>
							<div class="sd_not_in_current_plan">
								Email tracking is not available on your current plan. <span class="sd_clickable">Click here to <span class="sd_primary_upper" href="#"><strong>Upgrade</strong></span></span>
							</div>
						</div>
					</div>
				</div>
			`,
			MESSAGE_VIEW: `
				<div class="sd_mv_top">
					<img src="${chrome.extension.getURL('img/readreceipts/activity.png')}">
					<div class="sd_inlblock">Opened {{msgLocation}} {{msgDevice}}</div>
				</div>
				<div class="sd_small_text">{{date}}</div>
			`,
			MESSAGE_LOCATION: ' in {{location}}',
			MESSAGE_DEVICE: ' on a {{device_OS}} device',
			PERMISSION_REQUIRED: `
				<div class="sd_permissions">
					<div class="sd_popup_heading"><strong>Email tracking</strong></div>
					<div class="sd_img_lrg"><img src="${chrome.extension.getURL('img/readreceipts/permissions.png')}"></div>
					<p class="sd_pr_col">
						Please click OK to give your browser<br>
						permission to enable these features
					</p>
					<div><button class="sortdButton primary">OK</button></div>
					<p class="sd_detail">Sortd needs these permissions to:</p>
					<p class="sd_detail">
						Exclude your own opens from being tracked and to<br>
						Display browser notifications (when you want them)
						<a href="{{infoURL}}" target="_blank"><img class="sdrm_info" src="${chrome.extension.getURL('img/readreceipts/info.svg')}"></a>
					</p>
				</div>
			`
		}
	};

	function SortdReadReceipts(options) {

		this.__FILE_NAME = 'sortd-read-receipts.js';

		this.config = options.config;
		this.messages = options.messages;
		this.errorHandler = options.errorHandler;
		this.tracker = options.tracker;
		this.hashedUser = options.hashedUser || null;
		this.viewIdCounter = 0;
		this.readReceiptsEnabled = false;	// True if the user is a pro user

		this.sortdDataService = SortdDataService.getInstance({ messages: this.messages });

		this.settings = {
			baseUrl: this.config.chrome.host.mail3.baseUrl + '/rr/{{guid}}?from={{uid}}',
			icon: {
				state: this.config.chrome.host.mail3.assetsUrl + '/img/read-receipt/{{state}}.svg',
				mail: this.config.chrome.host.mail3.assetsUrl + '/img/read-receipt/{{state}}.svg',
			}
		};

		this.readReceiptStateForView = {};

		this.isLoggedIn = false;
		this.newThreadsInList = [];
		this.getThreadInfoTimer = null;
		this.receiptData = {
			receipts: [],
			emitter: null,
			stream: null,
			messageStream: null
		};

		this.receiptData.stream = Kefir
			.stream((emitter) => {
				this.receiptData.emitter = emitter;
				this.receiptData.emitter.emit(this.receiptData.receipts);
			})
			.toProperty(() => 0);

		this.receiptData.messageStream = this.receiptData.stream.map((receipts) => {
			return receipts.reduce((acc, receipt) => {
				acc[receipt.messageId] = receipt;
				return acc;
			}, {});
		});

		this.receiptData.threadStream = this.receiptData.stream.map((receipts) => {
			return receipts.reduce(function(acc, receipt) {

				if (acc[receipt.threadId]) {
					acc[receipt.threadId].views = acc[receipt.threadId].views.concat(receipt.views);
				} else {
					acc[receipt.threadId] = {
						threadId: receipt.threadId,
						views: receipt.views.slice(0)
					};
				}

				return acc;
			}, {});
		});

		this.messages.addListener('readReceipts', this.handleReadReceiptsReceived.bind(this));
		this.messages.addListener('setSortdUser', this.handleSetSortdUserReceived.bind(this));
		this.messages.addListener('isLoggedIn', this.handleUserLogin.bind(this));

		this.addTicksToSentNavMenu(readReceiptConstants.SENT_ITEMS_NAV.RETRIES);
	}

	SortdReadReceipts.prototype.handleError = function(error) {
		this.errorHandler.handle(error, this.__FILE_NAME);
	};

	SortdReadReceipts.prototype.handleUserLogin = function(message) {

		if (message && message.data) {
			this.isLoggedIn = message.data.loggedIn || false;

			if (this.isLoggedIn && !this.getThreadInfoTimer) {
				this.getThreadsInfo();
			}
		}
	};

	SortdReadReceipts.prototype.handleOpenedComposeView = function(composeView) {

		Promise.all(
			[
				this.sortdDataService.getReadReceiptEnabled(),
				this.sortdDataService.getUserSettingValue(constants.USER_SETTING_KEYS.READ_RECEIPTS)
			])
			.then((result) => {

				this.readReceiptsEnabled = result[0];

				const settingValue = result[1];
				const enabled = this.isLoggedIn && this.readReceiptsEnabled && settingValue && settingValue.enabled || false;
				const readReceiptState = enabled ? 'enabled' : 'disabled';
				const viewId = this.viewIdCounter;
				let notifications = false;
				if (settingValue && settingValue.notifications) {
					// Currently only desktop notifications are supported
					notifications = settingValue.notifications.desktop || false;
				}

				this.viewIdCounter++;
				this.readReceiptStateForView[viewId] = {
					enabled,
					notifications,
					guid: ObjectId.generate()
				};

				composeView.addButton({
					title: 'Tracking is currently ' + readReceiptState,
					iconUrl: this.settings.icon.state.replace('{{state}}', readReceiptState),
					iconClass: `sd_toggle_receipts${viewId}`,
					hasDropdown: true,
					onClick: (event) => {
						this.handleComposeButtonClicked(composeView, event, viewId);
					}
				});

				composeView.on('presending', () => this.composePresendHandler(composeView, viewId));
				composeView.on('sent', (event) => this.composeSentHandler(composeView, event, viewId));
				composeView.on('destroy', () => delete this.readReceiptStateForView[viewId]);
			})
			.catch((err) => this.handleError(err));
	};

	SortdReadReceipts.prototype.handleOpenedMessageView = function(messageView) {

		try {
			if (messageView.isLoaded()) {
				messageView.getMessageIDAsync()
					.then((messageId) => {

						messageView.addAttachmentIcon(this.receiptData.messageStream.map((messages) => {
							const message = messages[messageId];

							if (message) {
								if (message.views.length > 0) {
									return {
										iconUrl: chrome.extension.getURL('img/readreceipts/viewed.svg'),
										tooltip: message.views.length + ' views',
										onClick: this.showMessageViews.bind(this, message)
									};
								} else {
									return {
										iconUrl: chrome.extension.getURL('img/readreceipts/sent.svg'),
										tooltip: 'No views',
										onClick: this.showMessageViews.bind(this, message)
									};
								}
							}
						}));
					});
			}
		} catch (ex) {
			this.handleError(ex);
		}
	};

	SortdReadReceipts.prototype.handleThreadRowViewAdded = function(threadRowView) {

		try {
			threadRowView.getThreadIDAsync()
				.then((threadId) => {
					this.threadRowAdded(threadRowView, threadId);
				})
				.catch((err) => {
					this.handleError(err);
				});
		} catch (ex) {
			this.handleError(ex);
		}
	};

	SortdReadReceipts.prototype.handleComposeButtonClicked = function(composeView, event, viewId) {

		if (!this.isLoggedIn) {
			this.addLoginRequired(event.dropdown);
		} else if (!this.readReceiptsEnabled) {
			this.addSubscriptionRequired(composeView, event.dropdown);
		} else {
			const state = this.readReceiptStateForView[viewId];
			event.dropdown.el.innerHTML = readReceiptConstants.HTML.COMPOSE_VIEW_DROP_DOWN
				.replace(/{{trackOpens}}/g, (state && state.enabled) ? 'checked' : '')
				.replace(/{{notifyMe}}/g, (state && state.notifications) ? 'checked' : '')
				.replace(/{{viewid}}/g, viewId)
				.replace(/{{gdprTrackingURL}}/g, this.config.chrome.sortd.readReceiptGdprURL);

			event.dropdown.el.querySelector('.sdrm_settings_link')
				.addEventListener('click', this.openSettings.bind(this, composeView, event.dropdown));

			event.dropdown.el.querySelector('.sdrm_toggle_container [type="checkbox"][name="trackOpens"]')
				.addEventListener('click', this.toggleReceipts.bind(this, event.dropdown.el, viewId));

			event.dropdown.el.querySelector('.sdrm_toggle_container [type="checkbox"][name="notifyMe"]')
				.addEventListener('click', this.toggleNotifications.bind(this, viewId));

			this.checkCePermissions(event.dropdown.el);
			this.updateNotificationCheckbox(event.dropdown.el, viewId);
		}
	};

	SortdReadReceipts.prototype.handleThreadRowClicked = function(event) {

		if (event.threadRowView) {

			event.threadRowView.getThreadIDAsync()
				.then((threadId) => {

					const receipt = this.receiptData.receipts.find((r) => r.threadId === threadId);
					this.showMessageViews(receipt, 'threads');
				});
		}
	};

	SortdReadReceipts.prototype.openSettings = function(composeView, dropdown) {

		try {
			this.messages.sendMessageToExtension('openUserSettings');
			dropdown.close();
			composeView.setMinimized(true);
		} catch (ex) {
			this.handleError(ex);
		}
	};

	SortdReadReceipts.prototype.toggleReceipts = function(container, viewId) {

		try {
			this.readReceiptStateForView[viewId].enabled = !(this.readReceiptStateForView[viewId] && this.readReceiptStateForView[viewId].enabled);
			this.setReceiptsEnabledIcons(container, viewId);
		} catch (ex) {
			this.handleError(ex);
		}
	};

	SortdReadReceipts.prototype.toggleNotifications = function(viewId) {

		try {
			this.readReceiptStateForView[viewId].notifications = !(this.readReceiptStateForView[viewId] && this.readReceiptStateForView[viewId].notifications);
		} catch (ex) {
			this.handleError(ex);
		}
	};

	SortdReadReceipts.prototype.addLoginRequired = function(dropdown) {

		dropdown.el.innerHTML = readReceiptConstants.HTML.COMPOSE_VIEW_DROP_DOWN_NOT_LOGGED_IN;
		dropdown.el.querySelector('#not-loggedin-notification-signin').addEventListener('click', (event) => {

			event.preventDefault();
			this.messages.sendMessageToExtension('showLogin', { fullscreen: false });
			dropdown.close();
		});
	};

	SortdReadReceipts.prototype.addSubscriptionRequired = function(composeView, dropdown) {

		dropdown.el.innerHTML = readReceiptConstants.HTML.COMPOSE_VIEW_DROP_DOWN_NO_SUBSCRIPTION;
		dropdown.el.querySelector('.sdrm_compose_dd .sd_primary_upper').addEventListener('click', (event) => {

			event.preventDefault();
			this.openSubscriptionPopupInSortd();
			dropdown.close();
			composeView.setMinimized(true);
		});
	};

	SortdReadReceipts.prototype.setReceiptsEnabledIcons = function(container, viewId) {

		const viewState = this.readReceiptStateForView[viewId];
		const enabled = (this.isLoggedIn && viewState && viewState.enabled);
		const state = enabled ? 'enabled' : 'disabled';

		const icon = document.querySelector(`.sd_toggle_receipts${viewId} img`);
		if (icon) {
			icon.src = this.settings.icon.state.replace('{{state}}', state);
			const parent = icon.closest('[data-tooltip]');
			if (parent) {
				parent.setAttribute('data-tooltip', 'Tracking is currently ' + state);
				parent.setAttribute('aria-label', 'Tracking is currently ' + state);
			}
		}

		this.updateNotificationCheckbox(container, viewId);
	};

	SortdReadReceipts.prototype.updateNotificationCheckbox = function(container, viewId) {

		const viewState = this.readReceiptStateForView[viewId];
		const enabled = (this.isLoggedIn && viewState && viewState.enabled);

		const el = container.querySelector(`.sdrm_toggle_container [type="checkbox"][name="notifyMe"]`);
		if (el) {
			if (enabled) {
				el.removeAttribute('disabled');
				if (viewState.notifications) {
					el.checked = true;
				}
			} else {
				el.setAttribute('disabled', 'disabled');
				el.checked = false;
			}
		}
	};

	SortdReadReceipts.prototype.composePresendHandler = function(composeView, viewId) {

		try {
			const viewState = this.readReceiptStateForView[viewId];
			if (this.isLoggedIn && viewState && viewState.enabled) {

				const guid = this.readReceiptStateForView[viewId].guid;
				const url = this.settings.baseUrl
					.replace('{{guid}}', guid)
					.replace('{{uid}}', this.hashedUser || '');

				const el = document.createElement('div');
				el.innerHTML = constants.HTML.READ_RECEIPT.replace('READ_RECEIPT_URL', url);
				composeView.getBodyElement().appendChild(el.firstElementChild);
			}
		} catch (ex) {
			this.handleError(ex);
		}
	};

	SortdReadReceipts.prototype.composeSentHandler = function(composeView, event, viewId) {

		try {
			const viewState = this.readReceiptStateForView[viewId];
			if (this.isLoggedIn && viewState && viewState.enabled) {
				Promise.all([event.getThreadID(), event.getMessageID()])
					.then((result) => {

						const recipients = composeView.getToRecipients()
							.map((recipient) => {
								return Object.assign({}, recipient, { type: constants.RECIPIENT_TYPE.TO });
							})
							.concat(composeView.getCcRecipients()
								.map((recipient) => {
									return Object.assign({}, recipient, { type: constants.RECIPIENT_TYPE.CC });
								}));

						this.messages.sendMessageToFrame('trackReadReceipt', {
							threadId: result[0],
							messageId: result[1],
							receiptId: viewState.guid,
							muted: !viewState.notifications,
							subject: composeView.getSubject(),
							recipients
						});
					})
					.catch((err) => this.handleError(err));
			}
		} catch (ex) {
			this.handleError(ex);
		}
	};

	SortdReadReceipts.prototype.threadRowAdded = function(threadRowView, threadId) {

		try {
			if (this.newThreadsInList.indexOf(threadId) === -1) {
				this.newThreadsInList.push(threadId);
			}

			if (this.getThreadInfoTimer) {
				clearTimeout(this.getThreadInfoTimer);
				this.getThreadInfoTimer = null;
			}

			this.getThreadInfoTimer = setTimeout(() => {

				this.getThreadInfoTimer = null;
				this.getThreadsInfo();
			}, 250);

			threadRowView.addAttachmentIcon(
				this.receiptData.threadStream.map((threads) => {
					const thread = threads[threadId];

					if (thread) {
						if (thread.views.length > 0) {
							return {
								iconUrl: chrome.extension.getURL('img/readreceipts/viewed.svg'),
								tooltip: thread.views.length + ' views'
							}
						} else {
							return {
								iconUrl: chrome.extension.getURL('img/readreceipts/sent.svg'),
								tooltip: 'No views'
							}
						}
					}
				}));
		} catch (ex) {
			this.handleError(ex);
		}
	};

	SortdReadReceipts.prototype.getViewEntry = function(view) {

		try {
			const city = view.location.city || '';
			const country = view.location.country || '';
			const device = view.device && view.device.toLowerCase();
			let deviceOS = null;
			let msgLocation = '';
			let msgDevice = '';

			let location;
			if (city && country) {
				location = city + ', ' + country;
			} else {
				location = city || country;
			}

			if (location && device !== 'gmail') {
				msgLocation = readReceiptConstants.HTML.MESSAGE_LOCATION.replace(/{{location}}/g, location);
			}

			if (device === 'other') {
				view.device = null;
			}

			if (view.os && view.device) {
				deviceOS = view.os + ', ' + view.device;
			} else {
				deviceOS = view.os || view.device;
			}

			if (deviceOS) {
				msgDevice = readReceiptConstants.HTML.MESSAGE_DEVICE.replace(/{{device_OS}}/g, deviceOS);
			}

			let msgDates = this.formatMessageViewsDate(new Date(view.dateViewed));
			if (view.datesViewed && view.datesViewed.length > 0) {
				const viewsLen = view.datesViewed.length;
				const moreThanLimit = (viewsLen >= readReceiptConstants.VIEWS_LENGTH_LIMIT);
				const numToAdd = (viewsLen > 5) ? 5 : viewsLen - 1;
				if (moreThanLimit) {
					msgDates = `... ${readReceiptConstants.VIEWS_LENGTH_LIMIT}+ more<br>${msgDates}`;
				} else if (viewsLen > numToAdd + 1) {
					msgDates = `... ${viewsLen - (numToAdd + 1)} more<br>${msgDates}`;
				}
				if (numToAdd > 0) {
					for (let i = (viewsLen - numToAdd); i < viewsLen; i++) {
						msgDates = `${this.formatMessageViewsDate(new Date(view.datesViewed[i]))}<br>${msgDates}`;
					}
				}
			}

			return readReceiptConstants.HTML.MESSAGE_VIEW
				.replace(/{{date}}/g, msgDates)
				.replace(/{{msgDevice}}/g, msgDevice)
				.replace(/{{msgLocation}}/g, msgLocation);
		} catch (ex) {
			this.handleError(ex);
			return '';
		}
	};

	SortdReadReceipts.prototype.getViewCount = function(view) {

		if (view.datesViewed && view.datesViewed.length > 0) {
			return view.datesViewed.length;
		}

		return 1;
	};

	SortdReadReceipts.prototype.showMessageViews = function(message, activeTab) {

		try {
			if (!message) {
				return this.showMessageNotTracked();
			}

			let messageViewCount = 0;
			let threadViewCount = 0;
			const messageViewsHtml = message.views
				.sort((a, b) => {
					const bLastDate = (b.datesViewed && a.datesViewed.length > 0) ? b.datesViewed[b.datesViewed.length - 1] : b.dateViewed;
					const aLastDate = (a.datesViewed && a.datesViewed.length > 0) ? a.datesViewed[a.datesViewed.length - 1] : a.dateViewed;
					return new Date(bLastDate).getTime() - new Date(aLastDate).getTime();
				})
				.reduce((acc, view) => {
					messageViewCount += this.getViewCount(view);
					return acc + this.getViewEntry(view);
				}, '');

			const threadViews = this.receiptData.receipts
				.filter((rcpt) => rcpt.threadId === message.threadId)
				.reduce((acc, msg) => acc.concat(msg.views), []);

			const threadViewsHtml = threadViews
				.sort((a, b) => {
					const bLastDate = (b.datesViewed && a.datesViewed.length > 0) ? b.datesViewed[b.datesViewed.length - 1] : b.dateViewed;
					const aLastDate = (a.datesViewed && a.datesViewed.length > 0) ? a.datesViewed[a.datesViewed.length - 1] : a.dateViewed;
					return new Date(bLastDate).getTime() - new Date(aLastDate).getTime();
				})
				.reduce((acc, view) => {
					threadViewCount += this.getViewCount(view);
					return acc + this.getViewEntry(view);
				}, '');

			if (threadViewCount >= readReceiptConstants.VIEWS_LENGTH_LIMIT) {
				threadViewCount = `${readReceiptConstants.VIEWS_LENGTH_LIMIT}+`;
			}

			if (messageViewCount >= readReceiptConstants.VIEWS_LENGTH_LIMIT) {
				messageViewCount = `${readReceiptConstants.VIEWS_LENGTH_LIMIT}+`;
			}

			const el = document.createElement('div');
			el.innerHTML = readReceiptConstants.HTML.MESSAGE_VIEWS
				.replace(/{{messageViews}}/g, messageViewsHtml)
				.replace(/{{threadViews}}/g, threadViewsHtml)
				.replace(/{{threadViewCount}}/g, threadViewCount)
				.replace(/{{messageViewCount}}/g, messageViewCount)
				.replace(/{{notifyMe}}/g, message.muted ? '' : 'checked');
			document.body.appendChild(el.firstElementChild);

			this.setActiveTab(activeTab || 'messages');

			document.querySelectorAll('#sd_rr_message_views .sd_tab')
				.forEach((tab) => tab.addEventListener('click', this.tabClickHandler.bind(this)));

			document.querySelector('#sd_rr_message_views #sdrm_overlay')
				.addEventListener('click', this.closeMessageViews.bind(this));

			document.querySelector('#sd_rr_message_views .sd_popup_close')
				.addEventListener('click', this.closeMessageViews.bind(this));

			document.querySelector('#sd_rr_message_views [type="checkbox"][name="notifyMe"]')
				.addEventListener('click', this.toggleMuted.bind(this, message));

			this.checkCePermissions(document.querySelector('#sd_rr_message_views #sdrm_popup'));
			this.checkReadReceiptPermissions(document.querySelector('#sd_rr_message_views #sdrm_popup'));

			const evt = constants.TRACKER.EVENTS.OPENED_EMAIL_TRACKING_STREAM;
			this.tracker.track(evt.event, {}, true, evt.trackOnce);
		} catch (ex) {
			this.handleError(ex);
		}
	};

	SortdReadReceipts.prototype.showMessageNotTracked = function() {

		const el = document.createElement('div');
		el.innerHTML = readReceiptConstants.HTML.MESSAGE_VIEWS_NOT_TRACKED;
		document.body.appendChild(el.firstElementChild);

		document.querySelector('#sd_rr_message_views .sd_popup_close')
			.addEventListener('click', this.closeMessageViews.bind(this));

		this.checkCePermissions(document.querySelector('#sd_rr_message_views #sdrm_popup'));
		this.checkReadReceiptPermissions(document.querySelector('#sd_rr_message_views #sdrm_popup'));
	};

	SortdReadReceipts.prototype.closeMessageViews = function() {

		const el = document.querySelector('#sd_rr_message_views');

		if (el) {
			el.remove();
		}
	};

	SortdReadReceipts.prototype.toggleMuted = function(message) {

		try {
			this.messages.sendMessageToFrame('muteReadReceipt', {
				receiptId: message.receiptId,
				muted: !message.muted
			})
		} catch (ex) {
			this.handleError(ex);
		}
	};

	SortdReadReceipts.prototype.getThreadsInfo = function() {

		if (this.isLoggedIn) {

			this.messages.sendMessageToFrame(
				'loadReadReceiptViews',
				{
					threadIds: this.newThreadsInList.splice(0)
				});
		}
	};

	SortdReadReceipts.prototype.handleReadReceiptsReceived = function(message) {

		try {
			const readReceipts = message && message.data && message.data.readReceipts;

			if (readReceipts && readReceipts.length > 0) {
				this.receiptData.receipts = readReceipts;
			}

			if (this.receiptData.emitter) {
				this.receiptData.emitter.emit(this.receiptData.receipts);
			}
		} catch (ex) {
			this.handleError(ex);
		}
	};

	SortdReadReceipts.prototype.handleSetSortdUserReceived = function(message) {

		try {
			if (message && message.data && message.data.hashedUser) {
				this.hashedUser = message.data.hashedUser;
			}
		} catch (ex) {
			this.handleError(ex);
		}
	};

	SortdReadReceipts.prototype.formatMessageViewsDate = function(date) {

		let diff = Date.now() - date.getTime();
		diff = Math.round(diff / 60000);	// In minutes

		if (diff < 60) {	// Less than an hour
			return `${diff}m ago`;
		} else if (diff < 1440) {	// Less than a day
			return `${Math.round(diff / 60)}h ago`;
		} else if (diff < 10080) {	// Less than a week
			return `${Math.round(diff / 1440)}d ago`;
		} else {
			return `${constants.MONTHS[date.getMonth()]} ${date.getDate()}, ${date.getFullYear()}`;
		}
	};

	SortdReadReceipts.prototype.setActiveTab = function(tabId) {

		try {
			document.querySelectorAll('#sd_rr_message_views .sd_tab').forEach((tab) => tab.classList.remove('active'));
			document.querySelector(`#sd_rr_message_views .sd_tab[data-tabid="${tabId}"]`).classList.add('active');

			document.querySelectorAll('#sd_rr_message_views .sd_tab_content').forEach((tab) => tab.classList.remove('active'));
			document.querySelector(`#sd_rr_message_views .sd_tab_content[data-tabid="${tabId}"]`).classList.add('active');

			document.querySelectorAll('#sd_rr_message_views .sd_tab_text').forEach((tab) => tab.classList.remove('active'));
			document.querySelector(`#sd_rr_message_views .sd_tab_text[data-tabid="${tabId}"]`).classList.add('active');
		} catch (ex) {
			this.handleError(ex);
		}
	};

	SortdReadReceipts.prototype.tabClickHandler = function(event) {

		try {
			this.setActiveTab(event.currentTarget.dataset.tabid);
		} catch (ex) {
			this.handleError(ex);
		}
	};


	SortdReadReceipts.prototype.checkCePermissions = function(container) {

		try {
			if (!container) {
				return;
			}

			this.messages.sendMessageToEventPage(
				'hasCePermissions',
				{
					permissions: readReceiptConstants.CE.PERMISSIONS,
					origins: readReceiptConstants.CE.ORIGINS
				},
				(err, result) => {

					try {
						if (!(err || result)) {
							const id = 'sd_permissions_' + Math.floor(Math.random() * (90000));
							const div = document.createElement('div');
							div.innerHTML = readReceiptConstants.HTML.PERMISSION_REQUIRED
								.replace(/{{infoURL}}/g, this.config.chrome.sortd.readReceiptPermissionsURL);
							const permissions = div.firstElementChild;
							permissions.id = id;
							container.insertBefore(permissions, container.firstChild);
							// Below fixes 1st time position
							setTimeout(() => document.getElementById(id).style.display = 'block', 200);

							const clickEl = permissions.querySelector('button.sortdButton');
							if (clickEl) {
								clickEl.addEventListener('click', (evt) => {

									if (evt) {
										evt.preventDefault();
										evt.stopPropagation();
									}

									this.requestCePermissions(container);
								});
							}
						}
					} catch (ex) {
						this.handleError(ex);
					}
				});
		} catch (ex) {
			this.handleError(ex);
		}
	};

	SortdReadReceipts.prototype.requestCePermissions = function(container) {

		try {
			if (!container) {
				return;
			}

			this.messages.sendMessageToEventPage(
				'requestCePermissions',
				{
					permissions: readReceiptConstants.CE.PERMISSIONS,
					origins: readReceiptConstants.CE.ORIGINS
				},
				(err, result) => {

					try {
						if (!err && result) {
							container.querySelectorAll('.sd_permissions').forEach((el) => el.style.display = 'none');
							this.enableDesktopNotifications();
						} else {
							window.open(this.config.chrome.sortd.readReceiptPermissionsURL);
						}
					} catch (ex) {
						this.handleError(ex);
					}
				});
		} catch (ex) {
			this.handleError(ex);
		}
	};

	SortdReadReceipts.prototype.checkReadReceiptPermissions = function(container) {

		return this.sortdDataService.getReadReceiptEnabled()
			.then((result) => {

				this.readReceiptsEnabled = result;

				if (!this.readReceiptsEnabled) {
					container.querySelectorAll('.sd_not_in_current_plan').forEach((el) => {
						el.style.display = 'block';

						el.querySelector('.sd_clickable').addEventListener('click', (event) => {

							event.preventDefault();
							this.closeMessageViews();
							this.openSubscriptionPopupInSortd();
						});
					});
				}
			});
	};

	SortdReadReceipts.prototype.enableDesktopNotifications = function() {
		this.messages.sendMessageToFrame('enabledReadReceiptDesktopNotification');
	};

	SortdReadReceipts.prototype.openSubscriptionPopupInSortd = function() {

		this.messages.sendMessageToExtension('openPanelAndSendMessage', {
			action: 'isReadReceiptsEnabled',
			showPopup: true
		});
	};

	SortdReadReceipts.prototype.addTicksToSentNavMenu = function(retries) {

		try {
			const el = document.querySelector(readReceiptConstants.SENT_ITEMS_NAV.MENU_SELECTOR);
			if (el) {
				this.addSentItemNavMenuWatcher(el);
			} else if (retries > 0) {
				retries--;
				setTimeout(() => this.addTicksToSentNavMenu(retries), 1000);
			}
		} catch (ex) {
			this.handleError(ex);
		}
	};

	SortdReadReceipts.prototype.addSentItemNavMenuWatcher = function(el) {

		const item = el.querySelector(readReceiptConstants.SENT_ITEMS_NAV.SENT_SELECTOR);
		if (item) {
			this.addTicksToSentNavMenuElement(item);
		}

		const observer = new MutationObserver((mutations) => {
			mutations.forEach((mutation) => {

				if (mutation.addedNodes) {
					for (let i = 0; i < mutation.addedNodes.length; i++) {

						const node = mutation.addedNodes[i];
						if (node.tagName === 'DIV' && node.querySelector) {
							const itm = node.querySelector(readReceiptConstants.SENT_ITEMS_NAV.SENT_SELECTOR);
							if (itm) {
								this.addTicksToSentNavMenuElement(itm);
							}
						}
					}
				}
			});
		});

		observer.observe(el, {
			childList: true,
			subtree: true,
			attributes: false,
			characterData: false
		});
	};

	SortdReadReceipts.prototype.addTicksToSentNavMenuElement = function(el) {

		if (el) {
			const img = document.createElement('img');
			img.src = chrome.extension.getURL('img/readreceipts/viewed.svg');
			img.classList.add('sdrr_sent_item_ticks');
			el.parentElement.appendChild(img);
		}
	};

	return SortdReadReceipts;
}());