/**
 * Sortd
 * Copyright (c) 2019, Sortd Inc
 * https://www.sortd.com
 *
 * @fileoverview Defines the sortd shortcut helper
 * @name sortd-shortcut-helper
 */

class SortdShortcutHelper {

    constructor(options) {

        this.__FILE_NAME = 'shortcut-helper.js';

		this.errorHandler = options.errorHandler;
		this.messages = options.messages;

		this.registeredListener = false;
		this.listeners = {};
		this.displayMode = constants.PANEL_DISPLAY_MODES.SIDEBAR;
	}

	/**
	 * Sends the keydown event to the frame
	 * @param {Event} evt - The keydown event
	 * @private
	 */
	sendKeyDownEventToFrame(evt) {

		const event = {
			altKey: evt.altKey,
			ctrlKey: evt.ctrlKey,
			shiftKey: evt.shiftKey,
			keyCode: evt.keyCode,
			charCode: evt.charCode,
			key: evt.key,
			metaKey: evt.metaKey,
			repeat: evt.repeat,
			which: evt.which
		};

		this.messages.sendMessageToFrame(evt.type.toLocaleLowerCase() === 'keydown' ? 'keyDownEvent' : 'keyUpEvent', event);
	};

	/**
	 * Keydown handler.
	 * Only check keys in fullscreen mode.
	 * Allows only keys in the ALLOWED_CODES array to pass through to gmail, block the rest
	 * @param {Event} evt - The keydown event
	 * @param {Boolean} isKeyUp - True when this is called from the KeyUp event
	 * @private
	 */
	onKeyDownEvent(evt, isKeyUp) {

		try {
			let eventName;
			let captureEvent = !(isKeyUp || evt.dontCaptureSortd);

			let src;
			if (evt.originalEvent && evt.originalEvent.srcElement) {
				src = evt.originalEvent.srcElement;
			} else {
				src = evt.srcElement || {};
			}

			const srcType = src.type || '';
			const srcEditable = (src.contentEditable || false).toString();
			const srcRole = evt.srcElement && evt.srcElement.getAttribute && evt.srcElement.getAttribute('role');

			if (srcRole === 'textbox' || srcType === 'textarea' || srcType === 'text' || srcEditable === 'true' ||
				(srcType === 'submit' && evt.keyCode === constants.SHORTCUTS.KEY_CODES.ENTER)) {
				// Don't capture events in text areas or when user submits the form
				return true;
			}

			if (this.displayMode === constants.PANEL_DISPLAY_MODES.FULLSCREEN &&
				evt.ctrlKey !== true &&
				evt.metaKey !== true &&
				evt.altKey !== true &&
				constants.SHORTCUTS.ALLOWED_CODES.indexOf(evt.keyCode) === -1
			) {
				if (evt.stopImmediatePropagation) {
					evt.stopImmediatePropagation();
				}
				if (evt.stopPropagation) {
					evt.stopPropagation();
				}
				if (evt.preventDefault) {
					evt.preventDefault();
				}

				this.sendKeyDownEventToFrame(evt);

				if (constants.SHORTCUTS.FRAME_FOCUS.indexOf(evt.keyCode) !== -1) {
					this.messages.sendMessageToExtension('focusSortd');
				}

				return false;
			} else if (captureEvent) {

				let prefix = '';

				if (evt.shiftKey) {
					prefix += constants.SHORTCUTS.KEY_CODES.SHIFT + '_';
				}
				if (evt.ctrlKey) {
					prefix += constants.SHORTCUTS.KEY_CODES.CTRL + '_';
				}
				if (evt.altKey) {
					prefix += constants.SHORTCUTS.KEY_CODES.ALT + '_';
				}
				if (evt.metaKey) {
					prefix += constants.SHORTCUTS.KEY_CODES.META + '_';
				}

				eventName = 'shortcut_' + prefix + evt.keyCode;

				if (this.hasListener(eventName)) {

					const propagateEvent = this.callListeners(eventName, evt);

					if (propagateEvent !== true) {

						if (evt.stopImmediatePropagation) {
							evt.stopImmediatePropagation();
						}
						if (evt.stopPropagation) {
							evt.stopPropagation();
						}
					}
				}
			}
		} catch (ex) {
			this.errorHandler.handle(ex);
		}

		return true;
	}

	/**
	 * Handler for keypress event from the iframe
	 * @param message - The message that was received
	 * @private
	 */
	onKeyPressHandler(message) {

		const evt = message.data;

		if (this.displayMode !== constants.PANEL_DISPLAY_MODES.FULLSCREEN ||
			evt.ctrlKey === true || evt.metaKey === true || evt.altKey === true ||
			constants.SHORTCUTS.ALLOWED_CODES.indexOf(evt.keyCode) !== -1) {

			const oEvent = document.createEvent('KeyboardEvent');

			/**
			 * Chrome does not set "keyCode" or "which" when firing keyboard events
			 * https://bugs.webkit.org/show_bug.cgi?id=16735
			 *
			 * This hack is supposed to bypass above bug, but does not seem to work in gmail's case
			 * http://stackoverflow.com/questions/10455626/keydown-simulation-in-chrome-fires-normally-but-not-the-correct-key
			 */
			Object.defineProperty(oEvent, 'keyCode', {
				get: () => this.keyCodeVal
			});
			Object.defineProperty(oEvent, 'which', {
				get: () => this.keyCodeVal
			});
			Object.defineProperty(oEvent, 'charCode', {
				get: () => this.keyCodeVal
			});

			/**
			 * initKeyboardEvent does not match DOM 3 spec
			 * https://bugs.webkit.org/show_bug.cgi?id=13368
			 */
			oEvent.initKeyboardEvent('keypress',
				true,	// canBubbleArg
				true,	// cancelableArg
				document.defaultView,	// viewArg,
				evt.key,	// keyIdentifierArg,
				0,		// locationArg,
				evt.ctrlKey,	// ctrlKeyArg,
				evt.altKey,		// altKeyArg,
				evt.shiftKey,	// shiftKeyArg,
				evt.metaKey,	// metaKeyArg,
				false	// altGraphKeyArg
			);

			oEvent.keyCodeVal = evt.keyCode;

			document.dispatchEvent(oEvent);
		}
	}

	/**
	 * Update the displayMode
	 * @param message - The message that was received
	 * @private
	 */
	setDisplayMode(message) {

		if (message.data && message.data.displayMode) {
			this.displayMode = message.data.displayMode;
		}
	}

	/**
	 * @private
	 */
	hasListener(name) {

		if (this.listeners[name] && this.listeners[name].length > 0) {
			return true;
		} else {
			return false;
		}
	}

	/**
	 * @private
	 */
	callListeners(name, evt) {

		let result = true;

		if (this.listeners[name]) {
			this.listeners[name].forEach((cb) => result = result && cb(evt));
		}

		return result;
	}

	/**
	 * Initialize the keyboard listener
	 */
	init() {

		if (!this.registeredListener) {

			document.addEventListener('keydown', (evt) => {

				try {
					return this.onKeyDownEvent(evt, false);
				} catch (ex) {
					this.errorHandler.handle(ex);
				}
			});

			document.addEventListener('keyup', (evt) => {

				try {
					return this.onKeyDownEvent(evt, true);
				} catch (ex) {
					this.errorHandler.handle(ex);
				}
			});

			this.messages.addListener('keyPressEvent', this.onKeyPressHandler.bind(this));
			this.messages.addListener('setDisplayMode', this.setDisplayMode.bind(this));

			this.registeredListener = true;
		}
	}

	addListener(keyCode, cb) {

		const name = 'shortcut_' + keyCode;

		if (!this.listeners[name]) {
			this.listeners[name] = [];
		}
		this.listeners[name].push(cb);
	};	

	removeListener(keyCode, cb) {

		const name = 'shortcut_' + keyCode;

		if (this.listeners[name]) {
			const index = this.listeners[name].indexOf(cb);
			if (index !== -1) {
				this.listeners[name].splice(index, 1);
			}
		}
	}
}

SortdShortcutHelper.instance = null;

SortdShortcutHelper.getInstance = function(options) {

	if (SortdShortcutHelper.instance === null) {
		if (!options) {
			throw new Error('ShortcutHelper has not been initialized correctly');
		}
		SortdShortcutHelper.instance = new SortdShortcutHelper(options);
		SortdShortcutHelper.instance.init();
	}

	return SortdShortcutHelper.instance;
}
