/**
 * Sortd
 * Copyright (c) 2019, Sortd Inc
 * https://www.sortd.com
 *
 * @fileoverview Tracks events
 * @name sortd-tracker
 */

var SortdTracker = (function () {

	'use strict';

	/**
	 * Initialize the class
	 * @class SortdTracker
	 */
	function SortdTracker (options) {

		this.FILE_NAME = 'sortd-tracker.js';

		this._config = options.config;
		this._errorHandler = options.errorHandler;
		this._messages = options.messages;
		this._http = new SortdHttp();

		this._emailAddress = null;
		this._uniqueId = null;
		this._linkedEmails = null;

		this._eventsToTrack = [];
		this._propertiesToTrack = [];

		this._messages.addListener('getLinkedAccounts', this.handleGetLinkedAccounts.bind(this));
		this._messages.addListener('setPageTrackingInfo', this.handleSetPageTrackingInfo.bind(this));
	}

	/**
	 * Update the uniqueId
	 * @param {string} uniqueId - The uniqueId
	 */
	SortdTracker.prototype.updateUniqueId = function(uniqueId) {
		this._uniqueId = uniqueId && uniqueId.toLowerCase();
	};

	/**
	 *
	 * @param {string} event - The event to track
	 * @param {object} properties - Key/values
	 * @param {boolean} [updateLastActivity] - False if the last activity should not be updated
	 * @param {boolean} [trackOnce] - Defaults to false. True if the action should only be tracked once
	 * @private
	 */
	SortdTracker.prototype.track = function (event, properties, updateLastActivity, trackOnce) {

		const self = this;
		const options = {};

		try {
			if (!(this._uniqueId && this._emailAddress)) {
				this._eventsToTrack.push({ event, properties, updateLastActivity, trackOnce });
				return;
			}

			if (trackOnce === true && this.getIsTracked(event)) {
				return;
			}

			options.email = self._emailAddress;
			options.event = event;
			properties = properties || {};
			properties.email = self._emailAddress;
			properties.UniqueIdentifier = self._uniqueId;
			options['track-event-properties'] = properties;
			if (typeof updateLastActivity !== 'undefined' && updateLastActivity === false) {
				options['update-last-activity'] = false;
			}

			this._http.post(
				self._config.chrome.host.mail3.baseUrl + '/tracker/track',
				options,
				function (err) {

					if (err) {
						self._errorHandler.handle(err, self.FILE_NAME);
					} else {
						self.setIsTracked(event);
					}
				});
		} catch (ex) {
			self._errorHandler.handle(ex, self.FILE_NAME);
		}
	};

	SortdTracker.prototype.getIsTracked = function(event) {

		var self = this;
		var key = constants.STORAGE.KEYS.IS_TRACKED.replace(constants.STORAGE.EMAIL_PLACEHOLDER, self._emailAddress);
		var tracked = JSON.parse(localStorage.getItem(key));

		return (tracked && tracked[event]) || false;
	};

	SortdTracker.prototype.setIsTracked = function(event) {

		var self = this;
		var key = constants.STORAGE.KEYS.IS_TRACKED.replace(constants.STORAGE.EMAIL_PLACEHOLDER, self._emailAddress);
		var tracked = JSON.parse(localStorage.getItem(key)) || {};

		tracked[event] = true;
		localStorage.setItem(key, JSON.stringify(tracked));
	};

	/**
	 * Set a tracked userproperty
	 * @param {object} properties - Key/values : user properties to log
	 * @param {boolea} setOnce - True to only set the info once for the user (Default is false)
	 * @private
	 */
	SortdTracker.prototype.setUserProperty = function (properties, setOnce) {

		if (!this._emailAddress) {
			this._propertiesToTrack.push({ properties });
			return;
		}

		const self = this;
		const options = {};

		try {
			options.email = self._emailAddress;
			options.setOnce = setOnce || false;
			if (properties) {
				options['properties'] = properties;
			}

			this._http.post(
				self._config.chrome.host.mail3.baseUrl + '/tracker/setuserproperty',
				options,
				function (err) {

					if (err) {
						self._errorHandler.handle(err, self.FILE_NAME);
					}
				});
		} catch (ex) {
			self._errorHandler.handle(ex, self.FILE_NAME);
		}
	};

	/**
	 * Load the linked accounts from the background page
	 * @param {function} next - The results callback
	 * @private
	 */
	SortdTracker.prototype.loadLinkedAccounts = function (next) {

		try {
			this._messages.sendMessageToEventPage('getLinkedAccounts', {}, (err, linkedAccounts) => {

				if (err) {
					this._errorHandler.handle(err, this.FILE_NAME);
				}

				next(err || null, linkedAccounts);
			});
		} catch (ex) {
			this._errorHandler.handle(ex, this.FILE_NAME);
			next(ex);
		}
	};

	SortdTracker.prototype.updateEmailAddresses = function(emailAddress, linkedEmails, next) {

		this._emailAddress = emailAddress.toLowerCase();

		this.loadLinkedAccounts((err, linkedAccounts) => {

			try {
				if (err) {
					if (next) {
						next(null);
					}
					return;
				}

				let uniqueId = this._emailAddress;
				if (linkedAccounts && linkedAccounts.uniqueId) {
					uniqueId = linkedAccounts.uniqueId;
					if (linkedAccounts.linkedEmails && linkedAccounts.linkedEmails.length > 0) {
						linkedEmails = linkedEmails.concat(linkedAccounts.linkedEmails);
					}
				} else {
					let registeredEmails = localStorage.getItem(constants.STORAGE.KEYS.REGISTERED_EMAILS) || [];
					if (registeredEmails) {
						if (typeof registeredEmails === 'string') {
							registeredEmails = registeredEmails.split(',');
						}
						if (registeredEmails && registeredEmails.length > 0) {
							uniqueId = registeredEmails.shift();
							linkedEmails = linkedEmails.concat(registeredEmails);
						}
					}
				}

				// Remove the text "(default)" in various languages from the end of the address
				// but don't replace (***) inside the address
				uniqueId = uniqueId.toLowerCase().replace(/\(([^\)]*)\)[ \t]*$/, '').trim();
				this._linkedEmails = linkedEmails.map((e) => e.toLowerCase().replace(/\(([^\)]*)\)[ \t]*$/, '').trim());

				if (this._linkedEmails.indexOf(uniqueId) === -1) {
					this._linkedEmails.push(uniqueId);
				}

				this._http.post(
					this._config.chrome.host.mail3.baseUrl + '/getuniqueid',
					{ uniqueId, linkedEmails: this._linkedEmails },
					(postErr, result) => {

						try {
							if (postErr) {
								this._uniqueId = uniqueId;
								this._errorHandler.handle(postErr, this.FILE_NAME);
							} else if (result) {
								if (result.uniqueId) {
									this._uniqueId = result.uniqueId.toLowerCase();
									this._hasGsuiteAccount = result.hasGsuiteAccount;
									this.setUserProperty({
										UniqueIdentifier: this._uniqueId,
										HasGsuiteAccount: this._hasGsuiteAccount
									});
								} else {
									this.setUserProperty({
										UniqueIdentifier: this._uniqueId
									});
								}
						} else {
								this._uniqueId = uniqueId;
							}
						} catch (ex) {
							this._errorHandler.handle(ex, this.FILE_NAME);
						} finally {
							this.processUntrackedEvents();
							if (next) {
								next(null);
							}
						}
					});
			} catch (ex) {
				this._errorHandler.handle(ex, this.FILE_NAME);
				if (next) {
					next(ex);
				}
			}
		});

		this.setUserProperty({ email: this._emailAddress });
		this.processUntrackedProperties();
	};

	SortdTracker.prototype.handleGetLinkedAccounts = function(message, callback) {

		if (callback) {
			callback(null, { uniqueId: this._uniqueId, linkedEmails: this._linkedEmails });
		}
	};

	SortdTracker.prototype.processUntrackedEvents = function() {
		
		this._eventsToTrack.forEach((e) => {
			this.track(e.event, e.properties, e.updateLastActivity, e.trackOnce);
		});
	}

	SortdTracker.prototype.processUntrackedProperties = function() {
		
		this._propertiesToTrack.forEach((p) => {
			this.setUserProperty(p.properties);
		});
	}

	SortdTracker.prototype.handleSetPageTrackingInfo = function(message) {

		const data = message.data;

		if (data) {
			const properties = {};
			const trackingOverrides = this._config.trackingOverrides || {};


			if (trackingOverrides.landingPage) {
				properties['Landing Page'] = trackingOverrides.landingPage;
			} else if (data.landingpage) {
				properties['Landing Page'] = data.landingpage;
			}

			let source = data.source;
			if (!source) {
				if (data.landingpage) {
					source = constants.TRACKING_SOURCE.WEBSITE;
				} else {
					source = constants.TRACKING_SOURCE.DEFAULT;
				}
			}

			if (trackingOverrides.source) {
				properties['Source'] = trackingOverrides.source;
			} else {
				properties['Source'] = source;
			}

			if (trackingOverrides.medium) {
				properties['Medium'] = trackingOverrides.medium;
			} else if (data.medium) {
				properties['Medium'] = data.medium;
			}

			if (trackingOverrides.campaign) {
				properties['Campaign'] = trackingOverrides.campaign;
			} else if (data.campaign) {
				properties['Campaign'] = data.campaign;
			}

			if (data.trackingid) {
				properties['TrackingId'] = data.trackingid;
			}

			if (data.referrer) {
				properties['Referrer'] = data.referrer;
			}

			if (data.adgroup) {
				properties['AdGroup'] = data.adgroup;
			}

			if (data.searchterm) {
				properties['SearchTerm'] = data.searchterm;
			}

			if (data.keywordmatchtype) {
				properties['KeywordMatchType'] = data.keywordmatchtype;
			}

			if (data.sourcedevice) {
				properties['SourceDevice'] = data.sourcedevice;
			}

			if (this._config.ceType) {
				properties['Chrome Extension'] = this._config.ceType;
			}

			this.setUserProperty(properties, true);
		}
	}

	return SortdTracker;
}());