/**
 * Sortd
 * Copyright (c) 2021, Sortd Inc
 * https://www.sortd.com
 *
 * @fileoverview Functions to handle displaying fullscreen video overlay in Gmail
 * @name sortd-video-frame-wrapper
 */

var SortdVideoFrameWrapper = (function () {

	'use strict';

	function SortdVideoFrameWrapper () {

		this.FILE_NAME = 'sortd-video-frame-wrapper.js';
		this._errorHandler = new SortdErrorHandler();

		var myOverrides = {};
		if (typeof overrides !== 'undefined') {
			myOverrides = overrides;
		}
		this.config = new config(myOverrides);

		this.messageOriginExtension = 'chrome-extension://' + chrome.runtime.id;
		this.allowedMessageOrigins = [
			this.config.messageOriginGmail,
			this.config.chrome.host.mail3.baseUrl,
			this.messageOriginExtension
		];
	}

	SortdVideoFrameWrapper.prototype._handleError = function (err) {
		this._errorHandler.handle(err, this.FILE_NAME);
	};

	SortdVideoFrameWrapper.prototype.sendMessage = function (action, data, destination) {

		let targetOrigin = this.config.messageOriginGmail;
		if (destination === 'frame') {
			targetOrigin = this.messageOriginExtension;
		}

		window.parent.postMessage({
			action: action,
			owner: 'sortd',
			source: 'videoWrapper',
			destination: (destination || 'jsce'),
			data: data
		}, targetOrigin);
	};

	SortdVideoFrameWrapper.prototype.startVideo = function(data) {

		const video = document.createElement('video');
		video.autoplay = true;
		video.controls = true;
		video.playsinline = true;
		if (data.poster) {
			video.poster = data.poster;
		}

		if (data.mp4) {
			const videoSource = document.createElement('source');
			videoSource.type = 'video/mp4';
			videoSource.src = data.mp4;
			video.appendChild(videoSource);
		}

		document.body.appendChild(video);

		setTimeout(() => {
			document.body.classList.add('show');
		}, 0);
	};

	SortdVideoFrameWrapper.prototype.stopVideo = function(sendMessage) {

		const video = document.querySelector('video');
		video.pause();
		video.currentTime = 0;
		document.body.classList.remove('show');

		setTimeout(() => {
			video.remove();

			if (sendMessage) {
				this.sendMessage('videoStopped');
			}
		}, 500);
	};

	SortdVideoFrameWrapper.prototype.handleMessage = function (evt) {

		try {
			const message = evt.data;

			if (message && message.owner === 'sortd' && message.destination === 'videoWrapper' && this.allowedMessageOrigins.indexOf(evt.origin) !== -1) {

				switch (message.action) {
					case 'playVideo':
						this.startVideo(message.data);
						break;
					case 'stopVideo':
						this.stopVideo(false);
						break;
				}
			}
		} catch (ex) {
			this._handleError(ex);
		}
	};

	SortdVideoFrameWrapper.prototype.init = function () {

		document.addEventListener("DOMContentLoaded", () => {
			window.addEventListener('message', this.handleMessage.bind(this), false);
			document.getElementById('video_overlay').addEventListener('click', () => this.stopVideo(true));
		});
	}

	return SortdVideoFrameWrapper;
}());

(function () {
	new SortdVideoFrameWrapper().init();
}());
